#!/opt/edge/venv/bin/python3.12

from typing import Dict, Optional

from edge import eventlog
from edge.support import cgi
from edge.webserver.sysconf.api_datetime import TimeApi


class _HandlerFactory:
    """
    GET   /api/v2/sys/conf/time   /<uiref>
    PUT   /api/v2/sys/conf/time   /<uiref>
    PUT   /api/v2/sys/conf/time   /<uiref>/manual
    """

    @classmethod
    def create(cls, request: cgi.CgiRequest) -> cgi.CgiHandler:
        method_to_handlers_map: cgi.CgiHandlerFactoryDict = {
            'GET': [
                cls._time__get_uiref
            ],
            'PUT': [
                cls._time__put_uiref,
                cls._time__put_uiref_manual
            ]
        }

        handlers = method_to_handlers_map.get(request.method, [])
        for handler in handlers:
            try:
                return handler(request)

            except cgi.CgiError404:
                pass

        eventlog.log_d('fail: http404')
        raise cgi.CgiError404

    # noinspection PyUnusedLocal
    @classmethod
    def _time__get_uiref(cls, request: cgi.CgiRequest, limits: Optional[Dict] = None) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_absent()

        def handler() -> cgi.CgiResponse:
            unit_ext_jnode = TimeApi.load_unit()
            return cgi.CgiResponse(body=unit_ext_jnode)

        return handler

    # noinspection PyUnusedLocal
    @classmethod
    def _time__put_uiref(cls, request: cgi.CgiRequest, limits: Optional[Dict] = None) -> cgi.CgiHandler:
        if len(request.path_info) != 1:
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            unit_ext_jnode = TimeApi.save_unit(request.body_jnode)
            return cgi.CgiResponse(body=unit_ext_jnode)

        return handler

    # noinspection PyUnusedLocal
    @classmethod
    def _time__put_uiref_manual(cls, request: cgi.CgiRequest, limits: Optional[Dict] = None) -> cgi.CgiHandler:
        if len(request.path_info) != 2:
            raise cgi.CgiError404

        elif request.path_info[1] != 'manual':
            raise cgi.CgiError404

        request.assert_query_absent()
        request.assert_body_present()

        def handler() -> cgi.CgiResponse:
            sync_status = TimeApi.sync_manual(sync_source=request.body_jnode)
            return cgi.CgiResponse(body=sync_status)

        return handler


if __name__ == '__main__':
    eventlog.open_log('cgi-time', 'debug')
    try:
        cgi.cgi_handle(_HandlerFactory.create)
    finally:
        eventlog.close_log()
