import io

from typing import Tuple, Union, Optional, List, Final
from shapely import geometry, ops, GeometryCollection
from PIL import Image, ImageDraw, ImageFont

from edge.image.types.markings import MarkingsConf
from edge.image.types.footer import ImFooterConf
from edge.services.types.cam import ICamService
from edge.types.exceptions import ParamsError
from edge.types.image import ImRgba, ImTrans
from edge.types.shape.quad import QuadI
from edge.types.jnode.shape.quad import encode__quadi
from edge.types.shape.rect import RectF, RectI


class ImHelpers:
    @staticmethod
    def get_jpeg_size(jpeg_url: str) -> Tuple[int, int]:
        return Image.open(jpeg_url).size

    @staticmethod
    def get_jpeg_size2(jpeg_data: bytes) -> Tuple[int, int]:
        return Image.open(io.BytesIO(jpeg_data)).size

    @staticmethod
    def mk_jpeg(source_image: Image) -> bytes:
        jpeg_stream = io.BytesIO()
        try:
            rgb_image = source_image.convert('RGB')
            rgb_image.save(jpeg_stream, 'JPEG')
            return jpeg_stream.getvalue()
        finally:
            jpeg_stream.close()

    @classmethod
    def mk_jpeg2(cls, blob_data: ImRgba) -> bytes:
        return blob_data.with_image(lambda im: cls.mk_jpeg(im))

    @staticmethod
    def mk_png(source_image: Image) -> bytes:
        png_stream = io.BytesIO()
        try:
            source_image.save(png_stream, format='PNG')
            return png_stream.getvalue()
        finally:
            png_stream.close()

    @classmethod
    def mk_png2(cls, blob_data: ImRgba) -> bytes:
        return blob_data.with_image(lambda im: cls.mk_png(im))

    @staticmethod
    def mk_resize_image(
            source_data: ImRgba,
            target_w: int | float,
            target_h: int | float,
            resample: int = Image.Resampling.NEAREST) -> Image:

        target_wh = int(target_w), int(target_h)
        with Image.frombuffer(
                'RGBA', source_data.wh, source_data.buffer,
                'raw', 'RGBA', source_data.stride, 1) as crop_im:

            return crop_im.resize(target_wh, resample)

    @staticmethod
    def mk_resize_jpeg(
            blob_data: ImRgba,
            target_w: Union[int, float],
            target_h: Union[int, float],
            resample: int = Image.Resampling.NEAREST) -> bytes:

        with ImHelpers.mk_resize_image(blob_data, target_w, target_h, resample) as pil_im:
            return ImHelpers.mk_jpeg(pil_im)

    @staticmethod
    def mk_resize_trans(source_data: ImRgba,
                        target_w: int | float,
                        target_h: int | float) -> ImTrans:

        #
        # (x', y') - scene-im point; (x'', y'') - recrop-image point.
        #
        # |x''|   | Ix Jx Ox |   |x'|
        # |y''| = | Iy Jy Oy | * |y'|
        # |1  |   | 0  0  1  |   |1 |
        #
        #     scale           move              trans
        # | Ix  0   0 |   | 1   0   0  |   | Ix  0   0  |
        # | 0   Jy  0 | * | 0   1   0  | = | 0   Jy  0  |
        # | 0   0   1 |   | 0   0   1  |   | 0   0   1  |
        #

        target_trans = ImTrans(
            ix=target_w / float(source_data.w),
            jy=target_h / float(source_data.h))

        return target_trans

    @staticmethod
    def mk_crop_resize_image(
            source_data: ImRgba,
            crop_rect: RectI,
            target_w: Union[int, float],
            target_h: Union[int, float],
            resample: int = Image.Resampling.NEAREST) -> Image:

        target_wh = int(target_w), int(target_h)
        return source_data.with_crop_image(crop_rect, lambda crop_im: crop_im.resize(target_wh, resample))

    @staticmethod
    def mk_crop_resize_jpeg(
            blob_data: ImRgba,
            crop_rect: RectF,
            target_w: Union[int, float],
            target_h: Union[int, float],
            resample: int = Image.Resampling.NEAREST) -> bytes:

        crop_recti = RectI(*crop_rect.ltrb)
        with ImHelpers.mk_crop_resize_image(blob_data, crop_recti, target_w, target_h, resample) as pil_im:
            return ImHelpers.mk_jpeg(pil_im)

    @staticmethod
    def mk_crop_resize_trans(crop_rect: RectF, target_w: Union[int, float], target_h: Union[int, float]) -> ImTrans:
        #
        # (x', y') - scene-im point; (x'', y'') - recrop-image point.
        #
        # |x''|   | Ix Jx Ox |   |x'|
        # |y''| = | Iy Jy Oy | * |y'|
        # |1  |   | 0  0  1  |   |1 |
        #
        #     scale           move              trans
        # | Ix  0   0 |   | 1   0   Ox |   | Ix  0   Ox*Ix |
        # | 0   Jy  0 | * | 0   1   Oy | = | 0   Jy  Oy*Jy |
        # | 0   0   1 |   | 0   0   1  |   | 0   0   1     |
        #

        ix: float = target_w / crop_rect.fw
        jy: float = target_h / crop_rect.fh
        return ImTrans(ix=ix, jy=jy, ox=-crop_rect.fl * ix, oy=-crop_rect.ft * jy)

    @staticmethod
    def draw_rect(pil_image: Image, rect: RectF, rect_trans: Optional[ImTrans] = None, color: str = 'red', line_w: int = 1) -> None:
        ltrb = [rect.fl, rect.ft, rect.fr, rect.fb]
        if rect_trans:
            ltrb = rect_trans.map_ltrb(ltrb)

        ImageDraw.Draw(pil_image).rectangle(ltrb, outline=color, width=line_w)

    @staticmethod
    def draw_poly(pil_image: Image, quad: QuadI, rect_trans: Optional[ImTrans] = None) -> None:
        if rect_trans:
            poly = encode__quadi(rect_trans.map_quad(quad))
        else:
            poly = encode__quadi(quad)

        ImageDraw.Draw(pil_image).polygon(poly, outline='red')

    @staticmethod
    def draw_footer(pil_image: Image, footer_args: ImFooterConf, font: ImageFont, font_size: int) -> None:
        draw = ImageDraw.Draw(pil_image)
        draw.rectangle([(0, 0), pil_image.size], fill='black')

        y_delta: float = 0.4 * font_size  # top-margin
        start_y: float = y_delta / 2.0

        leter_width: int = font.getlength('M')

        for row in footer_args.rows:
            start_x: float = leter_width / 2.0  # left-margin = 0.5em
            for column in row.cols:
                column_width: int = column.width
                column_text: str = column.template
                column_align: int = column.align

                if column_align == -1:
                    x_delta: float = 0.0

                elif column_align == 0:
                    text_width = font.getlength(column_text)
                    x_delta: float = (column_width - text_width) / 2.0 - leter_width / 2.0

                elif column_align == 1:
                    text_width = font.getlength(column_text)
                    x_delta: float = float(column_width - text_width - leter_width)

                else:
                    raise NotImplementedError

                draw.text((start_x + x_delta, start_y), column_text, font=font)
                start_x = start_x + column_width

            start_y = start_y + font_size + y_delta

    @staticmethod
    def draw_track(im: Image, trace: List[Tuple[float, float]], skip_zones: List[RectF]):

        skip_polygons: List[geometry.Polygon] = []
        for zone in skip_zones:
            skip_polygons.append(geometry.Polygon([
                (zone.fl, zone.ft),
                (zone.fr, zone.ft),
                (zone.fr, zone.fb),
                (zone.fl, zone.fb)
            ]))

        skip_multipolygon = ops.unary_union(geometry.MultiPolygon(skip_polygons))

        traceline = geometry.LineString(trace)
        trancelines: GeometryCollection = ops.split(traceline, skip_multipolygon)

        im_draw = ImageDraw.Draw(im)
        for line in trancelines.geoms:
            if skip_multipolygon.contains(line):
                continue

            im_draw.line(line.coords, fill='white')
            for pt in line.coords:
                im_draw.rectangle((pt[0] - 2, pt[1] - 2, pt[0] + 2, pt[1] + 2), fill='red')

    @staticmethod
    def pixelate_roi(image: Image, bbox: RectI, pixel_size: int = 10):
        # Crop the image to the region of interest
        cropped_image = image.crop(bbox.ltrb)

        # Resize the cropped image to the desired pixel size
        resized_image = cropped_image.resize((bbox.iw // pixel_size, bbox.ih // pixel_size), Image.Resampling.NEAREST)

        # Scale the resized image back up to the original size
        pixelated_image = resized_image.resize(cropped_image.size, Image.Resampling.NEAREST)

        # Replace the region of interest in the original image with the pixelated version
        image.paste(pixelated_image, bbox.ltrb)


class LineDrawer:
    def __init__(self,
                 target_draw: ImageDraw,
                 target_w: int, target_h: int,
                 source_w: int, source_h: int) -> None:
        self._target_draw: Final[ImageDraw] = target_draw
        self._target_x_ratio: Final[float] = float(target_w) / source_w
        self._target_y_ratio: Final[float] = float(target_h) / source_h

    def draw_line(self,
                  source_line: tuple[int, int, int, int],
                  color: tuple[int, int, int]) -> None:

        source_x1, source_y1, source_x2, source_y2 = source_line
        target_line = (
            float(source_x1) * self._target_x_ratio,
            float(source_y1) * self._target_y_ratio,
            float(source_x2) * self._target_x_ratio,
            float(source_y2) * self._target_y_ratio)

        self._target_draw.line(target_line, color, width=1)


class MarkingsDrawer:
    def __init__(self, cam: ICamService, conf: MarkingsConf, source_w: int, source_h: int) -> None:
        self._line_to_color_map = dict()

        try:
            for line2 in conf.lines2d_f:
                p0_x, p0_y = int(line2.ref_point.x * source_w), int(line2.ref_point.y * source_h)
                p0_f, p0_l = cam.mk_world_point(p0_x, p0_y, 0)
                p1_x, p1_y = cam.mk_image_point(p0_l, 0, p0_f + line2.dist3)
                self._line_to_color_map[(p0_x, p0_y, p1_x, p1_y)] = line2.rgb

            for line3 in conf.lines3d_f:
                p0_x, p0_y = cam.mk_image_point(line3.ref_point.l, line3.ref_point.u, line3.ref_point.f)
                p1_x, p1_y = cam.mk_image_point(line3.ref_point.l, line3.ref_point.u, line3.ref_point.f + line3.dist3)
                self._line_to_color_map[(p0_x, p0_y, p1_x, p1_y)] = line3.rgb

            for line2 in conf.lines2d_l:
                p0_x, p0_y = int(line2.ref_point.x * source_w), int(line2.ref_point.y * source_h)
                p0_f, p0_l = cam.mk_world_point(p0_x, p0_y, 0)
                p1_x, p1_y = cam.mk_image_point(p0_l + line2.dist3, 0, p0_f)
                self._line_to_color_map[(p0_x, p0_y, p1_x, p1_y)] = line2.rgb

            for line3 in conf.lines3d_l:
                p0_x, p0_y = cam.mk_image_point(line3.ref_point.l, line3.ref_point.u, line3.ref_point.f)
                p1_x, p1_y = cam.mk_image_point(line3.ref_point.l + line3.dist3, line3.ref_point.u, line3.ref_point.f)
                self._line_to_color_map[(p0_x, p0_y, p1_x, p1_y)] = line3.rgb

        except ParamsError:
            pass

    def draw_markings(self, drawer: LineDrawer) -> None:
        for line, color in self._line_to_color_map.items():
            drawer.draw_line(line, color)
