from typing import Tuple, Optional

from edge import eventlog


class WhScaler:
    def __init__(self, max_wh: Tuple[int, int] = None):
        self._max_wh = max_wh

        self._last_source_wh: Optional[Tuple[int, int]] = None
        self._last_target_wh: Optional[Tuple[int, int]] = None

    def _get_target_wh(self, source_wh: Tuple[int, int]) -> Optional[Tuple[int, int]]:
        # resolution constraints are absent
        if self._max_wh is None:
            eventlog.log_i(f'scale images from {source_wh} to {source_wh}')
            return None

        # resolution constraints are met
        source_w, source_h = source_wh
        max_w, max_h = self._max_wh
        if source_w <= max_w and source_h <= max_h:
            eventlog.log_i(f'scale images from {source_wh} to {source_wh}')
            return None

        # define resize-ratio to meet resolution constraints
        exact_width_ratio: float = float(max_w) / source_w
        exact_height_ratio: float = float(max_h) / source_h
        exact_resize_ratio: float = exact_height_ratio if exact_height_ratio < exact_width_ratio else exact_width_ratio

        # align target image width to 16 px. it's to be more friendy with quicksync-h264-encoder.
        _ALIGN_BOUNDARY_PX: int = 16
        _ALIGN_BOUNDARY_SEARCH_DEPTH: int = 3 * _ALIGN_BOUNDARY_PX

        source_wh_ratio: float = float(source_w) / source_h
        final_target_wigth: int = int(source_w * exact_resize_ratio) // _ALIGN_BOUNDARY_PX * _ALIGN_BOUNDARY_PX
        final_target_height: int = int(final_target_wigth / source_wh_ratio)

        aligned_target_wigth: int = final_target_wigth
        while aligned_target_wigth >= (final_target_wigth - _ALIGN_BOUNDARY_SEARCH_DEPTH):
            aligned_target_height: float = aligned_target_wigth / source_wh_ratio
            if (aligned_target_height % 1) == 0:
                final_target_wigth = aligned_target_wigth
                final_target_height = int(aligned_target_height)
                break

            aligned_target_wigth -= _ALIGN_BOUNDARY_PX

        if (final_target_height % 1) == 0:
            target_size: Tuple[int, int] = final_target_wigth, int(final_target_height)
        else:
            target_size: Tuple[int, int] = final_target_wigth, int(final_target_height) + 1

        eventlog.log_d(
            'source={} -> max-target={} -> pre-target={} -> target={}'
            .format(source_wh, self._max_wh, (final_target_wigth, final_target_height), target_size))

        eventlog.log_i(f"scale images from {source_wh} to {target_size}")

        return target_size

    def scaledown(self, source_wh: Tuple[int, int]) -> Tuple[int, int]:
        if source_wh != self._last_source_wh:
            self._last_target_wh = self._get_target_wh(source_wh)
            self._last_source_wh = source_wh

        if self._last_target_wh:
            return self._last_target_wh

        return source_wh
