from datetime import datetime
from typing import Any, Dict, Final, List, Self, Optional

from edge.services.types.about import IAboutService
from edge.services.types.tz import ITzService
from edge.support.speed import map_speed_uint
from edge.support.template import FormatTemplate
from edge.types.dir import Dir
from edge.types.exceptions import NotFoundError
from edge.types.geoposition import Geoposition
from edge.types.jnode.dir import encode_dir_text
from edge.types.location import Location
from edge.types.trackset.spotspeed import SpotSpeed


class ImFooterColConf:
    def __init__(self, width: int, align: int, template: str):
        self.width = width
        self.align = align
        self.template = template

    def eval(self, mapping: Dict[str, str]) -> Self:
        value = FormatTemplate(self.template).substitute(mapping)
        return ImFooterColConf(self.width, self.align, value)


class ImFooterRowConf:
    def __init__(self, enabled: bool, cols: List[ImFooterColConf]):
        self.enabled = enabled
        self.cols = cols

    def eval(self, mapping: Dict[str, str]) -> Self:
        cols = [col.eval(mapping) for col in self.cols]
        return ImFooterRowConf(self.enabled, cols)


class IImFooterVars:
    def get_value(self, var_name: str) -> str:
        raise NotFoundError()


class DefImFooterVars(IImFooterVars):
    def __init__(self, defvars: Dict[str, Any]):
        self._defvars = defvars

    def get_value(self, var_name: str) -> str:
        var_value = self._defvars.get(var_name)
        return '' if var_value is None else var_value


class ImFooterVars(IImFooterVars):
    def __init__(self,
                 tz_service: ITzService,
                 about_service: IAboutService,
                 main_ts: int,
                 time_sync: bool = False,
                 control_ts: Optional[int] = None,
                 defvars: Optional[Dict] = None,
                 location: Optional[Location] = None,
                 geolocation: Optional[Geoposition] = None,
                 plate_text: Optional[str] = None,
                 direction: Dir = Dir.CODE_U,
                 avg_direction: Dir = Dir.CODE_U,

                 speed_limit: Optional[int] = None,

                 section_speed: Optional[int] = None,
                 section_overspeed: Optional[int] = None,

                 spot_speed: Optional[int] = None,  # TODO: remove
                 spot_speed_main: Optional[SpotSpeed] = None,
                 spot_speed_control: Optional[SpotSpeed] = None,
                 spot_speed_ra_n: Optional[int] = None,
                 spot_speed_cam_n: Optional[int] = None,
                 spot_speed_cam_2: Optional[int] = None,
                 spot_speed_wim: Optional[int] = None,

                 spot_overspeed_main: Optional[int] = None,
                 spot_overspeed_ra_n: Optional[int] = None,
                 spot_overspeed_wim: Optional[int] = None,

                 halt_time: Optional[int] = None,
                 forward_cm: Optional[int] = None,
                 mmr_class: Optional[str] = None):

        self._parent: Final[IImFooterVars] = DefImFooterVars(defvars or dict())  # TODO: pass to ctor

        self._time_sync: Final[bool] = time_sync
        self._main_dt_local: Final[datetime] = tz_service.map_ts_usec_to_datetime_local(main_ts)  # TODO: pass to ctor
        self._control_dt_local: Final[datetime | None] = control_ts and tz_service.map_ts_usec_to_datetime_local(  # TODO: pass to ctor
            control_ts)

        self._hw_vendor: Final[str] = about_service.hw_vendor  # TODO: pass to ctor
        self._hw_model: Final[str] = about_service.hw_model  # TODO: pass to ctor
        self._hw_serial: Final[str] = about_service.hw_serial  # TODO: pass to ctor

        self._plate_text: Final[str | None] = plate_text
        self._location: Final[Location | None] = location
        self._geoposition: Final[Geoposition | None] = geolocation
        self._direction: Final[Dir] = direction
        self._avg_direction: Final[Dir] = avg_direction
        self._halt_time: Final[int | None] = halt_time

        self._spot_speed_main: Final[int | None] = map_speed_uint(spot_speed_main and spot_speed_main.measured)
        self._spot_speed_control: Final[int | None] = map_speed_uint(spot_speed_control and spot_speed_control.measured)

        self._spot_speed_ra_n: Final[int | None] = map_speed_uint(spot_speed_ra_n)
        self._spot_speed_cam_2: Final[int | None] = map_speed_uint(spot_speed_cam_2)
        self._spot_speed_cam_n: Final[int | None] = map_speed_uint(spot_speed_cam_n)
        self._spot_speed: Final[int | None] = map_speed_uint(spot_speed)
        self._spot_speed_wim: Final[int | None] = map_speed_uint(spot_speed_wim)

        self._spot_overspeed_main: Final[int | None] = spot_overspeed_main
        self._spot_overspeed_ra_n: Final[int | None] = spot_overspeed_ra_n
        self._spot_overspeed_wim: Final[int | None] = spot_overspeed_wim

        self._section_speed: Final[int | None] = section_speed
        self._section_overspeed: Final[int | None] = section_overspeed

        self._speed_limit: Final[int | None] = speed_limit
        self._forward_cm: Final[int | None] = forward_cm

        self._mmr_class: Final[str | None] = mmr_class

    def get_value(self, var_name: str) -> str:
        if var_name == 'datetime':
            # <22.01.2018 17:00:33>
            return self._main_dt_local.strftime('%d.%m.%Y %H:%M:%S')

        elif var_name == 'datetime-ms':
            # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
            return self._main_dt_local.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        elif var_name == 'datetime-ms-gnss':
            if self._geoposition is None or self._time_sync is False:
                return ''
            else:
                # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
                return self._main_dt_local.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        elif var_name == 'datetime-ms-sync':
            if self._time_sync is False:
                return ''
            else:
                # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
                return self._main_dt_local.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        elif var_name == 'datetime-control':
            if self._control_dt_local is None:
                return ''
            else:
                # <22.01.2018 17:00:33>
                return self._control_dt_local.strftime('%d.%m.%Y %H:%M:%S')

        elif var_name == 'datetime-control-ms':
            if self._control_dt_local is None:
                return ''
            else:
                # srtip 3 last digits of microsecond to get string like this <22.01.2018 17:00:33.544>
                return self._control_dt_local.strftime('%d.%m.%Y %H:%M:%S.%f')[:-3]

        elif var_name == 'time-control-ms':
            if self._control_dt_local is None:
                return ''
            else:
                # srtip 3 last digits of microsecond to get string like this <17:00:33.544>
                return self._control_dt_local.strftime('%H:%M:%S.%f')[:-3]

        elif var_name == 'post-name':
            if self._location is None or self._location.place == '':
                return self._parent.get_value(var_name)
            else:
                return self._location.place

        elif var_name == 'lane':
            if self._location is None or self._location.lane == '':
                return self._parent.get_value(var_name)
            else:
                return self._location.lane

        elif var_name == 'geolocation':
            return str(self._geoposition) if self._geoposition else self._parent.get_value(var_name)

        elif var_name == 'geolocation-sync':
            return str(self._geoposition) if  (self._geoposition and self._time_sync) else self._parent.get_value(var_name)

        elif var_name == 'plate':
            return self._plate_text or self._parent.get_value(var_name)

        elif var_name == 'speed':
            return str(int(abs(self._spot_speed or 0)) or self._parent.get_value(var_name))

        elif var_name == 'dir':
            alias_name = encode_dir_text(self._direction)
            return self._parent.get_value(alias_name) or alias_name

        elif var_name == 'dir-avg':
            alias_name = encode_dir_text(self._avg_direction)
            return self._parent.get_value(alias_name) or alias_name

        elif var_name == 'section-speed':
            return str(self._section_speed or self._parent.get_value(var_name))

        elif var_name == 'section-overspeed':
            return str(self._section_overspeed or self._parent.get_value(var_name))

        elif var_name == 'main-speed':
            return str(self._spot_speed_main or self._parent.get_value(var_name))

        elif var_name == 'main-speed2':
            if self._spot_speed_main is None:
                return self._parent.get_value(var_name)
            elif self._spot_speed_main < 20:
                return '<20'
            elif self._spot_speed_main > 240:
                return '>240'
            else:
                return str(self._spot_speed_main)

        elif var_name == 'control-speed':
            return str(self._spot_speed_control or self._parent.get_value(var_name))

        elif var_name == 'radar-speed':
            return str(self._spot_speed_ra_n or self._parent.get_value(var_name))

        elif var_name == 'radar-speed2':
            if self._spot_speed_ra_n is None:
                return self._parent.get_value(var_name)
            elif self._spot_speed_ra_n < 20:
                return '<20'
            elif self._spot_speed_ra_n > 240:
                return '>240'
            else:
                return str(self._spot_speed_ra_n)

        elif var_name == 'video-speed':
            return str(self._spot_speed_cam_2 or self._parent.get_value(var_name))

        elif var_name == 'video-speed2':
            return str(self._spot_speed_cam_n or self._parent.get_value(var_name))

        elif var_name == 'wim-speed':
            return str(self._spot_speed_wim or self._parent.get_value(var_name))

        elif var_name == 'wim-overspeed':
            return str(self._spot_overspeed_wim or self._parent.get_value(var_name))

        elif var_name == 'main-overspeed':
            return str(self._spot_overspeed_main or self._parent.get_value(var_name))

        elif var_name in ('speed-limit', 'max-speed'):
            return str(self._speed_limit or self._parent.get_value(var_name))

        elif var_name == 'radar-overspeed':
            return str(self._spot_overspeed_ra_n or self._parent.get_value(var_name))

        elif var_name == 'vendor':
            return self._hw_model

        elif var_name == 'model':
            return self._hw_model

        elif var_name == 'serial':
            return self._hw_serial

        elif var_name == 'halt-time':
            return str(self._halt_time)

        elif var_name == 'position-f':
            # forward_cm is in cm, convert to m and take 2 symbols after point
            if self._forward_cm is None:
                return self._parent.get_value(var_name)

            return str(round(self._forward_cm / 100, 2))

        elif var_name == 'mmr-class':
            if self._mmr_class:
                return self._mmr_class
            else:
                return self._parent.get_value(var_name)

        else:
            raise NotFoundError()


class ImFooterConf:
    def __init__(self, height: int, font_size: int, rows: List[ImFooterRowConf]):
        """
        :param height:
        :param font_size:
            Font size, pixels
        :param rows:
        """
        self.height = height
        self.font_size = font_size
        self.rows = rows

    def eval(self, footer_vars: IImFooterVars) -> Self:
        """
        :raises: NotFoundError
        """
        var_names = set()
        for row in self.rows:
            for col in row.cols:
                var_names.update(FormatTemplate.get_pholders(col.template))

        mapping: Dict[str, str] = dict()
        for var_name in var_names:
            mapping[var_name] = footer_vars.get_value(var_name)

        rows = [r.eval(mapping) for r in self.rows]
        return ImFooterConf(self.height, self.font_size, rows)
