import ctypes
from glob import glob
from typing import Final, Callable

from edge import eventlog
from edge.services.types.cam import CamServiceConf, ICamService, SpeedcamConf
from edge.support.speed import map_speed_int, SpeedEstimateError, SpeedValidateError, get_speed_rel_error3, \
    get_speed_abs_error
from edge.types.exceptions import InternalError, ParamsError
from edge.types.track import LrTrackItem
from edge.types.dir import Dir


class CQuad(ctypes.Structure):
    _fields_ = [
        ('x0', ctypes.c_int), ('y0', ctypes.c_int),
        ('x1', ctypes.c_int), ('y1', ctypes.c_int),
        ('x2', ctypes.c_int), ('y2', ctypes.c_int),
        ('x3', ctypes.c_int), ('y3', ctypes.c_int)
    ]


class CPoint(ctypes.Structure):
    _fields_ = [
        ('x', ctypes.c_int), ('y', ctypes.c_int)
    ]


kLLM_CAM_STAT__DONE = 0
kLLM_CAM_STAT__FAIL_ESTIMATE = 1
kLLM_CAM_STAT__FAIL_VALIDATE = 2

LOG_N = 0
LOG_E = 1
LOG_W = 2
LOG_I = 3
LOG_D = 4
LOG_T = 5

LOG_FLAGS_NONE = 0
LOG_FLAGS_STDERR = 1


class _Api:
    _lib = ctypes.CDLL(glob('/opt/edge/lib/liblandmark*.so')[0])

    # bool
    # LlmCamCreate3(const char *aCamIntrinsicConf, const char *aCamExtrinsicConf, llm_cam_t *aSelf)
    camCreate3 = _lib.LlmCamCreate3
    camCreate3.argtypes = (
        ctypes.c_char_p, ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_void_p))

    camCreate3.restype = ctypes.c_bool

    # bool
    # LlmCamCreate5(const char *aReprojectionConf, llm_cam_t *aSelf)
    camCreate5 = _lib.LlmCamCreate5
    camCreate5.argtypes = (
        ctypes.c_char_p,
        ctypes.POINTER(ctypes.c_void_p))

    camCreate5.restype = ctypes.c_bool

    # void LlmCamDestroy(void* aSelf);
    camDestroy = _lib.LlmCamDestroy
    camDestroy.argtypes = (ctypes.c_void_p,)

    # bool
    # LlmCamMkImagePoint(
    #     void *aSelf,
    #     double aSceneL,
    #     double aSceneU,
    #     double aSceneF,
    #     double *anImageL,
    #     double *anImageT);
    mkImagePoint = _lib.LlmCamMkImagePoint
    mkImagePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double),
        ctypes.POINTER(ctypes.c_double))
    mkImagePoint.restype = ctypes.c_bool

    # bool
    # LlmCamMkScenePointU(void *aSelf,
    #     double aFrameL, double aFrameT, double aSceneU,
    #     double *aSceneL, double *aSceneF);
    mkScenePoint = _lib.LlmCamMkScenePointU
    mkScenePoint.argtypes = (
        ctypes.c_void_p, ctypes.c_double, ctypes.c_double, ctypes.c_double,
        ctypes.POINTER(ctypes.c_double), ctypes.POINTER(ctypes.c_double))

    mkScenePoint.restype = ctypes.c_bool

    # int
    # LlmCamAssertPointSpeed(llm_cam_t aSelf,
    #     const char *aPlateText, uint64_t aPointTs1, uint64_t aPointTs2,
    #     struct llm_point *aPoint1, struct llm_point *aPoint2, double aPointU,
    #     bool aMainDirU,
    #     double aMainSpeed, double *aControlSpeed);
    assertPointSpeed = _lib.LlmCamAssertPointSpeed
    assertPointSpeed.argtypes = (
        ctypes.c_void_p,
        ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(CPoint), ctypes.POINTER(CPoint), ctypes.c_double,
        ctypes.c_bool, ctypes.c_double, ctypes.POINTER(ctypes.c_double))

    assertPointSpeed.restype = ctypes.c_int

    # int
    # LlmCamAssertPlateSpeed(llm_cam_t aSelf,
    #     const char *aPlateText, uint64_t aPlateTs1, uint64_t aPlateTs2,
    #     struct llm_quad *aPlateQ1, struct llm_quad *aPlateQ2, double *aPlateU,
    #     bool aMainDirU,
    #     double aMainSpeed, double *aControlSpeed);
    assertPlateSpeed = _lib.LlmCamAssertPlateSpeed
    assertPlateSpeed.argtypes = (
        ctypes.c_void_p,
        ctypes.c_char_p, ctypes.c_uint64, ctypes.c_uint64,
        ctypes.POINTER(CQuad), ctypes.POINTER(CQuad), ctypes.POINTER(ctypes.c_double),
        ctypes.c_bool, ctypes.c_double, ctypes.POINTER(ctypes.c_double))

    assertPlateSpeed.restype = ctypes.c_int

    # void
    # LlmLogOpen(const char *aName, int aMaxLevel, int aFlags);
    logopen = _lib.LlmLogOpen
    logopen.restype = ctypes.c_void_p
    logopen.argtypes = (ctypes.c_char_p, ctypes.c_int, ctypes.c_int)


class _SpeedcamAlgoStats:
    def __init__(self) -> None:
        self._success_count: int = 0
        self._failure_count: int = 0

        self._failure_estimate_count: int = 0
        self._failure_validate_count: int = 0
        self._hist_abs: list[int] = [0] * 11
        self._hist_rel: list[int] = [0] * 11

    @property
    def count(self) -> [int, int, int]:
        """
        :returns:
            [total, success, failure]
        """
        return self._success_count + self._failure_count, self._success_count, self._failure_count

    def inc_success(self, main_speed: int, control_speed: int, main_dir_u: bool) -> None:
        self._success_count += 1
        if main_speed < 40:
            abs_error = get_speed_abs_error(main_speed, control_speed, main_dir_u)
            if abs_error > 10:
                self._hist_abs[10] += 1
            else:
                self._hist_abs[int(abs_error)] += 1
        else:
            rel_error = get_speed_rel_error3(main_speed, control_speed, main_dir_u)
            if rel_error > 10:
                self._hist_rel[10] += 1
            else:
                self._hist_rel[int(rel_error)] += 1

    def inc_failure_estimate(self) -> None:
        self._failure_count += 1
        self._failure_estimate_count += 1

    def inc_failure_validate(self) -> None:
        self._failure_count += 1
        self._failure_validate_count += 1

    def encode(self) -> dict:
        jnode = {
            'count': self.count,
            'hist-abs': self._hist_abs,
            'hist-rel': self._hist_rel,
            'errors': {
                'estimate': self._failure_estimate_count,
                'validate': self._failure_validate_count
            }
        }

        return jnode


class _SpeedcamStats:
    def __init__(self) -> None:
        self.plate_a = _SpeedcamAlgoStats()
        self.plate_r = _SpeedcamAlgoStats()

        self.point_a = _SpeedcamAlgoStats()
        self.point_r = _SpeedcamAlgoStats()

    def encode(self) -> dict:
        return {
            'approaching': {
                'plate': self.plate_a.encode(),
                'point': self.point_a.encode()
            },
            'receding': {
                'plate': self.plate_r.encode(),
                'point': self.point_r.encode()
            }
        }


class _CamService(ICamService):
    def __init__(self, conf: CamServiceConf) -> None:
        self._handle = ctypes.c_void_p(0)
        answer = _Api.camCreate5(
            bytes(conf.repro_path, 'ascii'),
            self._handle)

        if not answer:
            eventlog.log_e('fail: camCreate5 (repro:<{}>)'.format(
                conf.repro_path))
            raise InternalError

        self._speedcam2_conf: Final[SpeedcamConf] = conf.speedcam2
        self._speedcam2_stats = _SpeedcamStats()

    def destroy(self) -> None:
        _Api.camDestroy(self._handle)

    def mk_image_point(self, scene_l: float, scene_u: float, scene_f: float) -> tuple[int, int]:
        image_l = ctypes.c_double()
        image_t = ctypes.c_double()
        success: bool = _Api.mkImagePoint(self._handle, scene_l, scene_u, scene_f, image_l, image_t)
        if not success:
            eventlog.log_i('fail: mkImagePoint')
            raise ParamsError
        return int(image_l.value), int(image_t.value)

    def mk_world_point(self, image_x: int, image_y: int, world_t: float) -> tuple[float, float]:
        scene_l = ctypes.c_double()
        scene_f = ctypes.c_double()
        success: bool = _Api.mkScenePoint(self._handle, image_x, image_y, world_t, scene_l, scene_f)
        if not success:
            eventlog.log_i('fail: mkScenePoint')
            raise ParamsError

        return scene_f.value, scene_l.value

    def _assert_cam_speed_vr(self,
                             vrp_text_utf8: bytes,
                             main_speed_kmh: int,
                             min_item_with_vr: LrTrackItem,
                             max_item_with_vr: LrTrackItem,
                             trackset_dir: Dir,
                             main_dir_u: bool = False) -> int:
        """
        :param vrp_text_utf8:
        :param main_speed_kmh:
        :param min_item_with_vr:
        :param max_item_with_vr:
        :param main_dir_u:
            The sign main_speed_kmh does not indicate the direction of movement
            of the vehicle relative to the camera.

        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        c_control_speed_kmh = ctypes.c_double()

        c_plate_text = None
        if self._speedcam2_conf.with_log:
            c_plate_text = vrp_text_utf8

        cpoint1 = CPoint(int((min_item_with_vr.vehicle_rect.il + min_item_with_vr.vehicle_rect.ir) / 2),
                         min_item_with_vr.vehicle_rect.ib)

        cpoint2 = CPoint(int((max_item_with_vr.vehicle_rect.il + max_item_with_vr.vehicle_rect.ir) / 2),
                         max_item_with_vr.vehicle_rect.ib)

        c_plate_up = ctypes.c_double(0)
        c_answer = _Api.assertPointSpeed(
            self._handle, c_plate_text, min_item_with_vr.ts, max_item_with_vr.ts,
            ctypes.byref(cpoint1), ctypes.byref(cpoint2), c_plate_up,
            main_dir_u, main_speed_kmh, ctypes.byref(c_control_speed_kmh))

        algo_stats = self._speedcam2_stats.point_a if trackset_dir == Dir.CODE_A else self._speedcam2_stats.point_r
        if c_answer == kLLM_CAM_STAT__FAIL_ESTIMATE:
            eventlog.log_i(f'fail: speedcam (plate:{vrp_text_utf8}, reason:fail-estimate)')
            algo_stats.inc_failure_estimate()
            raise SpeedEstimateError()

        elif c_answer == kLLM_CAM_STAT__FAIL_VALIDATE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:fail-validate, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            algo_stats.inc_failure_validate()
            raise SpeedValidateError()

        control_speed_kmh_f: float = c_control_speed_kmh.value
        control_speed_kmh_i: int = map_speed_int(control_speed_kmh_f)
        algo_stats.inc_success(main_speed_kmh, control_speed_kmh_i, main_dir_u)

        return control_speed_kmh_i

    def _assert_cam_speed_vrp(self,
                              vrp_text_utf8: bytes,
                              main_speed_kmh: int,
                              min_lr_item: LrTrackItem,
                              max_lr_item: LrTrackItem,
                              trackset_dir: Dir,
                              main_dir_u: bool = False) -> int:
        """
        :param vrp_text_utf8:
        :param main_speed_kmh:
        :param min_lr_item:
        :param max_lr_item:
        :param main_dir_u:
            The sign main_speed_kmh does not indicate the direction of movement
            of the vehicle relative to the camera.

        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        if min_lr_item.ts == max_lr_item.ts:
            eventlog.log_i('fail')
            raise ParamsError

        c_double_p = ctypes.POINTER(ctypes.c_double)
        c_plate_up_p: c_double_p | None = None
        if self._speedcam2_conf.avg_plate_u is not None:
            c_plate_up = ctypes.c_double(self._speedcam2_conf.avg_plate_u)
            c_plate_up_p = c_double_p(c_plate_up)

        c_control_speed_kmh = ctypes.c_double()

        c_plate_text = None
        if self._speedcam2_conf.with_log:
            c_plate_text = vrp_text_utf8

        quad1 = min_lr_item.plate_quad
        quad2 = max_lr_item.plate_quad

        cquad1 = CQuad(quad1.x0, quad1.y0, quad1.x1, quad1.y1, quad1.x2, quad1.y2, quad1.x3, quad1.y3)
        cquad2 = CQuad(quad2.x0, quad2.y0, quad2.x1, quad2.y1, quad2.x2, quad2.y2, quad2.x3, quad2.y3)

        c_answer = _Api.assertPlateSpeed(
            self._handle, c_plate_text, min_lr_item.ts, max_lr_item.ts,
            ctypes.byref(cquad1), ctypes.byref(cquad2), c_plate_up_p,
            main_dir_u,
            main_speed_kmh, ctypes.byref(c_control_speed_kmh))

        algo_stats = self._speedcam2_stats.plate_a if trackset_dir == Dir.CODE_A else self._speedcam2_stats.plate_r
        if c_answer == kLLM_CAM_STAT__FAIL_ESTIMATE:
            eventlog.log_i(f'fail: speedcam (plate:{vrp_text_utf8}, reason:fail-estimate)')
            algo_stats.inc_failure_estimate()
            raise SpeedEstimateError()

        elif c_answer == kLLM_CAM_STAT__FAIL_VALIDATE:
            eventlog.log_i(
                'fail: speedcam (plate:{}, reason:fail-validate, main-speed:{:3f})'.format(
                    vrp_text_utf8, main_speed_kmh))
            algo_stats.inc_failure_validate()
            raise SpeedValidateError()

        control_speed_kmh_f: float = c_control_speed_kmh.value
        control_speed_kmh_i: int = map_speed_int(control_speed_kmh_f)
        algo_stats.inc_success(main_speed_kmh, control_speed_kmh_i, main_dir_u)

        return control_speed_kmh_i

    def assert_cam_speed_2(self,
                           vrp_text_utf8: bytes,
                           main_speed_kmh: int,
                           lr_items: list[LrTrackItem],
                           trackset_dir: Dir,
                           main_dir_u: bool = False) -> int:
        """
        :param vrp_text_utf8:
        :param main_speed_kmh:
        :param lr_items:
        :param trackset_dir:
        :param main_dir_u:
            The sign main_speed_kmh does not indicate the direction of movement
            of the vehicle relative to the camera.

        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        min_item_with_vr = None
        for item in lr_items:
            if item.vehicle_rect is not None:
                min_item_with_vr = item
                break

        max_item_with_vr = None
        for item in reversed(lr_items):
            if item.vehicle_rect is not None:
                max_item_with_vr = item
                break

        prefer_vr = False
        if None not in [min_item_with_vr, max_item_with_vr]:
            if min_item_with_vr.ts != max_item_with_vr.ts:
                prefer_vr = True

        if prefer_vr:
            # vehicle rect
            return self._assert_cam_speed_vr(
                vrp_text_utf8, main_speed_kmh,
                min_item_with_vr, max_item_with_vr, trackset_dir, main_dir_u)
        else:
            # vehicle registration plate
            return self._assert_cam_speed_vrp(
                vrp_text_utf8, main_speed_kmh,
                lr_items[0], lr_items[-1], trackset_dir, main_dir_u)

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler({
            'type': 'tracker~speedcam2',
            'name': 'tracker.0~speedcam2.0',
            'stat': self._speedcam2_stats.encode()
        })


def init_service(conf: CamServiceConf) -> ICamService:
    return _CamService(conf)
