import os
import pwd
import grp
import datetime
import ujson

import psycopg2
import psycopg2.pool
import psycopg2.extras

from typing import Dict, List, Optional, Self
from psycopg2.extensions import connection as db_link, cursor

from edge import eventlog
from edge.services.types.service import IService
from edge.support import db

DEF_LIST_OWNER = 'local'
DEF_LIST_NAME = 'local'

DEF_PROF_NAME = 'default'
DEF_PROF_OLD = 'none'
DEF_PROF_NEW = 'set'
DEF_PROF_SAME = 'set'

DEF_PROF_TMPL = {
    'rowsep': '[\n\r]+',
    'colsep': ',',
    'fields': {
        'plate': {
            'format': '{0}'
        }
    }
}


class ProfileInfo:
    def __init__(self, name: str):
        self.name = name


class Profile:
    def __init__(self, name: str, desc: str, tmpl: str):
        self.name = name
        self.desc = desc
        self.tmpl = tmpl


class ListInfo:
    def __init__(self, owner: str, name: str, ver_dt: datetime.datetime, ver_desc: str):
        self.owner = owner
        self.ident = name
        self.utime = ver_dt
        self.comment = ver_desc

    def serialize(self) -> Dict:
        jnode = {
            'owner': self.owner,
            'name': self.ident,
            'time': self.utime,
        }

        if len(self.comment):
            jnode['comment'] = self.comment

        return jnode


class ListVersion:
    def __init__(self, ver_dt: datetime.datetime, ver_rev: str, ver_desc: str):
        self.utime = ver_dt
        self.revision = ver_rev
        self.comment = ver_desc


class ListMatch:
    def __init__(self, plate_list: ListInfo, plate_text: str, plate_dist: int, plate_extra: Dict):
        self.list = plate_list
        self.plate = plate_text
        self.dist = plate_dist
        self.extra = plate_extra

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        owner = jnode['list-owner']
        name = jnode['list-ident']
        ver_dt = jnode['list-date']
        ver_desc = jnode.get('list-description', '')
        list_info = ListInfo(owner, name, ver_dt, ver_desc)

        plate_text = jnode['plate']
        plate_dist = jnode['plate-distance']

        keys_to_delete = ['list-owner', 'list-ident', 'list-date', 'list-description', 'plate', 'plate-distance']

        for key in keys_to_delete:
            if jnode.get(key) is not None:
                del jnode[key]

        return cls(list_info, plate_text, plate_dist, jnode)

    def serialize(self) -> Dict:
        jnode = {
            'plate-list': self.list.serialize(),
            'plate-text': self.plate,
            'plate-dist': self.dist,
            'plate-extra': self.extra
        }

        return jnode


class ListHistoryItem:
    def __init__(self, moment, revision: str, sha1: str, template: str,
                 description: Optional[str], operator: Optional[str],
                 inserted: int, updated: int, deleted: int):

        self.moment = moment
        self.revision = revision
        self.sha1 = sha1
        self.template = template
        self.description = description
        self.operator = operator
        self.inserted = inserted
        self.updated = updated
        self.deleted = deleted

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        moment = jnode['moment']
        revision = jnode['revision']
        sha1 = jnode['sha1']

        template = jnode['template']
        description = jnode.get('description')
        operator = jnode.get('operator')

        inserted = jnode['inserted']
        updated = jnode['updated']
        deleted = jnode['deleted']

        return cls(moment, revision, sha1, template, description, operator, inserted, updated, deleted)


class PlateHistoryItem:
    def __init__(self, moment, revision: str, op: str, sha1: str, operator: Optional[str], description: Optional[str]):
        self.moment = moment
        self.revision = revision
        self.sha1 = sha1
        self.description = description
        self.operator = operator
        self.op = op

    @classmethod
    def unpack(cls, jnode: Dict) -> Self:
        moment = jnode['moment']
        revision = jnode['revision']
        sha1 = jnode['sha1']

        op = jnode['op']
        description = jnode.get('description')
        operator = jnode.get('operator')

        return cls(moment, revision, op, sha1, operator, description)


def link_guard(fn):
    def proxy_fn(self, *args, **kwargs):
        if not self._link:
            eventlog.log_e('fail: db-link (use context manager)')
            raise db.DbError

        return fn(self, *args, **kwargs)
    return proxy_fn


class DbWantedService(IService):
    """
    :type _link: db_link | None
    """
    def __init__(self, link_pool: psycopg2.pool.SimpleConnectionPool):
        self._link_pool = link_pool
        self._link = None

    def __enter__(self, *args, **kwargs):
        try:
            self._link = self._link_pool.getconn()
            self._link.__enter__(*args, **kwargs)
            return self

        except psycopg2.pool.PoolError as x:
            eventlog.log_e('fail: db-link ({})'.format(str(x)))
            raise db.DbError

    def __exit__(self, *args, **kwargs):
        self._link.__exit__(*args, **kwargs)
        self._link_pool.putconn(self._link)
        self._link = None

    def destroy(self) -> None:
        self._link_pool.closeall()

    @link_guard
    def load_profile_infos(self) -> List[ProfileInfo]:
        with self._link.cursor() as db_cursor:
            db_cursor.execute('SELECT ident FROM hwp.watchlistpl;')
            names = db_cursor.fetchall()
            return list(map(ProfileInfo, names))

    @link_guard
    def load_profile(self, name: str) -> Optional[Profile]:
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute("SELECT * FROM hwp.watchlistpl('list',%s,'',%s);", [name, None])
                answer_row = db_cursor.fetchone()
                answer_jnode: List[Dict] = answer_row[0]
                if answer_jnode is None:
                    return None
                elif len(answer_jnode) == 0:
                    return None
                elif len(answer_jnode) > 1:
                    raise NotImplementedError

                p_jnode = answer_jnode[0]

                desc = p_jnode.get('description', '')
                tmpl = p_jnode['template']
                return Profile(name, desc, tmpl)

        except psycopg2.Error as x:
            eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
            raise NotImplementedError

    @link_guard
    def save_profile(self, name: str, desc: str, tmpl: Dict) -> None:
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    "SELECT * FROM hwp.watchlistpl('set',%s,%s,%s);",
                    [name, desc, db.DbJsonArg(tmpl)])

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
            raise db.DbError

    @link_guard
    def drop_profile(self, name: str) -> None:
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute("SELECT * FROM hwp.watchlistpl('delete',%s);", [name])
        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
            raise db.DbError

    @link_guard
    def load_list_infos(self) -> List[ListInfo]:
        with self._link.cursor() as db_cursor:
            db_cursor.execute('SELECT owner,ident,utime,description FROM hwp.watchlist;')
            answer_rows = db_cursor.fetchall()

            list_infos = []
            for owner, name, ver_dt, ver_desc in answer_rows:
                list_infos.append(ListInfo(owner, name, ver_dt, ver_desc))

            return list_infos

    @link_guard
    def load_list_versions(self, owner: str, name: str) -> List[ListVersion]:
        with self._link.cursor() as db_cursor:
            db_cursor.execute(
                'SELECT moment,revision,description FROM hwp.watchlistlog '
                'WHERE owner=%s and ident=%s ORDER BY moment DESC;',
                [owner, name])
            answer_rows = db_cursor.fetchall()

        list_versions = []
        for ver_dt, ver_rev, ver_desc in answer_rows:
            list_versions.append(ListVersion(ver_dt, ver_rev, ver_desc))

        return list_versions

    @link_guard
    def drop_list(self, owner: str, name: str) -> int:
        """
        :returns:
            Affected row count.
        """
        with self._link.cursor() as db_cursor:  # type: cursor
            db_cursor.execute(
                'SELECT affected FROM hwp.watchlist_delete(%s,%s);',
                [owner, name])

            answer_row = db_cursor.fetchone()
            return answer_row[0]

    @link_guard
    def save_list(self, list_owner: str, list_name: str, list_file: str,
                  list_sha1: Optional[str], list_desc: Optional[str],
                  prof_tmpl: Optional[str], prof_name: Optional[str],
                  prof_old: str, prof_new: str, prof_same: str,
                  operator: str = 'dbuser'):
        """
        :param list_owner: str
        :param list_name: str

        :param list_file:
            list-file-path or list-file-data
        :param list_sha1:
        :param list_desc:

        :param prof_tmpl:
            Json. Has higher priority then prof_name.
        :param prof_name:
        :param prof_old:
        :param prof_new:
        :param prof_same:
        :param operator:
        """
        if os.path.isfile(list_file):
            uid = pwd.getpwnam('postgres').pw_uid
            gid = grp.getgrnam('postgres').gr_gid
            os.chown(list_file, uid, gid)

        if prof_tmpl:
            prof_tmpl = ujson.dumps(prof_tmpl)

        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT total,inserted,deleted,updated,errors '
                    'FROM hwp.watchload(%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)',
                    [list_owner, list_name, prof_tmpl, prof_name, list_file, list_sha1,
                     list_desc, operator, prof_old, prof_same, prof_new])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
                raise db.DbError

            total, inserted, deleted, updated, errors = db_cursor.fetchone()
            eventlog.log_i(
                'done: db (total:{}, inserted:{}, deleted:{}, updated:{}, errors:{})'
                .format(total, inserted, deleted, updated, len(errors)))

            if len(errors) != 0:
                for error in errors:
                    line, raw_data, error = error
                    data = raw_data.encode('utf-8')
                    eventlog.log_e(f'line:{line}, data:<{data}>, error:<{error}>')
                raise db.DbError

    @link_guard
    def cleanup_list(self, owner: str, name: str,
                     ver_max_rev: Optional[str] = None,
                     ver_max_hash: Optional[str] = None,
                     ver_max_time: Optional[str] = None):
        """
        :type owner: str
        :type name: str

        :type ver_max_rev: str | None
        :type ver_max_hash: str | None
        :type ver_max_time: str | None

        :return: Affected rows count.
        :rtype: int
        """
        try:
            eventlog.log_i(
                'exec: SELECT affected FROM hwp.watchlist_cleanup(%s,%s,%s,%s,%s);' %
                (owner, name, ver_max_time, ver_max_rev, ver_max_hash))

            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    'SELECT affected FROM hwp.watchlist_cleanup(%s,%s,%s,%s,%s);',
                    [owner, name, ver_max_time, ver_max_rev, ver_max_hash])

                answer_row = db_cursor.fetchone()
                return answer_row[0]

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
            raise db.DbError

    @link_guard
    def revert_list(self, owner: str, name: str,
                    ver_min_rev: Optional[str] = None,
                    ver_count: Optional[str] = None) -> None:
        try:
            with self._link.cursor() as db_cursor:  # type: cursor
                db_cursor.execute(
                    'SELECT inserted,updated,deleted FROM hwp.watchlist_revert(%s,%s,%s,%s);',
                    [owner, name, ver_count, ver_min_rev])

                inserted, updated, deleted = db_cursor.fetchone()
                eventlog.log_i(f'done: inserted:{inserted}, updated:{updated}, deleted:{deleted}')

        except psycopg2.Error as x:
            db.trap_hint_error(x)
            eventlog.log_e(f'fail: pgcode:{x.pgcode}, pgerror:{x.pgerror}')
            raise db.DbError

    @link_guard
    def check_plate(self, plate_text: str, max_dist: int = 0) -> List[ListMatch]:
        with self._link.cursor() as db_cursor:  # type: cursor
            db_cursor.execute(
                'SELECT watchlist FROM hwp.watchlist(%s,%s);',
                [plate_text, max_dist])

            answer_row = db_cursor.fetchone()
            # todo :check
            if len(answer_row):
                return list(map(ListMatch.unpack, answer_row[0]))
            else:
                return []

    @link_guard
    def get_list_history(self, list_owner: str, list_name: str,
                         limit: int = 1, offset: int = 0) -> List[ListHistoryItem]:

        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT listwatchlog '
                    'FROM hwp.listwatchlog(%s,%s,%s,%s)',
                    [list_owner, list_name, limit, offset])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
                raise db.DbError

            answer_row = db_cursor.fetchone()

        if len(answer_row):
            return list(map(ListHistoryItem.unpack, answer_row[0]))
        else:
            return []

    @link_guard
    def get_plate_history(self, list_owner: str, list_name: str, plate_text: str,
                          limit: int = 1, offset: int = 0) -> List[PlateHistoryItem]:

        with self._link.cursor() as db_cursor:
            try:
                db_cursor.execute(
                    'SELECT listwatchplatehistory '
                    'FROM hwp.listwatchplatehistory(%s,%s,%s,%s,%s)',
                    [list_owner, list_name, plate_text, limit, offset])

            except psycopg2.Error as x:
                db.trap_hint_error(x)
                eventlog.log_e(f'fail: db (pgcode:{x.pgcode}, pgerror:{x.pgerror})')
                raise db.DbError

            answer_row = db_cursor.fetchone()

        if len(answer_row):
            return list(map(PlateHistoryItem.unpack, answer_row[0]))
        else:
            return []


def init_service(conf: db.DbArgs) -> DbWantedService:
    """
    :raise db.DbError:
    """
    try:
        link_pool = psycopg2.pool.SimpleConnectionPool(
            minconn=1,
            maxconn=1,
            user=conf.db_user,
            password=conf.db_pass,
            database=conf.db_name,
            host=conf.db_host,
            port=conf.db_port)

    except psycopg2.Error as x:
        eventlog.log_e(f'fail: create db-link pool <{x}>')
        raise db.DbError

    return DbWantedService(link_pool)
