from typing import Final, Dict, List

from edge import eventlog
from edge.services.react import ReactSource
from edge.services.tl.trafficlight import ITrafficLight
from edge.services.tl.tl_signal import Signal
from edge.services.tl.trafficlight.jnode.react import decode__traffic_light
from edge.services.types.tl import ITlService, TlServiceConf, TlState, TlHaltSpanInfo, TlRoute
from edge.types.exceptions import JnodeError
from edge.types.timespan import TimeSpan


class _Service(ITlService):
    def __init__(self, twisted_reactor, conf: TlServiceConf):
        self._source = ReactSource(twisted_reactor)
        self._source.set_handler(self._handle_react)
        self._source.set_endpoints({conf.endpoint})
        self._conf: Final[TlServiceConf] = conf

        self._traffic_light: ITrafficLight | None = None

    def destroy(self) -> None:
        self._source.destroy()

    # noinspection PyUnusedLocal
    def _handle_react(self, react_name: str, react_jnode: Dict) -> None:
        try:
            self._traffic_light = decode__traffic_light(
                react_jnode, self._conf.tl_ident, self._conf.red_only, self._conf.accuracy_usec)

        except JnodeError as x:
            return

    def get_state(self, ts: int) -> TlState:
        if self._traffic_light is None:
            eventlog.log_d('fail: tl log is absent')
            return TlState()

        return self._traffic_light.get_state(ts)

    def get_halt_spans(self, min_ts: int, max_ts: int, route: TlRoute | None = None) -> List[TimeSpan]:
        return [] if self._traffic_light is None else self._traffic_light.get_signal_h(min_ts, max_ts)

    def get_halt_span_info(self, halt_span_ts: int, route: TlRoute | None = None) -> TlHaltSpanInfo:
        assert self._traffic_light is not None
        return self._traffic_light.get_halt_span_info(halt_span_ts)


class _StubService(ITlService):
    def get_state(self, ts: int) -> TlState:
        return TlState()

    def get_halt_spans(self, min_ts: int, max_ts: int, route: TlRoute | None = None) -> List[TimeSpan]:
        return Signal()

    def get_halt_span_info(self, ts: int, route: TlRoute | None = None) -> TlHaltSpanInfo:
        return TlHaltSpanInfo()
