from typing import Final, Optional

from edge.services.tl.tl_signal import Signal
from edge.services.tl.trafficlight import ITrafficLight
from edge.services.types.tl import TlState, TlHaltSpanInfo, TlRoute
from edge.types.timespan import TimeSpan


class TrafficLightR(ITrafficLight):
    def __init__(self, ref: int, time_span: TimeSpan, signal_h: Signal, signal_r: Signal) -> None:
        """
        :param ref:
            The reference of the traffic light.
        :param signal_h:
            The timeline of the halt signal.
        :param signal_r:
            The timeline of the red signal.
        """
        self.ref: Final[int] = ref
        self._time_span: Final[TimeSpan] = time_span

        self._signal_h: Final[Signal] = signal_h
        self._signal_r: Final[Signal] = signal_r

    def _get_red_msec(self, ts: int) -> Optional[int]:
        h_spans = [span for span in self._signal_h if ts in span]
        if not len(h_spans):
            return None

        h_span0: TimeSpan = h_spans[0]

        # find red intervals in prohibited item
        r_spans = [span for span in self._signal_r if span.test_overlaps(h_span0)]

        # red started before ts
        r_spans = [span for span in r_spans if span.min_ts < ts]
        if not len(r_spans):
            return None

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - r_spans[0].min_ts

        return any_red_time // 1000

    def get_state(self, ts: int) -> TlState:
        return TlState(r_active=self._signal_r.get_state(ts))

    def get_signal_h(self, min_ts: int, max_ts: int, route: TlRoute | None = None) -> Signal:
        time_span = TimeSpan(min_ts, max_ts)
        if self._time_span.test_overlaps(time_span):
            return Signal([x for x in self._signal_h if x.test_overlaps(time_span)])

        return Signal()

    def get_halt_span_info(self, ts: int, route: TlRoute | None = None) -> TlHaltSpanInfo:
        red_time = self._get_red_msec(ts)
        return TlHaltSpanInfo(halt_time=red_time, red_time=red_time)
