from typing import Final, Optional

from edge.services.tl.tl_signal import Signal
from edge.services.tl.trafficlight import ITrafficLight
from edge.services.types.tl import TlState, TlHaltSpanInfo, TlRoute
from edge.types.timespan import TimeSpan


class TrafficLightRag(ITrafficLight):
    def __init__(self, ref: int,
                 time_span: TimeSpan,
                 signal_h: Signal,
                 signal_r: Signal,
                 signal_a: Signal,
                 signal_g: Signal) -> None:
        """
        :param ref:
            The reference of the traffic light.
        :param signal_h:
            The timeline of the halt signal.
        :param signal_r:
            The timeline of the red signal.
        :param signal_a:
            The timeline of the amber signal.
        :param signal_g:
            The timeline of the green signal.
        """

        self.ref: Final[int] = ref
        self._time_span: Final[TimeSpan] = time_span

        self._signal_h: Final[Signal] = signal_h
        self._signal_r: Final[Signal] = signal_r
        self._signal_a: Final[Signal] = signal_a
        self._signal_g: Final[Signal] = signal_g

    def _get_green_amber_msec(self, ts: int) -> Optional[int]:
        try:
            oldest_h_span: TimeSpan = next(
                h_span for h_span in self._signal_h
                if ts in h_span)

            oldest_a_span: TimeSpan = next(
                a_span for a_span in self._signal_a
                if a_span.test_overlaps(oldest_h_span) and a_span.min_ts < ts)

        except StopIteration:
            return None

        a_min_ts: int = oldest_a_span.min_ts
        r_min_ts: int | None = self._signal_r.get_min_ts(min_ts=oldest_h_span.min_ts, max_ts=oldest_h_span.max_ts)

        if r_min_ts is None or ts < r_min_ts:
            return (ts - a_min_ts) // 1000

        else:
            return (r_min_ts - a_min_ts) // 1000

    def _get_red_msec(self, ts: int) -> Optional[int]:
        try:
            oldest_h_span: TimeSpan = next(
                h_span for h_span in self._signal_h
                if ts in h_span)

            oldest_r_span: TimeSpan = next(
                r_span for r_span in self._signal_r
                if r_span.test_overlaps(oldest_h_span) and r_span.min_ts < ts)

        except StopIteration:
            return None

        # check "just red" instead of "any red"
        # just_red = any_red - any_amber
        any_red_time = ts - oldest_r_span.min_ts

        post_amber_time = self._get_red_amber_msec(ts) or 0
        return (any_red_time - post_amber_time) // 1000

    def _get_red_amber_msec(self, ts: int) -> Optional[int]:
        try:
            oldest_h_span: TimeSpan = next(
                h_span for h_span in self._signal_h
                if ts in h_span)

            newest_a_span = next(
                a_span for a_span in self._signal_a[::-1]
                if a_span.test_overlaps(oldest_h_span) and a_span.min_ts < ts)

        except StopIteration:
            return None

        r_min_ts: int | None = self._signal_r.get_min_ts(min_ts=oldest_h_span.min_ts, max_ts=oldest_h_span.max_ts)
        if r_min_ts is not None and r_min_ts < newest_a_span.min_ts:
            return (ts - newest_a_span.min_ts) // 1000

        else:
            return None  # ?

    def get_state(self, ts: int) -> TlState:
        return TlState(
            r_active=self._signal_r.get_state(ts),
            a_active=self._signal_a.get_state(ts),
            g_active=self._signal_g.get_state(ts))

    def get_signal_h(self, min_ts: int, max_ts: int, route: TlRoute | None = None) -> Signal:
        time_span = TimeSpan(min_ts, max_ts)
        if self._time_span.test_overlaps(time_span):
            return Signal([x for x in self._signal_h if x.test_overlaps(time_span)])

        return Signal()

    def get_halt_span_info(self, ts: int, route: TlRoute | None = None) -> TlHaltSpanInfo:
        """
        :returns:
            (green-amber-msec, red-msec, red-amber-msec, halt-msec)
        """
        red_msec = self._get_red_msec(ts)
        red_amber_msec = self._get_red_amber_msec(ts)
        green_amber_msec = self._get_green_amber_msec(ts)

        halt_msec = 0
        if green_amber_msec is not None:
            halt_msec += green_amber_msec

        if red_msec is not None:
            halt_msec += red_msec

        if red_amber_msec is not None:
            halt_msec += red_amber_msec

        return TlHaltSpanInfo(halt_time=halt_msec,
                              green_amber_time=green_amber_msec,
                              red_amber_time=red_amber_msec,
                              red_time=red_msec)
