import os
import subprocess
from glob import glob
from time import sleep

import ujson
from edge.support.debinterface.interfaces import Interfaces

from edge import eventlog
from edge.support import sh, fs
from edge.types.exceptions import InternalError

PATH_CONF = '/etc/network/interfaces'
PATH_CONFD = '/etc/network/interfaces.d'
PATH_CONFD_PRIMARY = '/etc/network/interfaces.d/primary.conf'
PATH_MANAGEMENT_MAC = '/etc/network/management-mac'
PATH_DEFAULT_NETWORKING = '/etc/default/networking'

PATH_ETC_HOSTNAME = '/etc/hostname'
PATH_ETC_HOSTS = '/etc/hosts'


def list_ifnames() -> list[str]:
    dir_matches = glob('/sys/class/net/*')
    raw_ifnames = [os.path.basename(x) for x in dir_matches]
    return [x for x in raw_ifnames if x != 'lo']


def map_ifmac_to_ifname(if_mac: str) -> str | None:
    # /sys/class/net/eth0/address
    # /sys/class/net/eth1/address
    # /sys/class/net/eth2/address
    # /sys/class/net/lo/address
    mac_files = glob('/sys/class/net/*/address')
    for mac_file in mac_files:
        next_if_name = mac_file.split('/')[-2]  # like eth0
        with open(mac_file, 'r') as text_file:
            next_if_mac = text_file.readlines()[0].strip()

        if next_if_mac == if_mac:
            return next_if_name


def map_ifname_to_ifmac(if_name: str) -> str | None:
    mac_file = f'/sys/class/net/{if_name}/address'
    for mac_file in glob(mac_file):
        with open(mac_file, 'r') as text_file:
            if_mac = text_file.readlines()[0].strip()
            return if_mac


def _poll_iface_up(max_attempts: int) -> str | None:
    cli_params = ['ip', '-o', '-j', 'link', 'show']

    for attempt in range(1, 1 + max_attempts):
        eventlog.log_i(f'exec: attempt:{attempt}/{max_attempts}')
        try:
            cli_result = subprocess.run(
                cli_params, capture_output=True, text=True, encoding='ascii',
                check=True, timeout=1.0)

            if_infos: list[dict] = ujson.loads(cli_result.stdout)
            for iface in if_infos:
                if iface['ifname'] == 'lo':
                    pass

                elif iface['operstate'] == 'UP':
                    eventlog.log_i(f'done: attempt:{attempt}/{max_attempts}, iface:{iface["ifname"]}, operstate:UP')
                    return iface['ifname']

                elif 'LOWER_UP' in iface['flags']:
                    eventlog.log_i(f'done: attempt:{attempt}/{max_attempts}, iface:{iface["ifname"]}, flags:LOWER_UP')
                    return iface['ifname']

        except subprocess.TimeoutExpired as x:
            eventlog.log_w(f'fail: <{x.cmd}> (timeout:{x.timeout})')

        except subprocess.CalledProcessError as x:
            eventlog.log_w(f'fail: <{x.cmd}> (returncode:{x.returncode})')

        sleep(1)


def infer_primary_ifname(ifnames: list[str]) -> str | None:
    eventlog.log_d(f'exec: ifnames:{ifnames}')

    # Test firstboot.
    if not os.path.exists(PATH_CONFD):
        eventlog.log_d('interfaces.d directory does not exist')

    elif os.listdir(PATH_CONFD):
        eventlog.log_e('fail: interfaces.d directory is not empty')
        raise InternalError

    # Read the configuration from the file.
    ifconf: Interfaces = Interfaces(interfaces_path=PATH_CONF)

    # Remove all except the 'loopback' interface.
    ifnames2 = [x.get_attr('name') for x in ifconf.adapters if x.get_attr('name') != 'lo']
    for ifname in ifnames2:
        ifconf.removeAdapterByName(ifname)

    # Set static configuration for all interfaces except the 'loopback' interface.
    for index in range(len(ifnames)):
        ifname = ifnames[index]
        ifconf.addAdapter({
            'name': ifname,
            'auto': True,
            'addrFam': 'inet',
            'source': 'static',
            'address': f'192.168.{123 + index}.223',
            'netmask': f'255.255.255.0'
        })

    # Write the configuration to the file.
    ifconf.writeInterfaces()

    eventlog.log_d(f'exec: <systemctl restart networking>')

    # Restart the networking service.
    sh.run_silent('systemctl restart networking')

    # Wait until the primary interface is up.
    primary_ifname: str | None = _poll_iface_up(60)

    # Remove all except the 'loopback' interface.
    ifnames2 = [x.get_attr('name') for x in ifconf.adapters if x.get_attr('name') != 'lo']
    for ifname in ifnames2:
        ifconf.removeAdapterByName(ifname)

    # Write the configuration to the file.
    ifconf.writeInterfaces()

    # Append 'source' directive to the end of the PATH_CONF file
    # This is a workaround for the debinterface v3.5.0.
    with open(PATH_CONF, 'a') as f:
        f.write(f'source {PATH_CONFD}/*.conf\n')

    assert primary_ifname is None or primary_ifname in ifnames
    return primary_ifname


def init_primary_mac(ifname: str) -> str:
    ifmac = map_ifname_to_ifmac(ifname)
    fs.write_text_file([ifmac], PATH_MANAGEMENT_MAC)
    return ifmac


def get_primary_mac() -> str:
    with open(PATH_MANAGEMENT_MAC, 'r') as text_file:
        return text_file.readline().strip()


def init_primary_ifconf_dhcp(ifname: str) -> None:
    ifmac = map_ifname_to_ifmac(ifname)
    byte4 = int(ifmac.split(':')[-1], 16)

    # Clear the configuration file.
    ifconf_path = PATH_CONFD_PRIMARY
    with open(ifconf_path, 'w'):
        pass

    ifconf: Interfaces = Interfaces(interfaces_path=ifconf_path, update_adapters=False)
    ifconf.addAdapter({
        'name': ifname,
        'auto': True,
        'addrFam': 'inet',
        'source': 'dhcp'
    }, interfaces_path=ifconf_path)

    ifconf.addAdapter({
        'name': f'{ifname}:1',
        'auto': True,
        'addrFam': 'inet',
        'source': 'static',
        'address': f'192.168.222.{byte4}',
        'netmask': f'255.255.255.0'
    }, interfaces_path=ifconf_path)

    ifconf.writeInterfaces()


def init_camera_ifconf_static(ifname: str, ifindex: int, mtu=6000) -> None:
    # Clear the configuration file.
    ifconf_path = os.path.join(PATH_CONFD, f'{ifname}.conf')
    with open(ifconf_path, 'w'):
        pass

    ifconf: Interfaces = Interfaces(interfaces_path=ifconf_path, update_adapters=False)
    ifconf.addAdapter({
        'name': ifname,
        'auto': True,
        'addrFam': 'inet',
        'source': 'static',
        'address': f'192.168.{ifindex + 2}4.1',
        'netmask': f'255.255.255.0',
        'mtu': f'{mtu}'
    }, interfaces_path=ifconf_path)

    ifconf.writeInterfaces()


def set_hostname(new_hostname: str) -> bool:
    with open(PATH_ETC_HOSTNAME, 'r') as hostname_file:
        old_hostname = hostname_file.readline()

    if old_hostname == new_hostname:
        return False

    with open(PATH_ETC_HOSTS, 'r') as old_hosts_file:
        old_hosts_lines = old_hosts_file.readlines()

    line_found = False
    new_host_lines = []
    for old_line in old_hosts_lines:
        stripped_old_line = old_line.strip()
        if not stripped_old_line or stripped_old_line.startswith('#'):
            new_host_lines.append(stripped_old_line)
            continue

        param_set = stripped_old_line.split()
        if len(param_set) < 2:
            new_host_lines.append(stripped_old_line)
            continue

        hostname = param_set[1]
        if hostname == old_hostname:
            param_set[1] = new_hostname
            new_host_lines.append('\t'.join(param_set))
            line_found = True

        else:
            new_host_lines.append(stripped_old_line)

    if not line_found:
        new_host_lines.append('\t'.join(['127.0.0.1', new_hostname]))

    with open(PATH_ETC_HOSTS, 'w') as new_hosts_file:
        new_hosts_file.write('\n'.join(new_host_lines))

    with open(PATH_ETC_HOSTNAME, 'w') as hostname_file:
        hostname_file.write(new_hostname)

    return True


def edit_default_networking() -> None:
    with open(PATH_DEFAULT_NETWORKING, 'r') as text_file:
        lines = text_file.readlines()

    new_lines = []
    for line in lines:
        if line.startswith('#CONFIGURE_INTERFACES='):
            new_lines.append('CONFIGURE_INTERFACES=no\n')

        # elif line.startswith('#WAIT_ONLINE_METHOD='):
        #     new_lines.append('WAIT_ONLINE_METHOD=none\n')

        else:
            new_lines.append(line)

    with open(PATH_DEFAULT_NETWORKING, 'w') as text_file:
        text_file.write(''.join(new_lines))

