from typing import Callable, Type, TypeVar

from edge import eventlog
from edge.services.types.service import IService
from edge.types.exceptions import GenericError, InternalError


TService = TypeVar('TService', bound=IService)


class ServicePack:
    def __init__(self):
        self._services: dict[str, IService] = dict()

    def add(self, service_type: Type[TService], service: TService | None) -> None:
        if service_type.__name__ in self._services:
            eventlog.log_e(f'fail: service with name <{service_type.__name__}> already exists')
            raise InternalError()

        elif service is None:
            eventlog.log_d(f'fail: service with name <{service_type.__name__}> is None')

        else:
            self._services[service_type.__name__] = service
            eventlog.log_d(f'done: name:{service_type.__name__}')

    def get(self, service_type: Type[TService]) -> TService:
        if service_type.__name__ not in self._services:
            eventlog.log_e(f'fail: service with name <{service_type.__name__}> not found')
            raise InternalError()

        else:
            return self._services[service_type.__name__]

    def try_get(self, service_type: Type[TService]) -> TService | None:
        if service_type.__name__ not in self._services:
            eventlog.log_e(f'fail: service with name <{service_type.__name__}> not found')
            return None

        else:
            return self._services[service_type.__name__]

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        for service in self._services.values():
            # noinspection PyBroadException
            try:
                service.collect_stats(json_handler)
            except GenericError:
                pass
            except Exception:
                eventlog.log_x()

    def destroy(self) -> None:
        for name, service in self._services.items():
            # noinspection PyBroadException
            try:
                service.destroy()
            except Exception:
                eventlog.log_e(f'fail: destroy service <{name}>')
                eventlog.log_x()

        eventlog.log_d('done')
