import ctypes

from typing import Any, Callable, Dict, List, Tuple, TypeVar

from PIL import Image

from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectF, RectI

TImRgbaHandlerResult = TypeVar('TImRgbaHandlerResult')


class ImRgba:
    def __init__(self):
        self._w = 0
        self._h = 0
        self._dh = 0
        self._ts = 0

        self._buffer: ctypes.Array[ctypes.c_char] = ctypes.create_string_buffer(0)

    @property
    def ts(self) -> int:
        return self._ts

    @property
    def w(self) -> int:
        return self._w

    @property
    def h(self) -> int:
        return self._h

    @property
    def wh(self) -> Tuple[int, int]:
        return self._w, self._h

    @property
    def dh(self) -> int:
        return self._dh

    @property
    def buffer(self) -> ctypes.Array[ctypes.c_char]:
        return self._buffer

    @property
    def stride(self) -> int:
        return 4 * self._w

    def reset(self, w: int, h: int, dh: int = 0, ts: int = 0) -> None:
        self._w = w
        self._h = h
        self._dh = dh
        self._ts = ts

        if (h + dh) * self.stride > len(self._buffer):
            self._buffer = ctypes.create_string_buffer(self.stride * (h + dh))

    def with_image(self, handler: Callable[[Image.Image], TImRgbaHandlerResult]) -> TImRgbaHandlerResult:
        with Image.frombuffer(
                'RGBA', (self.w, self.h + self.dh), self.buffer,
                'raw', 'RGBA', self.stride, 1) as im:

            im.readonly = 0
            return handler(im)

    def with_crop_image(self,
                        crop_rect: RectI,
                        handler: Callable[[Image.Image], TImRgbaHandlerResult]) -> TImRgbaHandlerResult:

        buffer_mv = memoryview(self.buffer)
        crop_offset = crop_rect.it * self.stride + 4 * crop_rect.il
        with Image.frombuffer(
                'RGBA', crop_rect.wh, buffer_mv[crop_offset:],
                'raw', 'RGBA', self.stride, 1) as im:

            im.readonly = 0
            return handler(im)

    def with_header_image(self, handler: Callable[[Image.Image], TImRgbaHandlerResult]) -> TImRgbaHandlerResult:
        with Image.frombuffer(
                'RGBA', (self.w, self.h), self.buffer,
                'raw', 'RGBA', self.stride, 1) as im:

            im.readonly = 0
            return handler(im)

    def with_footer_image(self, handler: Callable[[Image.Image], TImRgbaHandlerResult]) -> TImRgbaHandlerResult:
        buffer_mv = memoryview(self.buffer)
        footer_offset = int(self.stride * self.h)
        with Image.frombuffer(
                'RGBA', (self.w, self.dh), buffer_mv[footer_offset:],
                'raw', 'RGBA', self.stride, 1) as im:

            im.readonly = 0
            return handler(im)


class ImJpeg:
    def __init__(self):
        self._ts = 0
        self._w = 0
        self._h = 0

        self._buffer: ctypes.Array = ctypes.create_string_buffer(0)
        self._buffer_used = 0

    @property
    def ts(self) -> int:
        return self._ts

    @property
    def w(self) -> int:
        return self._w

    @property
    def h(self) -> int:
        return self._h

    @property
    def wh(self) -> Tuple[int, int]:
        return self._w, self._h

    @property
    def buffer_size(self) -> int:
        return len(self._buffer)

    @property
    def buffer_used(self) -> int:
        return self._buffer_used

    @property
    def buffer(self) -> ctypes.Array:
        return self._buffer

    @property
    def buffer_data(self) -> List[Any]:
        return self._buffer[0:self._buffer_used]

    def reset(self, w: int, h: int, buffer_used: int, ts: int = 14) -> None:
        self._ts = ts
        self._w = w
        self._h = h

        if buffer_used > self.buffer_size:
            self._buffer = ctypes.create_string_buffer(buffer_used)

        self._buffer_used = buffer_used


class ImTrans:

    #
    # (x', y') - scene-im point; (x'', y'') - recrop-image point.
    #
    # |x''|   | a  b  x |   |x'|
    # |y''| = | c  d  y | * |y'|
    # |1  |   | 0  0  1 |   |1 |
    #
    #
    # |x''|   | Ix Jx Ox |   |x'|
    # |y''| = | Iy Jy Oy | * |y'|
    # |1  |   | 0  0  1  |   |1 |
    #

    def __init__(self,
                 ix: float = 1.0,
                 iy: float = 0.0,
                 jx: float = 0.0,
                 jy: float = 1.0,
                 ox: float = 0.0,
                 oy: float = 0.0):

        self._ix = ix
        self._iy = iy
        self._jx = jx
        self._jy = jy
        self._ox = ox
        self._oy = oy

    def map_ltrb(self, source_ltrb: List[float]) -> List[float]:
        l, t, r, b = source_ltrb
        target_ltrb = [
            l * self._ix + t * self._jx + self._ox,
            l * self._iy + t * self._jy + self._oy,

            r * self._ix + b * self._jx + self._ox,
            r * self._iy + b * self._jy + self._oy
        ]

        return target_ltrb

    def map_point(self, fl: float, ft: float) -> Tuple[float, float]:
        return (
            fl * self._ix + ft * self._jx + self._ox,
            fl * self._iy + ft * self._jy + self._oy
        )

    def map_track(self, source_track: List[Tuple[float, float]]) -> List[Tuple[float, float]]:
        target_track = []
        for fl, ft in source_track:
            target_track.append(self.map_point(fl, ft))

        return target_track

    def map_quad(self, source_quad: QuadI) -> QuadI:
        target_quad = QuadI(
            int(source_quad.x0 * self._ix + source_quad.y0 * self._jx + self._ox),
            int(source_quad.x0 * self._iy + source_quad.y0 * self._jy + self._oy),

            int(source_quad.x1 * self._ix + source_quad.y1 * self._jx + self._ox),
            int(source_quad.x1 * self._iy + source_quad.y1 * self._jy + self._oy),

            int(source_quad.x2 * self._ix + source_quad.y2 * self._jx + self._ox),
            int(source_quad.x2 * self._iy + source_quad.y2 * self._jy + self._oy),

            int(source_quad.x3 * self._ix + source_quad.y3 * self._jx + self._ox),
            int(source_quad.x3 * self._iy + source_quad.y3 * self._jy + self._oy))

        return target_quad

    def map_rectf(self, source_rect: RectF) -> RectF:
        sl = source_rect.fl
        st = source_rect.ft
        sr = sl + source_rect.fw
        sb = st + source_rect.fh
        target_ltrb = self.map_ltrb([sl, st, sr, sb])

        return RectF(*target_ltrb)

    def serialize(self) -> Dict:
        return {
            'o': [self._ox, self._oy],
            'i': [self._ix, self._iy],
            'j': [self._jx, self._jy]
        }
