from typing import Final, List, Optional, Self, Set, Tuple

from edge.types.dir import Dir
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectI


class ITrack:
    @property
    def vrp_center_trace(self) -> List[Tuple[int, int]]:
        raise NotImplementedError


class LrTrackItem:
    def __init__(self, ts: int, plate_weight: int, plate_quad: QuadI, vehicle_rect: Optional[RectI]):
        self.ts: Final[int] = ts
        self.plate_weight: Final[int] = plate_weight
        self.plate_quad: Final[QuadI] = plate_quad
        self.vehicle_rect: Final[Optional[RectI]] = vehicle_rect


class RaTrackItem:
    def __init__(self,
                 frame_ts: int,
                 frame_l_rel: float,
                 frame_t_rel: float,
                 scene_ts: int,
                 scene_f_cm: int,
                 scene_l_cm: int,
                 speed_kmh: int) -> None:
        """
        :param frame_ts:
            The time when the data from the camera was received.
        :param frame_l_rel:
        :param frame_t_rel:
        :param scene_ts:
            The time when the data from the radar was received.
        :param scene_f_cm:
            The forward distance from the origin of the control zone to the vehicle, cm.
        :param scene_l_cm:
            The left distance from the origin of the control zone to the vehicle, cm.
        :param speed_kmh:
            Positive value means that the vehicle is moving in receeding direction relative to the control zone
            Negative value means that the vehicle is moving in approaching direction relative to the control zone
        """

        self.frame_ts: Final[int] = frame_ts
        self.frame_l_rel: Final[float] = frame_l_rel
        self.frame_t_rel: Final[float] = frame_t_rel

        self.scene_ts: Final[int] = scene_ts
        self.scene_f_cm: Final[int] = scene_f_cm
        self.scene_l_cm: Final[int] = scene_l_cm

        self.speed_kmh: Final[int] = speed_kmh

    @property
    def speed_dir(self):
        if self.speed_kmh > 0:
            return Dir.CODE_R
        elif self.speed_kmh < 0:
            return Dir.CODE_A
        else:
            return Dir.CODE_U


class DirTrackItem:
    def __init__(self, item_ts: int, item_dir: Dir) -> None:
        self.ts: Final[int] = item_ts
        self.dir: Final[Dir] = item_dir


class DirTrack(list[DirTrackItem]):
    """
    Contains items with a defined direction (approaching or receeding).
    """
    def __init__(self, items: List[DirTrackItem], items_dir: Dir) -> None:
        super().__init__(items)
        self.dir: Final[Dir] = items_dir

    def get_item(self, ts) -> Optional[DirTrackItem]:
        for item in self:
            if item.ts == ts:
                return item


class LaneTrackItem:
    def __init__(self, item_ts: int, item_lane: str) -> None:
        self.ts: Final[int] = item_ts
        self.lane: Final[str] = item_lane


class LaneTrack(list[LaneTrackItem]):
    def __init__(self, items: List[LaneTrackItem], lanes: Set[str]) -> None:
        super().__init__(items)
        self.lanes: Final[Set[str]] = lanes


class TrTrackItem:
    """
    Track item of an aggregated track.
    Contains data about the vehicle from different sources at the same timestamp.
    """
    def __init__(self, ts: int,
                 lr_item: Optional[LrTrackItem] = None,
                 ra_item: Optional[RaTrackItem] = None,
                 dir_item: Optional[DirTrackItem] = None,
                 lane_item: Optional[LaneTrackItem] = None) -> None:

        self.ts: Final[int] = ts

        self.lr_item: Final[LrTrackItem | None] = lr_item
        self.ra_item: Final[RaTrackItem | None] = ra_item
        self.dir_item: Final[DirTrackItem | None] = dir_item
        self.lane_item: Final[LaneTrackItem | None] = lane_item

    def __lt__(self, other: Self) -> bool:
        return self.ts < other.ts

    @property
    def vrp_text_quad(self) -> QuadI:
        if self.lr_item:
            return self.lr_item.plate_quad

        raise NotImplementedError()
