from typing import Final, Self

from edge.types.shape.rect import RectI
from edge.types.track import LrTrackItem
from edge.types.timespan import TimeSpan


class ImPlace:
    def __init__(self, r: int, x: int, y: int) -> None:
        """
        :param r:
            License plate's bbox height, pixels.
        :param x:
            Abbscissa of the center of the license plate's bbox, pixels.
        :param y:
            Ordinate of the center of the license plate's bbox, pixels.
        """
        self._r: Final[int] = r
        self._min_x: Final[int] = x - r
        self._max_x: Final[int] = x + r
        self._min_y: Final[int] = y - r
        self._max_y: Final[int] = y + r

    def __contains__(self, point_xy: tuple[int, int]) -> bool:
        same_x: bool = self._min_x < point_xy[0] < self._max_x
        same_y: bool = self._min_y < point_xy[1] < self._max_y
        return same_x and same_y

    @property
    def xy(self) -> tuple[int, int]:
        return self._min_x + self._r, self._min_y + self._r

    @property
    def r(self) -> int:
        return self._r


class LrTrendItem:
    def __init__(self, place: ImPlace, span: TimeSpan, event: bool = False) -> None:
        self.place: Final[ImPlace] = place
        self.range: Final[TimeSpan] = span
        self.event = event

    @property
    def min_ts(self) -> int:
        return self.range.min_ts

    @property
    def max_ts(self) -> int:
        return self.range.max_ts

    @property
    def duration_sec(self) -> int:
        return self.range.duration // 1000000

    def set_event(self) -> None:
        self.event = True

    def clone(self) -> Self:
        return LrTrendItem(self.place, TimeSpan(self.range.min_ts, self.range.max_ts), self.event)


class LrTrend:
    """
    Aggregates the history of the license plate position in the same area of the control zone image
    """
    def __init__(self,
                 lost_tout_usec: int,
                 dead_tout_usec: int,
                 max_vrp_bbox_wh: tuple[int, int] = (0, 0),
                 items: list[LrTrendItem] = ()) -> None:
        """
        :param lost_tout_usec:
            Timeout of plate loss in camera field of view, usec.

            The current max_trend_item will be extended if the next track_item is within this timeout.
            The new max_trend_item will be created if the next track_item is outside this timeout.

        :param dead_tout_usec:
            It is considered that the vehicle has not left the control zone
            if it was absent in the camera's field of view for less than this timeout

            It is considered that the vehicle left the control zone
            if it was absent in the camera's field of view more than this timeout

        :param max_vrp_bbox_wh:
            The current maximum size of the license plate's bbox over the history of observations, pixels.

        :param items:
            The history of observations.
        """
        self._lost_tout_usec: Final[int] = lost_tout_usec
        self._dead_tout_usec: Final[int] = dead_tout_usec
        self._items: list[LrTrendItem] = list(items)

        self.max_vrp_bbox_w: int = max_vrp_bbox_wh[0]
        self.max_vrp_bbox_h: int = max_vrp_bbox_wh[1]

    @property
    def items(self) -> list[LrTrendItem]:
        return self._items

    @property
    def min_ts(self) -> int:
        return self._items[0].range.min_ts if len(self._items) else 0

    @property
    def max_ts(self) -> int:
        return self._items[-1].range.max_ts if len(self._items) else 0

    def set_event(self, event_ts: int) -> None:
        for item in self._items:  # type: LrTrendItem
            if event_ts in item.range:
                item.set_event()
                break

    def extend(self, track_items: list[LrTrackItem]) -> None:
        """
        Extends the LrTrend with the new LrTrackItems.

        :param track_items:
            It is expected that the list is sorted in ascending order of timestamp
        """
        max_trend_item: LrTrendItem | None = None
        for track_item in track_items:
            if max_trend_item is None:
                # cleanup
                if track_item.ts > self.max_ts + self._dead_tout_usec:
                    self._items.clear()

                if self._items:
                    max_trend_item = self._items[-1]

            # extend max_trend_item
            vrp_bbox: RectI = track_item.plate_quad.bbox_recti
            if max_trend_item:
                # update current max_vrp_bbox.
                if vrp_bbox.iw > self.max_vrp_bbox_w:
                    if vrp_bbox.ih > self.max_vrp_bbox_h:
                        self.max_vrp_bbox_w = vrp_bbox.iw
                        self.max_vrp_bbox_h = vrp_bbox.ih

                if vrp_bbox.center in max_trend_item.place:
                    if track_item.ts < (max_trend_item.range.max_ts + self._lost_tout_usec):
                        max_trend_item.range.max_ts = track_item.ts
                        continue

            # create new max_trend_item
            max_trend_item = LrTrendItem(
                ImPlace(vrp_bbox.ih, *vrp_bbox.center),
                TimeSpan(track_item.ts, track_item.ts))

            self._items.append(max_trend_item)

    def get_vrp_bbox(self, item: LrTrendItem) -> RectI:
        plate_wh_ratio = float(self.max_vrp_bbox_w) / self.max_vrp_bbox_h

        center_x, center_y = item.place.xy
        half_h: float = item.place.r / 2.0
        half_w: float = half_h * plate_wh_ratio

        return RectI(
            center_x - half_w,
            center_y - half_h,
            center_x + half_w,
            center_y + half_h)

    def retrend(self, lost_tout: int = 0) -> Self:
        new_item: LrTrendItem = self._items[0].clone()
        new_items: list[LrTrendItem] = [new_item]

        for old_item in self._items:
            if old_item.place.xy in new_item.place:
                if old_item.range.min_ts < (new_items[-1].range.max_ts + lost_tout):
                    new_item.range.max_ts = old_item.range.max_ts
                    if old_item.event:
                        new_item.set_event()

                    continue

            new_item = old_item.clone()
            new_items.append(new_item)

        return LrTrend(self._lost_tout_usec, self._dead_tout_usec,
                       (self.max_vrp_bbox_w, self.max_vrp_bbox_h), new_items)
