import os
from typing import cast

from edge import eventlog
from edge.support import fs, net, sh
from edge.support.cgi import CgiBackup
from edge.types.exceptions import InternalError, ParamsError
from edge.webserver.sysconf.api_datetime.chronyd import setup_chronyd
from edge.webserver.sysconf.api_datetime.gpsd import map_enpoint_gnss_to_pps
from edge.webserver.sysconf.api_datetime.ntpd import setup_ntpd
from edge.webserver.sysconf.api_datetime.types import SourceConf, GnssSourceConf, NtpSourceConf
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.utype import *


def _get_gnss_sync_source(sync_sources: list[SourceConf], gnss_jnode: dict | None) -> GnssSourceConf:
    gnss_sources: list[SourceConf] = [source for source in sync_sources if source['type'] == 'gnss']
    gnss_sources_count = len(gnss_sources)
    if gnss_sources_count > 1:
        eventlog.log_d(f'fail: ParamsError (gnss-sources-count:{gnss_sources_count})')
        raise ParamsError

    autosync: bool = gnss_sources[0].get('autosync', False) if gnss_sources_count == 1 else False


    if gnss_jnode:
        pps_offset: float = gnss_jnode.get('pps-offset') or 0.0
        gnss_offset: float = gnss_jnode.get('gnss-offset') or 0.0
        gnss_endpoint: str = gnss_jnode['endpoint']

        from os.path import realpath
        gnss_endpoint_rel: str = realpath(gnss_endpoint)
        pps_endpoint_rel: str | None = map_enpoint_gnss_to_pps(gnss_endpoint_rel)

        return GnssSourceConf(autosync, gnss_endpoint_rel, pps_endpoint_rel, gnss_offset, pps_offset)

    elif autosync:
        eventlog.log_e('fail: target-autosync:true, time.json#/ro/gnss:null')
        raise InternalError

    else:
        return GnssSourceConf()


def _get_ntp_sync_source(sync_sources: list[SourceConf]) -> NtpSourceConf:
    ntp_sources: list[SourceConf] = [source for source in sync_sources if source['type'] == 'ntp']
    ntp_sources_count = len(ntp_sources)
    if ntp_sources_count == 1:
        return cast(NtpSourceConf, ntp_sources[0])

    elif ntp_sources_count == 0:
        return {
            'autosync': False,
            'type': 'ntp',
            'servers': []
        }

    else:
        eventlog.log_d('fail: ParamsError (#/args/sources)')
        raise ParamsError


class TimeApi:
    _GNSS_SOURCE_LIMITS = {
        'type': 'gnss',
        'sync-auto': True,
        'sync-manual': False
    }

    _GNSS_SOURCE_CONF: GnssSourceConf = {
        'type': 'gnss',
        'autosync': False
    }

    @classmethod
    def load_unit(cls) -> dict:
        TimeUiconf.install()

        uiconf = TimeUiconf.load_conf()
        uiconf_pub: dict = uiconf['rw']

        ext_uiconf_pub = uiconf_pub
        ext_sources = ext_uiconf_pub['args']['sources']

        # patch config if GNSS device detected
        if len([x for x in ext_sources if x['type'] == 'gnss']) == 0:
            if net.query_gps_count() > 0:
                ext_sources.append(cls._GNSS_SOURCE_CONF)

        if len([x for x in ext_uiconf_pub['limits']['sources'] if x['type'] == 'gnss']) == 0:
            ext_uiconf_pub['limits']['sources'].append(cls._GNSS_SOURCE_LIMITS)

        TimeUiconf.save_conf(uiconf)

        unit_ext_jnode = {
            'type': UTYPE_TIME,
            'conf': ext_uiconf_pub['args'],
            'limits': ext_uiconf_pub['limits']
        }

        return unit_ext_jnode

    @classmethod
    @fs.Lock('/var/run/edge/datetime', 2)
    @CgiBackup('/etc/edge/setupstore/backup/datetime', {
        'time.json': TimeUiconf.CONF_PATH
    })
    def save_unit(cls, new_uconf: dict) -> dict:
        fs.validate_json({'args': new_uconf['conf']}, TimeUiconf.SCHEMA_PATH)  # TODO
        new_uconf_conf = new_uconf['conf']

        unit_uconf = TimeUiconf.load_conf()
        unit_uconf_pri: dict = unit_uconf['ro']
        unit_uconf_pub: dict = unit_uconf['rw']

        if new_uconf_conf['tz'] not in unit_uconf_pub['limits']['tz']:
            eventlog.log_d('fail: ParamsError (#/args/tz)')
            raise ParamsError

        gnss_jnode: dict | None = unit_uconf_pri.get('gnss')
        next_gnss_source: GnssSourceConf = _get_gnss_sync_source(new_uconf_conf.get('sources', []), gnss_jnode)
        next_ntp_source: NtpSourceConf = _get_ntp_sync_source(new_uconf_conf.get('sources', []))

        timekeeper_type = os.environ.get('EDGE_TIMEKEEPER__TYPE') or 'ntp'
        if timekeeper_type == 'ntp':
            setup_ntpd(next_ntp_source, next_gnss_source)
        elif timekeeper_type == 'chrony':
            setup_chronyd(next_ntp_source, next_gnss_source)
        else:
            eventlog.log_d('fail: EDGE_TIMEKEEPER__TYPE')
            raise ParamsError

        # if tz changed restart
        if unit_uconf_pub['args']['tz'] != new_uconf_conf['tz']:
            eventlog.log_i('exec: restart overlayrender, speedcontrol, webserver, tracker')
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_RENDERER)
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_SPEEDCONTROL)
            dg.LifeCycle.sigkill_dgmods(dg.EDGE_TYPE_TRACKER)
            dg.LifeCycle.restart_ws_dgmod()
            # TODO: restart edgeserver if next_time < current_time

        # write new jnode to disk
        unit_uconf_pub['args'] = new_uconf_conf
        TimeUiconf.save_conf(unit_uconf)

        return cls.load_unit()  # TODO: strip limits from uiconf

    @classmethod
    def sync_manual(cls, sync_source: dict) -> dict:
        import datetime

        fs.validate_json(sync_source, TimeUiconf.SCHEMA_MANUAL_PATH)
        source_type = sync_source.get('type')
        if source_type == 'ntp':
            cls._sync_manual_ntp(sync_source)
            return {
                'type': 'ntp',
                'time': datetime.datetime.now().isoformat()
            }

        elif source_type == 'user':
            cls._sync_manual_user(sync_source)
            return {
                'type': 'user',
                'time': datetime.datetime.now().isoformat()
            }

        else:
            eventlog.log_i('fail: ParamsError (#/type)')
            raise ParamsError

    @staticmethod
    def _sync_manual_ntp(sync_source: dict) -> None:
        # set OS time
        servers = [sh.quote(server['address']) for server in sync_source.get('servers', [])]
        if not servers:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        timekeeper_type = os.environ.get('EDGE_TIMEKEEPER__TYPE') or 'ntp'
        if timekeeper_type == 'ntp':
            sh_cmd = f'ntpdate -u {" ".join(servers)}'
        elif timekeeper_type == 'chrony':
            sh_cmd = f'chronyc makestep'
        else:
            eventlog.log_d('fail: EDGE_TIMEKEEPER__TYPE')
            raise ParamsError

        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

    @staticmethod
    def _sync_manual_user(sync_source: dict) -> None:
        import iso8601
        tz_datetime = iso8601.parse_date(sync_source['time'])
        utc_datetime = tz_datetime - tz_datetime.utcoffset()

        # set OS time
        sh_cmd = 'date --set="{:%Y-%m-%d %H:%M:%S}"'.format(utc_datetime)
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: http500. sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError

        # sync BIOS time to OS time
        sh_cmd = 'hwclock -w'
        answer = sh.run_silent(sh_cmd)
        if answer == 0:
            eventlog.log_i(f'done: sh=<{sh_cmd}>')
        else:
            eventlog.log_e(f'fail: http500. sh=<{sh_cmd}> rc=<{answer}>')
            raise InternalError
