from typing import Final

from edge import eventlog
from edge.support import fs, sh
from edge.webserver.sysconf.api_datetime.gpsd import GpsdConfTweaker
from edge.webserver.sysconf.api_datetime.types import NtpSourceConf, GnssSourceConf


class _NtpdateConfTweaker:
    NTPSEC_NTPDATE_DEB__CONF_PATH = '/etc/default/ntpsec-ntpdate'
    NTPDATE_DEB__CONF_PATH = '/etc/default/ntpdate'

    @classmethod
    def _get_conf_path(cls):
        if fs.path_exists(cls.NTPSEC_NTPDATE_DEB__CONF_PATH):
            # bookworm
            return cls.NTPSEC_NTPDATE_DEB__CONF_PATH
        elif fs.path_exists(cls.NTPDATE_DEB__CONF_PATH):
            # buster
            return cls.NTPDATE_DEB__CONF_PATH
        else:
            eventlog.log_e('fail: unsupported')
            raise NotImplementedError

    @classmethod
    def tweak(cls, ntp_autosync: bool) -> None:
        conf_path = cls._get_conf_path()

        with open(conf_path, 'r') as text_file:
            curr_lines = text_file.readlines()

        # Remove lines starting with 'exit 0'
        next_lines = [line for line in curr_lines if not line.startswith('exit 0')]

        if ntp_autosync:
            # do sync time using ntpdate on system startup
            pass
        else:
            # do not sync time using ntpdate on system startup
            next_lines.insert(0, 'exit 0\n')

        with open(conf_path, 'w') as text_file:
            text_file.write(''.join(next_lines))


class _NtpConfBuider:
    CONF_PATH = '/etc/ntp.conf'

    def __init__(self) -> None:
        self._footer: Final[list[str]] = [
            '#',
            'driftfile /var/lib/ntp/ntp.drift',
            '#statsdir /var/log/ntpstats/',
            'statistics loopstats peerstats clockstats',
            'filegen loopstats file loopstats type day enable',
            'filegen peerstats file peerstats type day enable',
            'filegen clockstats file clockstats type day enable',
            '#',
            'restrict -4 default kod notrap nomodify nopeer noquery',
            'restrict -6 default kod notrap nomodify nopeer noquery',
            'restrict 127.0.0.1',
            'restrict ::1',
            '#',
            'tinker panic 0'
        ]

    @staticmethod
    def _encode_ntp_source(lines: list[str], ntp_source: NtpSourceConf) -> bool:
        if not ntp_source.get('autosync', False):
            return False

        for server_conf in ntp_source['servers']:
            address = server_conf['address']
            if server_conf['prefered']:
                lines.append(f'server {address} prefer')
            else:
                lines.append(f'server {address}')

        for server_conf in ntp_source['servers']:
            address = server_conf['address']
            lines.append(f'restrict {address} noquery notrap')

        return True

    @staticmethod
    def _encode_gnss_source(lines: list[str], gnss_source: GnssSourceConf) -> None:
        assert gnss_source.gnss_enpoint is not None
        lines.extend([
            f'# GPS',
            f'server 127.127.28.0 minpoll 4 maxpoll 4 prefer',
            f'fudge 127.127.28.0 time1 {gnss_source.gnss_offset} flag1 1 refid GPS',
            f''
        ])

        if gnss_source.pps_enpoint is not None:
            pps_port = gnss_source.pps_enpoint[-1]
            lines.extend([
                f'# PPS',
                f'server 127.127.22.{pps_port}',
                f'fudge 127.127.22.{pps_port} flag2 1 flag3 1 refid PPS',
                f'tos mindist 0.01',
                f''
            ])

    def build(self, ntp_source: NtpSourceConf, gnss_source: GnssSourceConf) -> bool:
        lines: list[str] = []

        ntp_autosync = self._encode_ntp_source(lines, ntp_source)
        if gnss_source.autosync:
            self._encode_gnss_source(lines, gnss_source)

        lines.extend(self._footer)

        with open(self.CONF_PATH, 'w') as text_file:
            text_file.write('\n'.join(lines))

        return ntp_autosync or gnss_source.autosync


def setup_ntpd(ntp_source: NtpSourceConf, gnss_source: GnssSourceConf) -> None:
    _NtpdateConfTweaker.tweak(ntp_source.get('autosync', False))
    autosync = _NtpConfBuider().build(ntp_source, gnss_source)
    GpsdConfTweaker.tweak(gnss_source.endpoints)

    sh.run_silent('systemctl stop ntp')
    sh.run_silent('systemctl disable ntp')
    if autosync:
        sh.run_silent('systemctl enable ntp')
        sh.run_silent('systemctl start ntp')

    sh.run_silent('systemctl restart gpsd')
