import subprocess
import os
import shutil
import time

from edge.support import debinterface

import edge.support.networking
from edge import eventlog
from edge.support import networking, fs
from edge.support.networking import set_hostname, PATH_ETC_HOSTNAME, PATH_ETC_HOSTS
from edge.types.exceptions import *
from edge.webserver.sysconf.uimod.utype import *


UI_NETWORK_SCHEMA_FILEPATH = '/etc/edge/setupstore/schemas/network.json'
_UICONF_PATH = '/etc/edge/setupstore/admin/actions/network.json'
_DEF_UICONF_PATH = '/opt/edge/share/setupstore/ui/network-jnode.json'

LOCK_FILEPATH = '/var/run/edge/network'

RESTART_NONE = 0
RESTART_HOSTNAME = 1
RESTART_NETWORKING = 2

BACKUP_DIRPATH = '/etc/edge/setupstore/backup/network.' + str(round(time.time() * 1000000)) + '/'


def _try_create_backup() -> bool:
    # noinspection PyBroadException
    try:
        os.makedirs(BACKUP_DIRPATH)
        shutil.copy(PATH_ETC_HOSTNAME, BACKUP_DIRPATH)
        shutil.copy(PATH_ETC_HOSTS, BACKUP_DIRPATH)
        shutil.copy(_UICONF_PATH, BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def _try_restore_backup() -> bool:
    # noinspection PyBroadException
    try:
        shutil.copy(BACKUP_DIRPATH + 'hostname', PATH_ETC_HOSTNAME)
        shutil.copy(BACKUP_DIRPATH + 'hosts', PATH_ETC_HOSTS)
        shutil.copy(BACKUP_DIRPATH + 'network.json', _UICONF_PATH)
        os.rmdir(BACKUP_DIRPATH)
        return True

    except Exception:
        return False


def _remove_backup() -> None:
    # noinspection PyBroadException
    try:
        shutil.rmtree(BACKUP_DIRPATH)
    except Exception:
        pass


def _check_iface_changed(old_jnode: dict, new_jnode: dict):
    if old_jnode['ip-mode'] != new_jnode['ip-mode']:
        return True

    if old_jnode['ip-mode'] == 'static':
        if old_jnode['ip-address'] != new_jnode['ip-address']:
            return True

        if old_jnode['ip-mask'] != new_jnode['ip-mask']:
            return True

        if old_jnode['ip-gateway'] != new_jnode['ip-gateway']:
            return True
    else:
        pass

    if old_jnode['dns-mode'] != new_jnode['dns-mode']:
        return True

    if old_jnode['dns-mode'] == 'static':
        if old_jnode['dns-1'] != new_jnode['dns-1']:
            return True

        if old_jnode['dns-2'] != new_jnode['dns-2']:
            return True
    else:
        pass

    return False


def _edit_iface(old_jnode: dict, new_jnode: dict) -> int:
    if not _check_iface_changed(old_jnode, new_jnode):
        return RESTART_NONE

    mac_address = new_jnode['mac']

    ifname = edge.support.networking.map_ifmac_to_ifname(mac_address)

    iface_conf = debinterface.interfaces.Interfaces(interfaces_path=networking.PATH_CONFD_PRIMARY)

    adapter_list = [item for item in iface_conf.adapters if item.export(options_list=['name'])['name'] == ifname]

    if len(adapter_list) != 1:
        raise ValueError
    adapter = adapter_list[0]

    adapter.reset()
    if new_jnode['ip-mode'] == 'static':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])
        adapter.setAddress(new_jnode['ip-address'])
        adapter.setNetmask(new_jnode['ip-mask'])
        adapter.setGateway(new_jnode['ip-gateway'])

    elif new_jnode['ip-mode'] == 'dhcp':
        adapter.setName(ifname)
        adapter.setAuto(True)
        adapter.setAddrFam('inet')
        adapter.setAddressSource(new_jnode['ip-mode'])

    else:
        raise NotImplementedError(f'unsupported ip-mode {new_jnode["ip-mode"]}')

    if new_jnode['dns-mode'] == 'static':
        dns_1 = new_jnode['dns-1']
        if dns_1:
            adapter.setDnsNameservers(dns_1)

        dns_2 = new_jnode['dns-2']
        if dns_2:
            adapter.setDnsNameservers(dns_1 + ' ' + dns_2)

    elif new_jnode['dns-mode'] == 'dhcp':
        pass

    else:
        raise NotImplementedError(f'unsupported dns-mode {new_jnode["dns-mode"]}')

    iface_conf.writeInterfaces()

    return RESTART_NETWORKING


# noinspection PyClassHasNoInit
class NetworkApi:
    # noinspection PyUnusedLocal
    @staticmethod
    def load_unit(uref: str) -> dict:
        def install_uiconf(def_uiconf_path: str, uiconf_path: str) -> None:
            if not os.path.exists(uiconf_path):
                try:
                    with open(networking.PATH_MANAGEMENT_MAC, 'r') as text_file:
                        mac_addr = text_file.readline().strip()

                except IOError:
                    eventlog.log_w('management-mac file not found, using eth0 as management interface')
                    mac_addr = edge.support.networking.map_ifname_to_ifmac('eth0')
                    if not mac_addr:
                        eventlog.log_d('fail: InternalError (get ifmac)')
                        raise InternalError

                def_uiconf = fs.load_json3(def_uiconf_path)
                def_uiconf_pri: dict = def_uiconf['ro']
                def_uiconf_pub: dict = def_uiconf['rw']

                def_uiconf_pri['mac'] = mac_addr
                def_uiconf_pub['hostname'] = subprocess.getoutput('hostname')

                fs.write_json_file(def_uiconf, uiconf_path, pretty=True)

        def install_uischema(uiconf_path: str, uischema_path: str) -> None:
            if not os.path.islink(uischema_path):
                fs.ln_schema(uischema_path, uiconf_path)

            elif not os.path.exists(os.readlink(uischema_path)):
                fs.ln_schema(uischema_path, uiconf_path)

        install_uiconf(_DEF_UICONF_PATH, _UICONF_PATH)
        install_uischema(_UICONF_PATH, UI_NETWORK_SCHEMA_FILEPATH)

        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        if unit_uconf is None:
            eventlog.log_e('fail: InternalError')
            raise InternalError

        unit_uconf_pub = unit_uconf['rw']  # type: dict
        unit_uconf_pri = unit_uconf['ro']  # type: dict

        unit_uconf_pub['mac'] = unit_uconf_pri['mac']
        return {
            'type': UTYPE_NETWORK,
            'conf': unit_uconf['rw']
        }

    # noinspection PyUnusedLocal
    @staticmethod
    def save_unit(uref: str, new_uconf: dict) -> dict:
        unit_uconf = fs.load_json3(_UICONF_PATH, default=None)
        if unit_uconf is None:
            eventlog.log_e('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf['conf']
        fs.validate_json(new_uconf_conf, UI_NETWORK_SCHEMA_FILEPATH)

        if not fs.try_lock_pid_file(LOCK_FILEPATH):
            eventlog.log_d('fail: BusyError')
            raise BusyError

        if not _try_create_backup():
            eventlog.log_e('fail: InternalError')
            raise InternalError

        unit_uconf_pub = unit_uconf['rw']
        try:
            restart_mask = RESTART_HOSTNAME if set_hostname(new_uconf_conf['hostname']) else RESTART_NONE
            restart_mask |= _edit_iface(unit_uconf_pub, new_uconf_conf) or restart_mask

            unit_uconf['rw'] = new_uconf_conf.copy()
            unit_uconf['rw'].pop('mac')
            fs.write_json_file(unit_uconf, _UICONF_PATH, pretty=True)

            if restart_mask & RESTART_HOSTNAME:
                os.system(f'hostnamectl set-hostname {new_uconf_conf["hostname"]}')

            if restart_mask & RESTART_NETWORKING:
                os.system('service networking restart')

            _remove_backup()
            unit_ext_jnode = {
                'type': UTYPE_NETWORK,
                'conf': new_uconf_conf
            }

            return unit_ext_jnode

        except KeyError:
            eventlog.log_d('fail: ParamsError (mac-address)')
            raise ParamsError

        except Exception:
            eventlog.log_x()
            if _try_restore_backup():
                _remove_backup()

            raise InternalError
