import os

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf.dgstats.camera import CamDgStats
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *

def _get_live_image(ui_pid: str) -> str | None:
    # get last overlay imagepath
    or_id = int(ui_pid) * 10
    or_store = f'/var/spool/edge/pipeline-{ui_pid}/or{or_id}-or{or_id}-store/'

    file_names: list[str] = os.listdir(or_store)

    if len(file_names):
        def decode_ts(file_name: str) -> int:
            return int(file_name.split('-')[-1])

        blob_name = max(file_names, key=decode_ts)
        blob_path = os.path.join(or_store, blob_name)
        return blob_path

# noinspection PyClassHasNoInit
class ReprojectionsApi:
    @staticmethod
    def load_unit(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_REPROJECTIONS)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        # unit_uconf = {
        #     'rw': {
        #         'conf': {
        #             'active': True,
        #             'camera': {
        #                 'intrinsic': {
        #                     'frustum': {
        #                         'pixel-size': 3.45,
        #                         'focal-dist': 50,
        #                         'sensor-size': [1920, 1200]
        #                     }
        #                 },
        #                 'extrinsic': {
        #                     'position': {
        #                         'forward': 0,
        #                         'left': 0,
        #                         'up': 4.75
        #                     },
        #                     'rotation': {
        #                         'elevation': 0,
        #                         'azimuth': 0,
        #                         'roll': 0
        #                     }
        #                 }
        #             },
        #             'markings': {
        #                 'line-f': [1, 1, 2, 2, 3, 3],
        #                 'line-l': [10, 10, 20, 20, 30, 30],
        #                 'grid': [{
        #                     'anchor': {
        #                         'forward': 50,
        #                         'left': -5,
        #                         'up': 0
        #                     },
        #                     'size-f': 6,
        #                     'size-l': 3
        #                 }]
        #             }
        #         }
        #     }
        # }

        stats_jnode = CamDgStats.create(unit_umod.pid).get_stats()

        reply = {
            'type': UTYPE_REPROJECTIONS,
            'conf': unit_uconf['rw']['conf'],
            'stats': stats_jnode
        }
        return reply

    @staticmethod
    def save_unit(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_REPROJECTIONS)

        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')
        fs.validate_json(new_uconf_conf, unit_umod.schema_path)

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['conf'] != new_uconf_conf:
            unit_uconf['rw']['conf'] = new_uconf_conf

            from edge.support import hg
            with hg.ConfRepo() as (hg_dir, hg_client):
                hgjpeg_path = fs.join_path(hg_dir, unit_umod.pid, '0.jpeg')
                hgconf_path = fs.join_path(hg_dir, unit_umod.pid, f'{UTYPE_REPROJECTIONS}.0.json')

                add_files: bool = not fs.path_exists(hgconf_path)

                jpeg_path: str | None = _get_live_image(unit_umod.pid)
                if jpeg_path is None:
                    eventlog.log_e(f'fail: get snapshot (term-ref:{unit_umod.pid})')
                    raise InternalError

                fs.silent_cp(jpeg_path, hgjpeg_path)
                fs.write_json_file(unit_uconf, hgconf_path, pretty=True)

                if add_files:
                    hg_client.add([hgconf_path.encode('ascii'), hgjpeg_path.encode('ascii')])

                hg_client.commit(message=f'edit({unit_umod.pid}/reprojections.0)', user='admin')

                uirefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
                creator.generate(unit_umod.create_backup_dirpath(), uirefs_l0_l1, {unit_umod.ref: unit_uconf})

        stats_jnode = CamDgStats.create(unit_umod.pid).get_stats()
        return {
            'type': UTYPE_REPROJECTIONS,
            'conf': new_uconf_conf,
            'stats': stats_jnode
        }
