from edge import eventlog
from edge.support import fs
from edge.support.template import FormatTemplate
from edge.types.exceptions import *
from edge.webserver.sysconf import edge_creator as creator
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.utype import *


class SeatbeltApi:
    @staticmethod
    def make_detector(term_uref: str, factory_conf: dict) -> tuple[str, dict]:
        if not unet.find_unit_r(rootinfo.decode_info().root, uref=term_uref):
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_utype = factory_conf.get('type')
        unit_name = factory_conf.get('name')

        if unit_utype != UTYPE_DET_SEATBELT or unit_name is None:
            eventlog.log_d('fail: ParamsError')
            raise ParamsError

        unit_uref, root_uinfo = ph.RootInfoUiconf.add_term_detector(term_uref, unit_utype, unit_name)
        unit_uconf = ph.build_term_unit_uconf(term_uref, unit_utype, unit_uref.ref, factory_conf)
        unit_umod = Umod(unit_uref.pid, unit_uref.type, unit_uref.id)
        unit_umod.install(unit_uconf)

        unit_conf = unit_umod.load_conf()
        unit_conf['rw']['args']['name'] = unit_name

        urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_conf})
        return unit_umod.ref, root_uinfo

    @staticmethod
    def drop_detector(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_DET_SEATBELT)
        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        root_uinfo, urefs_l1 = rootinfo.drop_unit_r(unit_umod.ref)
        creator.generate(unit_umod.create_backup_dirpath(), urefs_l1, forced=True)
        unit_umod.uninstall()

        reply = {
            'info': [root_uinfo]
        }
        return reply

    @staticmethod
    def load_detector(uref: str) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_DET_SEATBELT)

        unit_uconf = unit_umod.load_conf(default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        unit_uconf_pub = unit_uconf['rw']
        reply = {
            'type': UTYPE_DET_SEATBELT,
            'conf': unit_uconf_pub['args']
        }

        reply.update(ph.get_detector_limits_stats(uref))
        return reply

    @staticmethod
    def _validate_format(jnode: list[dict]) -> bool:
        key_to_true_map = {
            'speed': True,
            'datetime': True,
            'datetime-ms': True,
            'post-name': True,
            'geolocation': True,
            'plate': True,
            'lane': True,
            'vendor': True,
            'model': True,
            'serial': True,
            'mmr-class': True
        }

        for row in jnode:
            for col in row['cols']:
                try:
                    FormatTemplate(template=col['format']).substitute(key_to_true_map)
                except KeyError as x:
                    eventlog.log_d(f'fail: invalid key ({str(x)})')
                    return False

        return True

    @staticmethod
    def save_detector(uref: str, new_uconf: dict) -> dict:
        unit_umod = Umod.create(uref)
        unit_umod.assert_type(UTYPE_DET_SEATBELT)

        if unit_umod.load_conf(default=None) is None:
            eventlog.log_d('fail: NotFoundError')
            raise NotFoundError

        new_uconf_conf = new_uconf.get('conf')

        def validate_body() -> None:
            fs.validate_json({'args': new_uconf_conf}, unit_umod.schema_path)

            # validate 'format' jnodes from json
            entry_images = new_uconf_conf['images']['violation']

            entry_v_rows = entry_images['vehicle']['footer']['rows']
            if SeatbeltApi._validate_format(entry_v_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

            entry_s_rows = entry_images['scene']['footer']['rows']
            if SeatbeltApi._validate_format(entry_s_rows) is False:
                eventlog.log_d('fail: ParamsError')
                raise ParamsError

        validate_body()

        unit_uconf = unit_umod.load_conf()
        if unit_uconf['rw']['args'] != new_uconf_conf:
            if new_uconf_conf['name'] != unit_uconf['rw']['args']['name']:
                rootinfo.set_unit_name(unit_umod.ref, new_uconf_conf['name'])

            unit_uconf['rw']['args'] = new_uconf_conf

            urefs_l0_l1 = rootinfo.get_linked_urefs(unit_umod.ref)
            creator.generate(unit_umod.create_backup_dirpath(), urefs_l0_l1, {unit_umod.ref: unit_uconf})

        reply = {
            'type': UTYPE_DET_SEATBELT,
            'conf': new_uconf_conf
        }
        reply.update(ph.get_detector_limits_stats(uref))
        return reply
