import os

from edge import eventlog
from edge.support import fs, sh
from edge.support.cgi import CgiBackup
from edge.types.exceptions import *
from edge.webserver.sysconf.uimod.utype import *


class _Assets:
    DEF_UICONF_PATH = '/opt/edge/share/setupstore/ui/zabbix-jnode.json'
    UISCHEMA_PATH = '/etc/edge/setupstore/schemas/zabbix.json'
    UICONF_PATH = '/etc/edge/setupstore/admin/actions/zabbix.json'

    WEB_CONF = '/etc/zabbix/zabbix_agentd.d/agent.web.conf'


def _update_zabix_agent(uconf: dict) -> None:
    from edge.webserver.sysconf.uimod import rootinfo
    from edge.webserver.sysconf.uimod import unet

    uconf_pub: dict = uconf['rw']
    conf = uconf_pub['conf']
    zabbix_hostname = conf['agent-host']
    root_uinfo = rootinfo.decode_info().root
    urefs = unet.find_urefs_r(root_uinfo, utypes=[UTYPE_TERMINAL])
    hostname = zabbix_hostname
    for uref in urefs:
        hostname += f',{zabbix_hostname}.{uref}'

    conf_lines: list[str] = [
        f'Hostname={hostname}',
        f'',
        f'# Active checks related',
        f'ServerActive=' + ', '.join([
            f'{info["host"]}:{info["port"]:d}' for info in conf['servers']
        ]),
        f'',
        f'# Passive checks related',
        f'Server=' + ', '.join([
            info['host'] for info in conf['servers']
        ]),
        f'ListenPort={conf["agent-port"]}',
        f''
    ]

    with open(_Assets.WEB_CONF, 'w') as text_file:
        text_file.write('\n'.join(conf_lines))

    if conf['enabled']:
        sh.run_silent('systemctl enable zabbix-agent')
        sh.run_silent('systemctl restart zabbix-agent')
    else:
        sh.run_silent('systemctl stop zabbix-agent')
        sh.run_silent('systemctl disable zabbix-agent')


class ZabbixApi:
    @staticmethod
    def load_unit() -> dict:
        def install_uiconf(def_uiconf_path: str, uiconf_path: str) -> None:
            if not fs.path_exists(uiconf_path):
                fs.create_jnode_from_default(def_uiconf_path, uiconf_path)

        def install_uischema(uiconf_path: str, uischema_path: str) -> None:
            if os.path.islink(uischema_path):
                if fs.path_exists(os.readlink(uischema_path)):
                    return

            fs.ln_schema(uischema_path, uiconf_path)

        install_uiconf(_Assets.DEF_UICONF_PATH, _Assets.UICONF_PATH)
        install_uischema(_Assets.UICONF_PATH, _Assets.UISCHEMA_PATH)

        unit_uconf = fs.load_json3(_Assets.UICONF_PATH, default=None)
        if unit_uconf is None:
            eventlog.log_d('fail: InternalError')
            raise InternalError

        unit_uconf['rw']['type'] = UTYPE_ZABBIX_AGENT
        return unit_uconf['rw']

    @classmethod
    @fs.Lock('/var/run/edge/zabbix', 2)
    @CgiBackup('/etc/edge/setupstore/backup/zabbix', {
        'zabbix.json': _Assets.UICONF_PATH,
        'web.conf': _Assets.WEB_CONF,
    })
    def save_unit(cls, unit_jnode: dict) -> dict:
        fs.validate_json(unit_jnode, _Assets.UISCHEMA_PATH)

        # edit conf files
        curr_uconf = fs.load_json3(_Assets.UICONF_PATH)
        curr_uconf_pub: dict = curr_uconf['rw']
        if curr_uconf_pub != unit_jnode:
            if not unit_jnode['conf']['servers']:
                unit_jnode['conf']['servers'] = curr_uconf_pub['conf']['servers']

            curr_uconf_pub['conf'] = unit_jnode['conf']
            fs.write_json_file(curr_uconf, _Assets.UICONF_PATH, pretty=True)

            _update_zabix_agent(curr_uconf)

        unit_ext_jnode = unit_jnode
        unit_ext_jnode['type'] = UTYPE_ZABBIX_AGENT
        return unit_jnode

    @classmethod
    def generate_web_conf(cls):
        # The configuration file is absent if there was no call to ZabbixApi::load_unit().
        uconf = fs.load_json3(_Assets.UICONF_PATH, default=None)
        if uconf:
            _update_zabix_agent(uconf)
