import os
import shutil

from abc import ABCMeta, abstractmethod
from typing import Final, Self

from edge import eventlog
from edge.support import fs

from edge.webserver.sysconf.types.conf import IUiConfProvider, IDgConfProvider
from edge.webserver.sysconf.types.typeref import ITypeRefProvider


class IBackup(metaclass=ABCMeta):
    @abstractmethod
    def remove(self) -> None:
        """
        The implementation must not raise any exceptions.
        """
        pass

    @abstractmethod
    def restore(self) -> None:
        """
        The implementation must not raise any exceptions.
        """
        pass


class UimodBackup(IBackup):
    def __init__(self, source_filepath: str, backup_filepath: str) -> None:
        self._source_filepath: Final[str] = source_filepath
        self._backup_filepath: Final[str] = backup_filepath

    @classmethod
    def create(cls, backup_dir: str, uimod: IUiConfProvider) -> Self | None:
        uiconf_source_path = uimod.conf_path
        uiconf_target_name = f'ui-{uimod.ref}-old.json'
        uiconf_target_path = fs.join_path(backup_dir, uiconf_target_name)
        try:
            shutil.copy2(uiconf_source_path, uiconf_target_path)
            return cls(uiconf_source_path, uiconf_target_path)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: cp {uiconf_source_path} to {uiconf_target_path} ({x})')

    def remove(self) -> None:
        os.remove(self._backup_filepath)

    def restore(self) -> None:
        try:
            shutil.copy2(self._backup_filepath, self._source_filepath)
            eventlog.log_d(f'done: restore {self._backup_filepath} to {self._source_filepath}')

        except (os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: cp {self._backup_filepath} to {self._source_filepath} ({x})')


class DgmodBackup(IBackup):
    def __init__(self, source_dirpath: str, backup_dirpath: str) -> None:
        self._source_dirpath: Final[str] = source_dirpath
        self._backup_dirpath: Final[str] = backup_dirpath

    @classmethod
    def create(cls, dir_path: str, dgmod: IDgConfProvider) -> Self | None:
        source_dirpath = dgmod.conf_dir
        backup_dirpath = fs.join_path(dir_path, f'edge-{dgmod.ref}')

        try:
            shutil.copytree(source_dirpath, backup_dirpath)
            return cls(source_dirpath, backup_dirpath)

        except (IOError, os.error, shutil.Error) as x:
            eventlog.log_e(f'fail: copytree {source_dirpath} to {backup_dirpath} ({x.message})')

    def remove(self) -> None:
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w(f'fail: rm {error_path} ({exc_info[1].message})')

        shutil.rmtree(self._backup_dirpath, ignore_errors=False, onerror=onerror_fn)

    def restore(self) -> None:
        def onerror_fn(error_func, error_path, exc_info):
            eventlog.log_w(f'fail: rm {error_path} ({exc_info[1].message})')

        try:
            eventlog.log_d(f'exec: restore {self._backup_dirpath} to {self._source_dirpath}')
            shutil.rmtree(self._source_dirpath, ignore_errors=False, onerror=onerror_fn)
            shutil.copytree(self._backup_dirpath, self._source_dirpath)
            eventlog.log_d('done')

        except shutil.Error as x:
            for srcname, dstname, why in x.args[0]:
                eventlog.log_e(f'fail: cp {srcname} to {dstname} ({why})')


class NewFileBackup(IBackup):
    def __init__(self, backup_filepath: str) -> None:
        self._backup_filepath: Final[str] = backup_filepath

    @classmethod
    def create(cls, backup_dir: str, uimod: ITypeRefProvider, new_uiconf: dict) -> Self | None:
        new_name = f'ui-{uimod.ref}-new.json'
        new_filepath = fs.join_path(backup_dir, new_name)
        # noinspection PyBroadException
        try:
            eventlog.log_d(f'exec: create {new_filepath}')
            fs.write_json_file(new_uiconf, new_filepath, pretty=True)

        except Exception:
            eventlog.log_x()

        # handle backup succcess
        jnode_backup = cls(new_filepath)
        eventlog.log_d('done')
        return jnode_backup

    def remove(self) -> None:
        os.remove(self._backup_filepath)

    def restore(self) -> None:
        pass
