from typing import Final

from edge.support import fs
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.eventstore import EsUimod
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def bs64_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_EVENTSTORE}):
        return []

    # if not forced:
    #     return []

    root_info = rootinfo.decode_info().root
    es_uirefs = unet.find_urefs_r(root_info, UTYPE_EVENTSTORE)
    es_uimods = list(map(EsUimod.create, es_uirefs))

    return [_BsCreator64(es_uimods)]


class _BsCreator64(IDgConfBuilder):
    def __init__(self, es_uimods: list[IUiConfProvider]) -> None:
        self._dgmod: Final[dg.DgMod] = dg.CTRL_BS_DGMOD

        self._uimods_es: Final[list[IUiConfProvider]] = es_uimods

        self._main_conf = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        jnode = {
            'links': [
                ['blobstore-server.0', 'blobstore-manager.0'],
                ['blobstore-stats.0', 'blobstore-server.0']
            ],
            'nodes': [
                {
                    'bundle': ['libunity_essentials', '1.0.0'],
                    'type': 'essentials',
                    'name': 'essentials.0',
                    'units': [
                        {
                            'type': 'counters',
                            'name': 'blobstore-stats.0',
                            'conf': {
                                'stat-path': fs.join_path('/var/spool/edge/system', 'zabbix', 'bs65535'),
                                'stat-tout': 60
                            }
                        }
                    ]
                },
                {
                    'bundle': ['libunity_blobstore', '1.0.0'],
                    'type': 'blobstore-server',
                    'name': 'blobstore-server.0',
                    'conf': {
                        'endpoint': f'/run/edge/blobstore.65535-uds',
                        'thread-count': 4
                    }
                },
                {
                    'bundle': ['libunity_blobstore', '1.0.0'],
                    'type': 'blobstore-manager',
                    'name': 'blobstore-manager.0',
                    'units': []
                }
            ]
        }

        manager_units: list = jnode['nodes'][2]['units']

        links = jnode['links']

        for es_uimod in self._uimods_es:
            es_uiconf = uiref_to_conf_map.get(es_uimod.ref) or es_uimod.load_conf()
            cleanup_uiconf = es_uiconf['rw']['args']['cleanup']
            chanref = es_uiconf['ro']['channel']

            archive_limits = dict()

            if cleanup_uiconf['max-du'][0]:
                archive_limits['max-du'] = cleanup_uiconf['max-du'][1]

            if cleanup_uiconf['min-df'][0]:
                archive_limits['min-df'] = cleanup_uiconf['min-df'][1]

            if cleanup_uiconf['max-abs-time'][0]:
                archive_limits['max-abs-time'] = cleanup_uiconf['max-abs-time'][1] * 60

            if cleanup_uiconf['max-rel-time'][0]:
                archive_limits['max-rel-time'] = cleanup_uiconf['max-rel-time'][1] * 60

            links.append(['blobstore-stats.0', f'blobstore-manager.0~chan.{chanref}'])
            links.append(['blobstore-stats.0', f'blobstore-manager.0~chan.{chanref}~blobs-archive.0'])
            links.append(['blobstore-stats.0', f'blobstore-manager.0~chan.{chanref}~blobs-cache.0'])

            manager_units.append({
                'type': 'blobstore-manager~chan',
                'name': f'blobstore-manager.0~chan.{chanref}',
                'conf': {
                    'ref': f'{chanref}',
                    'logs-name': f'bs-chan{chanref}',
                    'temp-dir': f'/var/opt/edge/blobstore/65535-cache/{chanref}-tmp'
                },
                'units': [
                    {
                        'type': 'blobstore-manager~chan~blobs-cache',
                        'name': f'blobstore-manager.0~chan.{chanref}~blobs-cache.0',
                        'conf': {
                            'directory': f'/var/opt/edge/blobstore/65535-cache/{chanref}',
                            'logs-name': f'bs-chan{chanref}-cache',
                            'drop-size': 8,
                            'limits': {
                                'max-abs-time': 3 * 60
                            }
                        }
                    },
                    {
                        'type': 'blobstore-manager~chan~blobs-archive',
                        'name': f'blobstore-manager.0~chan.{chanref}~blobs-archive.0',
                        'conf': {
                            'directory': f'/var/opt/edge/blobstore/65535/{chanref}',
                            'logs-name': f'bs-chan{chanref}-archive',
                            'limits': archive_limits
                        }
                    }
                ]
            })

        self._main_conf = jnode

    def build(self) -> None:
        dgmod = self.dgmod
        dgmod.install()

        unity_conf_dir = fs.join_path(dgmod.conf_dir, 'unit.conf.d')
        fs.silent_rm_r(unity_conf_dir)
        fs.silent_mkdir_r(unity_conf_dir)

        conf_path = fs.join_path(unity_conf_dir, 'blobstore.conf')
        fs.write_json_file(self._main_conf, conf_path, pretty=True)

        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        pass
