import subprocess

from typing import Final, Self

from edge import eventlog
from edge.support import fs, sh
from edge.webserver.sysconf.types.conf import IDgConfProvider
from edge.types.exceptions import GenericError

EDGE_TYPE_BLOBSTORE = 'blobstore'
EDGE_TYPE_BLOBSTOREW = 'blobstorew'
EDGE_TYPE_EVENTSTORE = 'eventstore'
EDGE_TYPE_GRABCONTROL = 'grabcontrol'
EDGE_TYPE_LEDDAR = 'leddar'
EDGE_TYPE_LPR = 'lpr'
EDGE_TYPE_MEDIACODEC = 'mediacodec'
EDGE_TYPE_MEDIACODEC_GST = 'mediacodecgst'
EDGE_TYPE_MEDIAGRABBER = 'mediagrabber'
EDGE_TYPE_RADAR = 'radar'
EDGE_TYPE_RAWCODEC = 'rawcodecgst'
EDGE_TYPE_RENDERER = 'overlayrender'
EDGE_TYPE_RTSP_GRABBER = 'rtspgrabber'
EDGE_TYPE_RTSP_SERVER = 'rtspserver'
EDGE_TYPE_SETUPSTORE = 'setupstore'
EDGE_TYPE_SPEEDCONTROL = 'speedcontrol'
EDGE_TYPE_TRACKER = 'tracker'
EDGE_TYPE_TRAFFIC_LIGHT = 'traffic-light'
EDGE_TYPE_WEBSERVER = 'webserver'
EDGE_TYPE_PSUCTL = 'psuctl'


# noinspection PyClassHasNoInit
class LifeCycle:
    @staticmethod
    def sigkill_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -9 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))

    @staticmethod
    def sigkill_dgmods(dgmod_type: str) -> None:
        cli_params = ['pkill', '-9', dgmod_type]
        cli_result = subprocess.run(cli_params)
        eventlog.log_i(f'done: <{cli_result.args}> (exitcode:{cli_result.returncode})')

    @classmethod
    def restart_ws_dgmod(cls) -> None:
        eventlog.log_i('exec')
        try:
            sh.exec_cli('/opt/edge/bin/setupstore-cli', ['restart', 'webserver'])
        except GenericError:
            eventlog.log_e('fail')

    @staticmethod
    def sighup_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -1 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))

    @staticmethod
    def sighup_dgmods(dgmod_type: str) -> None:
        cli_params = ['pkill', '-1', dgmod_type]
        cli_result = subprocess.run(cli_params)
        eventlog.log_i(f'done: <{cli_result.args}> (exitcode:{cli_result.returncode})')

    @staticmethod
    def sighup_ss_dgmod() -> None:
        eventlog.log_i('exec')
        try:
            sh.exec_cli('/opt/edge/bin/setupstore-cli', ['setup', 'setupstore'])
        except GenericError:
            eventlog.log_e('fail')

    @staticmethod
    def sigint_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -2 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))

    @staticmethod
    def sigusr1_dgmod(dgmod_type: str, dgmod_id: str) -> None:
        eventlog.log_i(f'exec: {dgmod_type}.{dgmod_id}')
        sh.run_silent(
            "kill -10 `ps ax |"
            "grep \"{} --id.{}\" |"
            "grep -v grep |"
            "awk '{{print $1}}'`".format(dgmod_type, dgmod_id))


class DgMod(IDgConfProvider):
    def __init__(self, dgmod_type: str, dgmod_id: str) -> None:
        self._type: Final[str] = dgmod_type
        self._ref: Final[str] = f'{dgmod_type}.{dgmod_id}'
        self._id: Final[str] = dgmod_id

    def __hash__(self) -> int:
        return hash(self._ref)

    def __str__(self) -> str:
        return f'<{self._ref}|{self._id}>'

    def __eq__(self, other: Self) -> bool:
        return isinstance(other, self.__class__) and getattr(other, 'ref', None) == self.ref

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def conf_dir(self) -> str:
        if self._type == EDGE_TYPE_SETUPSTORE:
            return f'/etc/edge/{self._type}'
        else:
            return f'/etc/edge/{self._ref}'

    def get_conf_path(self, conf_name: str) -> str:
        return fs.join_path(self.conf_dir, conf_name)

    def load_json_conf(self, conf_name: str, **kwargs) -> dict | None:
        conf_path = self.get_conf_path(conf_name)
        return fs.load_json3(conf_path, **kwargs)

    def load_main_conf(self, **kwargs) -> dict | None:
        return self.load_json_conf('main.conf', **kwargs)

    def save_json_conf(self, conf_name: str, conf_jnode: dict) -> None:
        conf_path = self.get_conf_path(conf_name)
        fs.write_json_file(conf_jnode, conf_path, pretty=True)

    def save_main_conf(self, conf_jnode: dict) -> None:
        return self.save_json_conf('main.conf', conf_jnode)

    def save_text_conf(self, conf_name: str, conf_text: str) -> None:
        conf_path = self.get_conf_path(conf_name)
        with open(conf_path, 'w') as text_file:
            text_file.write(conf_text)

    def save_haltaction_conf(self, conf_text: str) -> None:
        if len(conf_text) == 0:
            fs.silent_rm_r(self.get_conf_path('haltaction.conf'))
        else:
            self.save_text_conf('haltaction.conf', conf_text)

    def save_setupstore_conf(self, active: bool, debug: bool = False) -> None:
        log_level = 'debug' if debug else 'info'

        if active:
            conf_text = f'Active=true\nArgSet=--id={self._id} --log-level={log_level}\n'
        else:
            conf_text = f'Active=false\nArgSet=--id={self._id} --log-level={log_level}\n'

        self.save_text_conf('setupstore.conf', conf_text)

    def install(self) -> None:
        fs.silent_mkdir_r(self.conf_dir)

    def uninstall(self) -> None:
        fs.silent_rm_r(self.conf_dir)

    def sigkill(self) -> None:
        LifeCycle.sigkill_dgmod(self._type, self._id)

    def sighup(self) -> None:
        LifeCycle.sighup_dgmod(self._type, self._id)

    def sigint(self) -> None:
        LifeCycle.sigint_dgmod(self._type, self._id)

    def sigusr1(self) -> None:
        LifeCycle.sigusr1_dgmod(self._type, self._id)


CTRL_BS_DGMOD = DgMod(EDGE_TYPE_BLOBSTORE, dgmod_id='65535')
CTRL_ES_DGMOD = DgMod(EDGE_TYPE_EVENTSTORE, dgmod_id='65535')
CTRL_SS_DGMOD = DgMod(EDGE_TYPE_SETUPSTORE, dgmod_id='0')
CTRL_WS_DGMOD = DgMod(EDGE_TYPE_WEBSERVER, dgmod_id='0')
