from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfBuilder, UrefToConfMap, Udeps, IDgConfDeleter
from edge.webserver.sysconf.dgmod.gc.gc_main import MainConfBuilder
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def gc_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_MV_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref)
    ])

    if not activate:
        return []

    return [_GcCreator(uimod_cam)]


class _GcCreator(IDgConfBuilder):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_gc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_GRABCONTROL, uimod_cam.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam

        self._main_active = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_gc

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_cam.pid)

        cam_uiconf = uiref_to_conf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()
        cam_uiconf_control = cam_uiconf['rw']['videostream']['control']

        cam_uiconf_pri = cam_uiconf['ro']

        #
        exp_time_conf = [0, 0, 0]
        gain_conf = [0, 0, 0]
        gain_step = 1.0
        exptime_step = 20
        if cam_uiconf_control['profile'] == 'auto-70-150':
            gain_conf = [cam_uiconf_control['gain-auto-min'], cam_uiconf_control['gain-auto-max'], gain_step]
            exp_time_conf = [
                cam_uiconf_control['exposure-auto-min'],
                cam_uiconf_control['exposure-auto-max'],
                exptime_step]
        #

        gamma_limits = [cam_uiconf_control.get('gamma-auto-min', 0.4),
                        cam_uiconf_control.get('gamma-auto-max', 1.5),
                        0.05]

        target_median = cam_uiconf_pri.get('target-median') or [60, 80]
        main_builder = MainConfBuilder(gamma_limits, target_median)

        control_ir = cam_uiconf_pri.get('control-ir', True)
        if control_ir:
            main_builder.setup_ir_control()

        main_builder.setup_gain_conf(gain_conf)
        main_builder.setup_exp_time_conf(exp_time_conf)
        main_builder.setup_add_scene_img_react(spool_dir, f'mg{self._uimod_cam.id}-gc{self._uimod_cam.id}-scene')
        main_builder.setup_add_scene_control_event(f'/etc/edge/mediagrabber.{self._uimod_cam.id}/dyn/')

        main_conf = main_builder.build()
        ss_active = cam_uiconf_control['profile'] == 'auto-70-150'

        self._main_active = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._main_active

        dgmod = self._dgmod_gc
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _GcDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_gc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_GRABCONTROL, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_gc.uninstall()


def gc_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_MV_CAMERA:
        return [_GcDeleter(target_uimod)]
    return []
