from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.mc.mc_main import MainConfBuilder
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def mc_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref)
    ])

    if not activate:
        return []

    return [_McCreator(uimod_cam)]


class _McCreator(IDgConfBuilder):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_mc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_MEDIACODEC_GST, uimod_cam.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam

        self._main_ss: tuple[dict, bool] | None = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_mc

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_cam.pid)
        cam_uiconf = uiref_to_conf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()

        quality = 30000
        if cam_uiconf:
            quality = cam_uiconf['ro'].get('h264Quality', 30000)

        # create main.conf
        main_builder = MainConfBuilder()
        main_builder.setup_blobstore(self._uimod_cam.id, '0')
        main_builder.setup_react(spool_dir, 'mg{0}-mcgst{0}-scene'.format(self._uimod_cam.id))
        main_builder.setup_codec(quality)
        main_conf = main_builder.build()

        rgbs = cam_uiconf['ro'].get('rgbs', False)
        mcgst_scene = cam_uiconf['ro'].get('mcgst-scene', True)
        ss_active = mcgst_scene and not rgbs

        self._main_ss = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._main_ss

        edge_mc = self._dgmod_mc
        edge_mc.install()
        edge_mc.save_main_conf(main_conf)
        edge_mc.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sighup()


class _McDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_mc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_MEDIACODEC_GST, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_mc.uninstall()


def mc_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type in (UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA):
        return [_McDeleter(target_uimod)]
    return []
