from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.identity import Identity


class StatsUnitBuilder:
    def __init__(self, unit_name: str, spool_path: str) -> None:
        self._unit_name = unit_name
        self._spool_path = spool_path

    def build(self) -> tuple[str, dict]:
        uiref = Identity(self._unit_name)
        return self._unit_name, {
            'bundle': ['mediagrabber', '1.0.0'],
            'type': uiref.type,
            'name': uiref.ref,
            'conf': {
                'stat-path': fs.join_path(self._spool_path, 'zabbix', 'vs0'),
                'stat-tout': 5
            }
        }


class HaltUnitBuilder:
    def __init__(self, unit_name: str) -> None:
        self._unit_name: Final[str] = unit_name

    def build(self) -> tuple[str, dict]:
        uiref = Identity(self._unit_name)
        return self._unit_name, {
            'bundle': ['mediagrabber', '1.0.0'],
            'type': uiref.type,
            'name': uiref.ref
        }


class StbUnitBuilder:
    def __init__(self, unit_name: str, spool_path: str) -> None:
        self._unit_name = unit_name
        self._spool_path = spool_path

        self._endpoints = set[str]()

    def setup_endpoint(self, endpoint: str) -> None:
        self._endpoints.add(fs.join_path(self._spool_path, endpoint))

    def build(self) -> tuple[str, dict]:
        if not self._endpoints:
            eventlog.log_e('fail')
            raise NotImplementedError()

        uiref = Identity(self._unit_name)
        return self._unit_name, {
            'bundle': ['mediagrabber', '1.0.0'],
            'type': uiref.type,
            'name': uiref.ref,
            'conf': {
                "endpoints": list(self._endpoints)
            }
        }


class VstUnitBuilder:
    def __init__(self, unit_name: str, spool_path: str) -> None:
        self._unit_name = unit_name
        self._spool_path = spool_path

        self._busstore_blob_dir: str = ''
        self._busstore_min_file_count: int = 0
        self._busstore_max_file_count: int = 0

        self._bus_endpoints: list[str] = list()
        self._bus_raw_format_mono8: bool | None = None
        self._bus_format: str = ''

    def setup_busstore(self, blob_dir: str, min_file_count: int, max_file_count: int) -> None:
        self._busstore_blob_dir = fs.join_path(self._spool_path, blob_dir)
        self._busstore_min_file_count = min_file_count
        self._busstore_max_file_count = max_file_count

    def setup_bus_format(self, blob_format: str) -> None:
        self._bus_format = blob_format
        self._bus_raw_format_mono8 = blob_format == 'MONO8T'

    def setup_bus(self, args_dirs: list[str], require_mono8: bool = False) -> None:
        if not require_mono8:
            for args_dir in args_dirs:
                self._bus_endpoints.append(fs.join_path(self._spool_path, args_dir))

        elif self._bus_raw_format_mono8 is None:
            eventlog.log_e('fail')
            raise NotImplementedError()

        elif self._bus_raw_format_mono8:
            for args_dir in args_dirs:
                self._bus_endpoints.append(fs.join_path(self._spool_path, args_dir))

    def build(self) -> tuple[str, dict]:
        uiref = Identity(self._unit_name)

        return self._unit_name, {
            'bundle': ['mediagrabber', '1.0.0'],
            'type': uiref.type,
            'name': uiref.ref,
            'conf': {
                'busstore': {
                    'Store': self._busstore_blob_dir,
                    'FileCountLow': self._busstore_min_file_count,
                    'FileCountHigh': self._busstore_max_file_count
                },
                'bus': {
                    'PixelFormat': self._bus_format,
                    'Dirs': self._bus_endpoints
                }
            }
        }


class DriverUnitBuilder:
    def __init__(self, unit_name: str) -> None:
        self._unit_name = unit_name

        self._jnode = {
            'halt-timer': {
                'SignalAbsentTout': 3000,
                'DeviceAbsentTout': 60000
            },
            'properties': {
                'VideoSource': dict(),
                'irbox': list()
            }
        }

    def setup(self, vendor: str, model: str, serial: str) -> None:
        if vendor == 'Basler':
            driver_file = '/opt/edge/libexec/mediagrabber/libmgbasler.so'

        elif vendor == 'Teledyne DALSA':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-dalsa.so'

        elif vendor == 'FLIR':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-flir.so'

        elif vendor == 'PointGrey':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-pointgrey.so'

        elif vendor == 'JAI Corporation':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-jai.so'

        elif vendor == 'galaxy':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-galaxy.so'

        elif vendor == 'lucid':
            driver_file = '/opt/edge/libexec/mediagrabber/libdriver-arena.so'

        else:
            eventlog.log_e('fail')
            raise NotImplementedError()

        if model == 'Blackfly S-BFS-PGE-50S5M':
            self._jnode['halt-timer'] = {
                'SignalAbsentTout': 60000,
                'DeviceAbsentTout': 90000
            }

        self._jnode.update({
            'bundle': driver_file,
            'vendor': vendor,
            'model': model,
            'serial': serial,
        })

    def setup_cam_vs_acq(self, w: int, h: int, x: int, y: int, fps: int, rotate180: bool):
        self._jnode['properties']['VideoSource']['Acquisition'] = {
            'Sensor': {
                'Width': w,
                'OffsetX': x,
                'OffsetY': y,
                'Height': h
            },
            'Fps': fps,
            'Rotate180': rotate180
        }

    def setup_cam_vs_sweep_manual(self, exp_time: int, gain: int, gamma: float):
        self._jnode['properties']['VideoSource']['ManSetup'] = {
            'GainAll': gain,
            'ExpTime': exp_time,
            'Gamma': gamma,
            'IrLevel': 1
        }

    def setup_cam_pinbox_ir(self, signal: str, exposure_ratio: int, pin: str, inverter: bool) -> None:
        assert exposure_ratio > 0
        self._jnode['properties']['irbox'].append({
            'pin': pin,
            'source': signal,
            'exposure-ratio': exposure_ratio,
            'inverter': inverter
        })

    def setup_cam_misc(self, data_size: int):
        self._jnode['properties']['Misc'] = {
            'TransferDataSize': data_size
        }

    def build(self) -> tuple[str, dict]:
        uiref = Identity(self._unit_name)

        return self._unit_name, {
            'bundle': ['mediagrabber', '1.0.0'],
            'type': uiref.type,
            'name': uiref.ref,
            'conf': self._jnode
        }
