from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.ra.ra_halt import HaltConfBuilder
from edge.webserver.sysconf.dgmod.ra.ra_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.ra.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.ra.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.ra.uiconf.location import LocationUiConfProvider
from edge.webserver.sysconf.dgmod.ra.uiconf.mv_camera import MvCameraUiConfPorvider
from edge.webserver.sysconf.dgmod.ra.uiconf.rtsp_camera import RtspCameraUiConfPorvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def ra_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_RADAR}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_repro = target_uideps.find_any({UTYPE_REPROJECTIONS})
        uimod_radar = target_uideps.find_any({UTYPE_RADAR})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_psuctl = target_uideps.try_find_any({UTYPE_PSUCTL})

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref),
        uiref_to_newconf_map.get(uimod_radar.ref),
        uiref_to_newconf_map.get(uimod_repro.ref),
    ])

    if not activate:
        return []

    return [_RaCreator(uimod_radar, uimod_cam, uimod_fusion, uimod_repro, uimod_psuctl)]


class _RaCreator(IDgConfBuilder):
    def __init__(self,
                 uimod_ra: IUiConfProvider,
                 uimod_cam: IUiConfProvider,
                 uimod_fusion: IUiConfProvider,
                 uimod_repro: IUiConfProvider,
                 uimod_psuctl: IUiConfProvider | None) -> None:
        self._dgmod_ra: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RADAR, uimod_ra.id)

        self._uimod_ra: Final[IUiConfProvider] = uimod_ra
        self._uimod_cam: Final[IUiConfProvider] = uimod_cam
        self._uimod_fusion: Final[IUiConfProvider] = uimod_fusion
        self._uimod_repro: Final[IUiConfProvider] = uimod_repro
        self._uimod_psuctl: Final[IUiConfProvider | None] = uimod_psuctl

        self._main_halt = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_ra

    def setup(self, uiref_to_newconf_map: UrefToConfMap) -> None:
        spool_dir = ph.get_spool_dir(self._uimod_ra.pid)

        api_builder = None
        main_builder = MainConfBuilder(self._uimod_ra.id, spool_dir)
        halt_builder = HaltConfBuilder(self._uimod_psuctl)

        providers = list[IUiConfProvider2]()

        ra_uiconf = uiref_to_newconf_map.get(self._uimod_ra.ref)
        ra_uiconf_changed = ra_uiconf is not None
        if ra_uiconf is None:
            ra_uiconf = self._uimod_ra.load_conf()

        repro_uiconf = uiref_to_newconf_map.get(self._uimod_repro.ref)
        repro_uiconf_changed = repro_uiconf is not None
        if repro_uiconf is None:
            repro_uiconf = self._uimod_repro.load_conf()

        self._dgmod_ra.save_json_conf('runtime/reprojections.conf', repro_uiconf)

        providers.extend([
            LocationUiConfProvider(ra_uiconf, repro_uiconf, ra_uiconf_changed or repro_uiconf_changed)
        ])

        # send sms to module radar with new config
        if ra_uiconf_changed or repro_uiconf_changed:
            eventlog.log_d(f'stat: newconf:<{ra_uiconf}> or <{repro_uiconf}>')

            ra_uds_filepath = fs.join_path(spool_dir, f'ra{self._uimod_ra.id}-ra{self._uimod_ra.id}-uds')
            ra_lock_dirpath = f'/run/edge/radar.{self._uimod_ra.pid}.api'

            from .ra_api import ApiConfBuilder
            api_builder = ApiConfBuilder(ra_uds_filepath, ra_lock_dirpath)

        if self._uimod_cam.type == UTYPE_MV_CAMERA:
            ra_pri_device = ra_uiconf['ro']['device']
            offset_frames = ra_pri_device.get('time-offset') or 0

            uimod = self._uimod_cam
            providers.extend([
                MvCameraUiConfPorvider(uimod.id, uiref_to_newconf_map.get(uimod.ref) or uimod.load_conf(), repro_uiconf, offset_frames)
            ])

        elif self._uimod_cam.type == UTYPE_RTSP_CAMERA:
            uimod = self._uimod_cam
            providers.extend([
                RtspCameraUiConfPorvider(uimod.id, uiref_to_newconf_map.get(uimod.ref) or uimod.load_conf(), repro_uiconf)
            ])

        else:
            eventlog.log_e('fail')
            raise NotImplementedError()

        providers.extend([
            FusionUiConfProvider(self._uimod_fusion.id)
        ])

        for provider in providers:
            provider.setup_builders(api_builder, main_builder, halt_builder)

        if api_builder is not None:
            api_builder.build()

        main_conf = main_builder.build()
        halt_conf = halt_builder.build()

        self._main_halt = main_conf, halt_conf

    def build(self) -> None:
        main_conf, halt_conf = self._main_halt

        dgmod = self._dgmod_ra
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_haltaction_conf(halt_conf)
        dgmod.save_setupstore_conf(active=True)

    def notify(self) -> None:
        pass


class _RaDeleter(IDgConfDeleter):
    def __init__(self, uimod_ra: IUiConfProvider) -> None:
        self._dgmod_ra: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RADAR, uimod_ra.id)

    def run(self) -> None:
        self._dgmod_ra.uninstall()


def ra_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_RADAR:
        return [_RaDeleter(target_uimod)]
    return []
