from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.rc.rc_main import MainConfBuilder
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def rc_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_MV_CAMERA}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_lpr = target_uideps.find_any({UTYPE_LPR})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    activate = forced or any([
        uiref_to_newconf_map.get(uimod_cam.ref)
    ])

    if not activate:
        return []

    return [_RcCreator(uimod_cam, uimod_lpr, uimod_fusion)]


class _RcCreator(IDgConfBuilder):
    def __init__(self,
                 uimod_cam: IUiConfProvider,
                 uimod_lpr: IUiConfProvider,
                 uimod_fusion: IUiConfProvider) -> None:
        self._dgmod_rc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RAWCODEC, uimod_cam.id)

        self._uimod_cam: Final[IUiConfProvider] = uimod_cam
        self._uimod_fusion: Final[IUiConfProvider] = uimod_fusion
        self._uimod_lpr: Final[IUiConfProvider] = uimod_lpr

        self._main_ss = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_rc

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        main_builder = MainConfBuilder(self._uimod_cam.id, ph.get_spool_dir(self._uimod_cam.pid))

        cam_uiconf = uiref_to_conf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()

        main_builder.setup_react('mg{0}-rc{0}-scene'.format(main_builder.uimod_id))
        main_builder.setup_event_blobs('rc{0}-rc{0}-store'.format(main_builder.uimod_id))
        main_builder.setup_event_args([
            f'rc{main_builder.uimod_id}-lr{self._uimod_lpr.id}-scene',
            f'rc{main_builder.uimod_id}-tr{self._uimod_fusion.id}-scene',
            'mg{0}-mcgst{0}-scene'.format(main_builder.uimod_id)
        ])

        ss_active = cam_uiconf['ro']['raw-format'] != 'MONO8T'
        if ss_active and cam_uiconf['rw']['videostream']['control']['profile'] == 'auto-70-150':
            main_builder.setup_event_args([
                'mg{0}-gc{0}-scene'.format(main_builder.uimod_id)
            ])

        main_conf = main_builder.build()
        self._main_ss = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._main_ss

        dgmod = self._dgmod_rc
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _RcDeleter(IDgConfDeleter):
    def __init__(self, uimod_cam: IUiConfProvider) -> None:
        self._dgmod_rc: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RAWCODEC, uimod_cam.id)

    def run(self) -> None:
        self._dgmod_rc.uninstall()


def rc_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_MV_CAMERA:
        return [_RcDeleter(target_uimod)]
    return []
