from typing import Final

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.dgmod.rn.rn_main import MainConfBuilder
from edge.webserver.sysconf.dgmod.rn.rn_setupstore import SetupStoreConfBuilder
from edge.webserver.sysconf.dgmod.rn.uiconf import IUiConfProvider2
from edge.webserver.sysconf.dgmod.rn.uiconf.fusion import FusionUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.gnss import GnssUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.redlight import RedlightUiConfProvider, RailroadUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.repro import ReproUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.roadway import RoadwayUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.spotspeed import SpotSpeedUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.streaming import HttpStreamingUiConfProvider, RtspStreamingUiConfProvider
from edge.webserver.sysconf.dgmod.rn.uiconf.webserver import WebServerUiConfProvider
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.terminal import TermAboutUiconf
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def rn_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_HTTP_STREAMING, UTYPE_RTSP_STREAMING}):
        return []

    try:
        uimod_streaming = target_uideps.find_any({UTYPE_HTTP_STREAMING, UTYPE_RTSP_STREAMING})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_term = target_uideps.find_any({UTYPE_TERMINAL})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_cam = target_uideps.try_find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
    uimod_gnss = target_uideps.try_find_any({UTYPE_GNSS})
    uimod_ra = target_uideps.try_find_any({UTYPE_RADAR})
    uimod_roadway = target_uideps.try_find_any({UTYPE_ROADWAY})
    uimod_ws = target_uideps.try_find_any({UTYPE_WEBSERVER})
    uimod_repro = target_uideps.try_find_any({UTYPE_REPROJECTIONS})

    uimods_spotspeed = target_uideps.try_find_all(UTYPE_DET_SPOT_SPEED)
    uimod_spotspeed = uimods_spotspeed[0] if len(uimods_spotspeed) == 1 else None

    uimods_railroad = target_uideps.try_find_all(UTYPE_DET_RAILROAD)
    uimods_redlight = target_uideps.try_find_all(UTYPE_DET_REDLIGHT)

    activate = forced or any([
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_railroad]),
        any([uiref_to_newconf_map.get(x.ref) for x in uimods_redlight]),
        uimod_cam and uiref_to_newconf_map.get(uimod_cam.ref),
        uimod_gnss and uiref_to_newconf_map.get(uimod_gnss.ref),
        uimod_repro and uiref_to_newconf_map.get(uimod_repro.ref),
        uimod_ra and uiref_to_newconf_map.get(uimod_ra.ref),
        uimod_roadway and uiref_to_newconf_map.get(uimod_roadway.ref),
        uimod_spotspeed and uiref_to_newconf_map.get(uimod_spotspeed.ref),
        uiref_to_newconf_map.get(uimod_streaming.ref),
        uiref_to_newconf_map.get(uimod_term.ref)
    ])

    if not activate:
        return []

    creator = _RnCreator(
        uimod_streaming, uimod_fusion, uimod_gnss, uimod_roadway, uimod_term, uimod_repro,
        uimods_redlight, uimods_railroad, uimod_cam,
        webserver_uimod=uimod_ws, radar_uimod=uimod_ra, spotspeed_uimod=uimod_spotspeed)

    return [creator]


class _RnCreator(IDgConfBuilder):
    def __init__(self,
                 streaming_uimod: IUiConfProvider,
                 fusion_uimod: IUiConfProvider,
                 gnss_uimod: IUiConfProvider,
                 roadway_uimod: IUiConfProvider,
                 terminal_uimod: IUiConfProvider,
                 repro_uimod: IUiConfProvider | None,
                 redlight_uimods: list[IUiConfProvider],
                 railroad_uimods: list[IUiConfProvider],
                 cam_uimod: IUiConfProvider | None = None,
                 webserver_uimod: IUiConfProvider | None = None,
                 radar_uimod: IUiConfProvider | None = None,
                 spotspeed_uimod: IUiConfProvider | None = None):

        super().__init__()

        self._dgmod_rn: Final = dg.DgMod(dg.EDGE_TYPE_RENDERER, streaming_uimod.id)

        self._uimod_camera: Final = cam_uimod
        self._uimod_fusion: Final = fusion_uimod
        self._uimod_gnss: Final = gnss_uimod
        self._uimod_repro: Final = repro_uimod
        self._uimod_radar: Final = radar_uimod
        self._uimod_roadway: Final = roadway_uimod
        self._uimod_spotspeed: Final = spotspeed_uimod
        self._uimod_streaming: Final = streaming_uimod
        self._uimod_terminal: Final = terminal_uimod
        self._uimod_webserver: Final = webserver_uimod

        self._uimods_railroad: Final = railroad_uimods
        self._uimods_redlight: Final = redlight_uimods

        self._main_ss = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_rn

    def setup(self, uiref_to_conf_map: UrefToConfMap) -> None:
        ss_builder = SetupStoreConfBuilder()
        main_builder = MainConfBuilder(self._uimod_streaming.id, ph.get_spool_dir(self._uimod_streaming.pid))
        main_builder.setup_service_origin(TermAboutUiconf.get_conf_path(self._uimod_streaming.pid))
        main_builder.setup_service_datetime(TimeUiconf.CONF_PATH)

        providers = list[IUiConfProvider2]()

        uimod = self._uimod_roadway
        providers.extend([
            FusionUiConfProvider(self._uimod_fusion.id),
            GnssUiConfProvider(self._uimod_gnss.id, self._uimod_gnss.conf_path),
            RoadwayUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
        ])

        if self._uimod_streaming.type == UTYPE_HTTP_STREAMING:
            uimod = self._uimod_streaming
            providers.extend([
                HttpStreamingUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf()),
                WebServerUiConfProvider('0')
            ])

        elif self._uimod_streaming.type == UTYPE_RTSP_STREAMING:
            uimod = self._uimod_streaming
            providers.extend([
                RtspStreamingUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        else:
            raise NotImplementedError()

        if self._uimod_radar:
            main_builder.setup_image_anno__ra_enabled(True)

        if self._uimod_repro:
            uimod = self._uimod_repro
            providers.extend([
                ReproUiConfProvider(uimod.id, uimod.conf_path)
            ])

        if self._uimod_spotspeed is not None:
            uimod = self._uimod_spotspeed
            providers.extend([
                SpotSpeedUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        if len(self._uimods_redlight) == 1:
            uimod = self._uimods_redlight[0]
            providers.extend([
                RedlightUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        elif len(self._uimods_railroad) == 1:
            uimod = self._uimods_railroad[0]
            providers.extend([
                RailroadUiConfProvider(uimod.id, uiref_to_conf_map.get(uimod.ref) or uimod.load_conf())
            ])

        for provider in providers:
            provider.setup_builders(main_builder, ss_builder)

        main_conf = main_builder.build()
        ss_active = ss_builder.build()

        self._main_ss = main_conf, ss_active

    def build(self) -> None:
        main_conf, ss_active = self._main_ss  # type: dict, bool

        dgmod = self._dgmod_rn
        dgmod.install()
        dgmod.save_main_conf(main_conf)
        dgmod.save_setupstore_conf(ss_active)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _RnDeleter(IDgConfDeleter):
    def __init__(self, uimod_streaming: IUiConfProvider) -> None:
        self._dgmod_rn: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_RENDERER, uimod_streaming.id)

    def run(self) -> None:
        self._dgmod_rn.uninstall()


def rn_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type in [UTYPE_HTTP_STREAMING, UTYPE_RTSP_STREAMING]:
        return [_RnDeleter(target_uimod)]

    return []
