from typing import Final

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.dgmod import dg
from edge.webserver.sysconf.dgmod.dgconf import IDgConfDeleter, IDgConfBuilder, UrefToConfMap, Udeps
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.uiconf.time import TimeUiconf
from edge.webserver.sysconf.uimod.utype import *


# noinspection PyUnusedLocal
def tr_create_dgconf_builder(
        target_uiref: str,
        target_uideps: Udeps,
        uiref_to_newconf_map: UrefToConfMap,
        forced: bool = False) -> list[IDgConfBuilder]:

    if not target_uideps.test_contains(target_uiref, {UTYPE_FUSION}):
        return []

    try:
        uimod_cam = target_uideps.find_any({UTYPE_MV_CAMERA, UTYPE_RTSP_CAMERA})
        uimod_fusion = target_uideps.find_any({UTYPE_FUSION})
        uimod_hstreaming = target_uideps.find_any({UTYPE_HTTP_STREAMING})
        uimod_lpr = target_uideps.find_any({UTYPE_LPR})
        uimod_roadway = target_uideps.find_any({UTYPE_ROADWAY})
        uimod_rules = target_uideps.find_any({UTYPE_TRAFFIC_RULES})
        uimod_term = target_uideps.find_any({UTYPE_TERMINAL})
        uimod_transit = target_uideps.find_any({UTYPE_DET_TRANSIT})
    except NotFoundError as x:
        eventlog.log_w(f'fail: link absent ({x})')
        return []

    uimod_repro = target_uideps.try_find_any({UTYPE_REPROJECTIONS})
    uimod_ra = target_uideps.try_find_any({UTYPE_RADAR})
    uimod_rstreaming = target_uideps.try_find_any({UTYPE_RTSP_STREAMING})

    activate = forced or any([
        uimod_ra and uiref_to_newconf_map.get(uimod_ra.ref),
        uimod_repro and uiref_to_newconf_map.get(uimod_repro.ref),
        uimod_rstreaming and uiref_to_newconf_map.get(uimod_rstreaming.ref),
        uiref_to_newconf_map.get(uimod_cam.ref),
        uiref_to_newconf_map.get(uimod_hstreaming.ref),
        uiref_to_newconf_map.get(uimod_roadway.ref),
        uiref_to_newconf_map.get(uimod_term.ref),
        uiref_to_newconf_map.get(uimod_transit.ref)
    ])

    if not activate:
        return []

    creator = _TrCreator(
        uimod_ra, uimod_cam, uimod_lpr, uimod_rstreaming, uimod_hstreaming,
        uimod_rules, uimod_transit, uimod_fusion, uimod_roadway, uimod_term, uimod_repro)

    return [creator]


class _TrCreator(IDgConfBuilder):
    def __init__(self, ra_uimod: IUiConfProvider | None,
                 cam_uimod: IUiConfProvider,
                 lpr_uimod: IUiConfProvider,
                 rstreaming_uimod: IUiConfProvider | None,
                 hstreaming_uimod: IUiConfProvider,
                 rules_uimod: IUiConfProvider,
                 transit_uimod: IUiConfProvider,
                 fusion_uimod: IUiConfProvider,
                 roadway_uimod: IUiConfProvider,
                 term_uimod: IUiConfProvider,
                 repro_uimod: IUiConfProvider | None) -> None:

        self._dgmod_tr: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_TRACKER, fusion_uimod.id)

        self._uimod_cam: Final[IUiConfProvider] = cam_uimod
        self._uimod_fusion: Final[IUiConfProvider] = fusion_uimod
        self._uimod_hstreaming: Final[IUiConfProvider] = hstreaming_uimod
        self._uimod_repro: Final[IUiConfProvider | None] = repro_uimod
        self._uimod_lpr: Final[IUiConfProvider] = lpr_uimod
        self._uimod_ra: Final[IUiConfProvider | None] = ra_uimod
        self._uimod_roadway: Final[IUiConfProvider] = roadway_uimod
        self._uimod_rstreaming: Final[IUiConfProvider | None] = rstreaming_uimod
        self._uimod_rules: Final[IUiConfProvider] = rules_uimod
        self._uimod_terminal: Final[IUiConfProvider] = term_uimod
        self._uimod_transit: Final[IUiConfProvider] = transit_uimod

        self._main_misc = None

    @property
    def dgmod(self) -> dg.DgMod | None:
        return self._dgmod_tr

    def setup(self, uiref_to_newconf_map: UrefToConfMap) -> None:
        cam_uiconf = uiref_to_newconf_map.get(self._uimod_cam.ref) or self._uimod_cam.load_conf()
        hstreaming_uiconf = uiref_to_newconf_map.get(self._uimod_hstreaming.ref) or self._uimod_hstreaming.load_conf()
        hstreaming_pub = hstreaming_uiconf['rw']
        transit_uiconf = uiref_to_newconf_map.get(self._uimod_transit.ref) or self._uimod_transit.load_conf()

        tr_dgmod = self._dgmod_tr
        spool_dir = ph.get_spool_dir(self._uimod_fusion.pid)
        if self._uimod_cam.type == UTYPE_MV_CAMERA:
            cam_pri = cam_uiconf['ro']
            if cam_pri.get('raw-format', 'MONO8T') == 'MONO8T':
                sourse_dir = fs.join_path(spool_dir, f'mg{self._uimod_cam.id}-tr{self._uimod_fusion.id}-scene')
            else:
                sourse_dir = fs.join_path(spool_dir, f'rc{self._uimod_cam.id}-tr{self._uimod_fusion.id}-scene')

        elif self._uimod_cam.type == UTYPE_RTSP_CAMERA:
            sourse_dir = fs.join_path(spool_dir, f'rg{self._uimod_cam.id}-tr{self._uimod_fusion.id}-scene')

        else:
            raise NotImplementedError

        plate_detected_mode = transit_uiconf['ro'].get('plate-detected-mode', False)

        terminal_uiconf = uiref_to_newconf_map.get(self._uimod_terminal.ref) or self._uimod_terminal.load_conf()

        road_conf_path = self._dgmod_tr.get_conf_path('road.conf')

        # create main.conf
        main_conf = {
            'services': {
                'road.conf': road_conf_path,
                'plate-info': '/opt/edge/share/lpr/plate-info.json',
                'time.conf': TimeUiconf.CONF_PATH,
                'stat': {
                    'tout': 5,
                    'path': fs.join_path(spool_dir, 'zabbix', 'tr0')
                },
                'scene-provider': {
                    'MgDir': sourse_dir,
                    'LrDir': fs.join_path(spool_dir, f'lr{self._uimod_lpr.id}-tr{self._uimod_fusion.id}-scene'),
                    'Tout': 750
                },
                'lr-track-provider': {
                    'mode': ['detected', 'lost'] if plate_detected_mode else ['lost'],
                    'endpoint': fs.join_path(spool_dir, f'lr{self._uimod_lpr.id}-tr{self._uimod_fusion.id}-track'),
                    'filter': terminal_uiconf['ro'].get('trackset-filter') or list(),
                    'lost-track-acc': {
                        'timeout': 3,
                        'store': {
                            'endpoint': fs.join_path(spool_dir, f'tr{self._uimod_fusion.id}-store'),
                            'max-file-count': 20  # max-file-count >> debounce-tout * lr_track_fps (or we'll lose some events)
                        }
                    }
                },
                'ra-track-provider': {
                    'filter': terminal_uiconf['ro']['ratrack-filter'],
                    'require-plate': transit_uiconf['ro'].get('require-plate', True)
                },
                'scene-emitter': [],
                'track-emitter': {
                    'endpoints': [
                        fs.join_path(spool_dir, f'tr{self._uimod_fusion.id}-sc{self._uimod_rules.id}-track')
                    ]
                }
            }
        }

        if self._uimod_repro:
            repro_uiconf = uiref_to_newconf_map.get(self._uimod_repro.ref) or self._uimod_repro.load_conf()
            repro_pri_speedcam_jnode = repro_uiconf['ro'].get('speedcam') or dict()
            main_conf['services']['cam'] = {
                'reprojections': self._uimod_repro.conf_path,
                'avg-plate-u': repro_pri_speedcam_jnode.get('avg-plate-u'),
                'log': repro_pri_speedcam_jnode.get('log') or False
            }

        # enable http streaming
        if hstreaming_pub['enabled']:
            # noinspection PyUnresolvedReferences
            main_conf['services']['scene-emitter'].append({
                'SkipRate': hstreaming_pub['live']['videostream']['skip-rate'],
                'Dir': fs.join_path(spool_dir, f'tr{self._uimod_fusion.id}-or{self._uimod_hstreaming.id}-scene')
            })

        # enable rtsp-streaming
        if self._uimod_rstreaming:
            rstreaming_uiconf = uiref_to_newconf_map.get(self._uimod_rstreaming.ref) or self._uimod_rstreaming.load_conf()
            rstreaming_pub = rstreaming_uiconf.get('rw')
            if rstreaming_pub['enabled']:
                # noinspection PyUnresolvedReferences
                main_conf['services']['scene-emitter'].append({
                    'SkipRate': rstreaming_uiconf['rw']['live']['videostream']['skip-rate'],
                    'Dir': fs.join_path(spool_dir, f'tr{self._uimod_fusion.id}-or{self._uimod_rstreaming.id}-scene')
                })

        # add ra-dir for speed pipelines
        if self._uimod_ra:
            ra_uiconf = uiref_to_newconf_map.get(self._uimod_ra.ref) or self._uimod_ra.load_conf()
            ra_cuboid = ra_uiconf['ro'].get('cuboid')
            if ra_cuboid is not None:
                main_conf['ra-cuboid'] = ra_cuboid

            main_conf['services']['scene-provider']['RaDir'] = fs.join_path(
                spool_dir, f'ra{self._uimod_ra.id}-tr{self._uimod_fusion.id}-scene')

        roadway_uiconf = uiref_to_newconf_map.get(self._uimod_roadway.ref) or self._uimod_roadway.load_conf()
        ui_road_jnode = roadway_uiconf['rw']['args']
        road_conf_jnode = {
            'road': {},
            'scene-wh': ui_road_jnode['scene-wh']
        }

        if len(ui_road_jnode['lanes']) > 0:
            road_conf_jnode['road']['lanes'] = ui_road_jnode['lanes']


        self._main_misc = main_conf, road_conf_jnode, road_conf_path

    def build(self) -> None:
        main_conf, road_conf_jnode, road_conf_path = self._main_misc  # type: dict, dict, str

        edge_tr = self._dgmod_tr
        edge_tr.install()
        edge_tr.save_json_conf(road_conf_path, road_conf_jnode)
        edge_tr.save_main_conf(main_conf)
        edge_tr.save_setupstore_conf(active=True)

    def notify(self) -> None:
        self.dgmod.sigkill()


class _TrDeleter(IDgConfDeleter):
    def __init__(self, uimod_fu: IUiConfProvider) -> None:
        self._dgmod_tr: Final[dg.DgMod] = dg.DgMod(dg.EDGE_TYPE_TRACKER, uimod_fu.id)

    def run(self) -> None:
        self._dgmod_tr.uninstall()


def tr_create_dgconf_deleter(target_uimod: IUiConfProvider) -> list[IDgConfDeleter]:
    if target_uimod.type == UTYPE_FUSION:
        return [_TrDeleter(target_uimod)]

    return []
