from typing import TypedDict, NotRequired, Self

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import CamServiceError

from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.uref import Uref


class CameraStats(TypedDict):
    stream: str
    snapshot: NotRequired[str]
    resolution: NotRequired[tuple[int, int]]


def get_terminal_vs0_stats(term_uref: str) -> CameraStats:
    term_umod: Umod = Umod.create(term_uref)
    term_uuid: str = TerminalUiconf(term_umod.conf_path).get_uuid()
    video_wh = CamDgStats.create(term_umod.pid).stream_wh
    jnode = {
        'stream': f'api/v2/origins/media/{term_uuid}/live/stream',
        'snapshot': f'api/v2/origins/media/{term_uuid}/live/snapshot'
    }

    if video_wh is not None:
        video_w, video_h = video_wh
        if video_w > 0 and video_h > 0:
            jnode['resolution'] = (video_w, video_h)

    return jnode


class CamDgStats:
    def __init__(self, term_uref: str, jnode_path: str) -> None:
        self._term_uref = term_uref
        self._jnode = fs.load_json3(jnode_path, default=None)

    @classmethod
    def create(cls, term_pid: str) -> Self:
        ui_id = '0'
        jnode_path = fs.join_path(f'/var/spool/edge/pipeline-{term_pid}/zabbix/vs{ui_id}')

        term_uref = Uref.create_term_uref(term_pid).ref
        return cls(term_uref, jnode_path)

    def _get_stream_stats(self) -> dict | None:
        streamer_stats: dict | None = None

        if self._jnode:
            units_jnode: list[dict] = self._jnode.get('units') or list()
            streamers = [x for x in units_jnode if x.get('type') == 'bus-streamer']

            for streamer in streamers:
                stat_jnode = streamer.get('stat') or dict()
                endpoint_count: int = stat_jnode.get('endpoints', 0)
                if endpoint_count:
                    streamer_stats = stat_jnode

        return streamer_stats

    def assert_stream_active(self) -> None:
        stream_stats: dict | None = self._get_stream_stats()
        if stream_stats is None:
            eventlog.log_w('fail: CamServiceError')
            raise CamServiceError

        stream_active = stream_stats.get('active', False)
        if not stream_active:
            eventlog.log_w('fail: CamServiceError')
            raise CamServiceError

    @property
    def stream_wh(self) -> tuple[int, int] | None:
        stream_stats: dict | None = self._get_stream_stats()
        if stream_stats:
            return stream_stats.get('wh')

    @property
    def frame_count(self) -> int | None:
        stream_stats: dict | None = self._get_stream_stats()
        if stream_stats:
            return stream_stats.get('frame-count')

    @property
    def temperature(self) -> float | None:
        if self._jnode:
            units_jnode: list[dict] = self._jnode.get('units') or list()
            devices = [x for x in units_jnode if x.get('type') == 'mediagrabber~device']

            for device in devices:
                stat_jnode = device.get('stat') or dict()
                return stat_jnode.get('temp')

        return None


    def get_stats(self) -> dict:
        return {
            'media': [
                get_terminal_vs0_stats(self._term_uref)
            ]
        }

class OrDgStats:
    def __init__(self, term_pid: str,  jnode_path: str) -> None:
        term_umod = Umod.create_term(term_pid)
        self._term_uref = term_umod.ref
        self._term_uuid = TerminalUiconf(term_umod.conf_path).get_uuid()
        self._jnode = fs.load_json3(jnode_path, default=None)

    @classmethod
    def create(cls, term_pid: str) -> Self:
        ui_id = '0'
        jnode_path = fs.join_path(f'/var/spool/edge/pipeline-{term_pid}/zabbix/or{ui_id}')
        return cls(term_pid, jnode_path)

    def _get_stream_stats(self) -> dict | None:
        streamer_stats: dict | None = None

        if self._jnode:
            streamer_stats = self._jnode.get('stream')

        return streamer_stats
    @property
    def stream_wh(self) -> tuple[int, int] | None:
        stream_stats: dict | None = self._get_stream_stats()
        if stream_stats:
            return stream_stats.get('wh')
