from collections.abc import Iterable

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import *
from edge.types.mmr import MMR_CLASSES
from edge.webserver.sysconf.dgstats.camera import get_terminal_vs0_stats
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.types.defs import *
from edge.webserver.sysconf.uiconf.builder import CTRL_DEF_UICONF_DIR, UconfBuilder
from edge.webserver.sysconf.uiconf.terminal import TerminalUiconf
from edge.webserver.sysconf.uimod import rootinfo
from edge.webserver.sysconf.uimod import unet
from edge.webserver.sysconf.uimod.eventstore import EsUimod
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.sysconf.uimod.umod2 import Umod2
from edge.webserver.sysconf.uimod.uref import Uref
from edge.webserver.sysconf.uimod.urole import *
from edge.webserver.sysconf.uimod.utype import *
from edge.webserver.sysconf.ulimits.detector import DetUlimits
from edge.webserver.sysconf.ulimits.eventstore import EsUlimits
from edge.webserver.sysconf.ulimits.roadway import RoadwayUlimits


def get_spool_dir(ui_pid: str) -> str:
    return fs.join_path(CTRL_SPOOL_DIR, f'pipeline-{ui_pid}')


def build_uconf(unit_utype: str, unit_uref: str, factory_name: str, factory_conf: dict | None = None) -> dict:
    factory_info = unet.decode_factory(fs.join_path(CTRL_DEF_UICONF_DIR, factory_name))
    uconf_builder = UconfBuilder(factory_info.defconf)
    return uconf_builder.build(unit_utype, unit_uref, factory_conf)


def build_term_unit_uconf(term_uref: str, unit_utype: str, unit_uref: str, factory_conf: dict | None = None) -> dict:
    term_umod = Umod.create(term_uref)
    term_uconf = TerminalUiconf(term_umod.conf_path)
    return build_uconf(unit_utype, unit_uref, term_uconf.factory_name, factory_conf)


def map_term_uuid_to_uref(term_uuid: str) -> str | None:
    root_uinfo = rootinfo.decode_info().root
    terms_uinfo: list[unet.UnitInfo] = unet.map_units_r(root_uinfo, utypes=[UTYPE_TERMINAL])

    for term_uinfo in terms_uinfo:
        next_term_umod = Umod.create(term_uinfo['ref'])
        next_term_uuid: str = TerminalUiconf(next_term_umod.conf_path).get_uuid()

        if next_term_uuid == term_uuid:
            return term_uinfo['ref']


def map_term_uuid_to_pid(term_uuid: str) -> str | None:
    term_uref = map_term_uuid_to_uref(term_uuid)
    if term_uref:
        return Umod.create(term_uref).pid


# noinspection PyClassHasNoInit
class RootInfoUiconf:
    @classmethod
    def get_term_event_types(cls, term_uinfo: unet.UnitInfo) -> list[str]:
        det_types = [
            UTYPE_DET_TRANSIT,
            UTYPE_DET_PARKING,
            UTYPE_DET_WRONG_WAY,
            UTYPE_DET_REDLIGHT,
            UTYPE_DET_WANTED,
            UTYPE_DET_SECTION_SPEED,
            UTYPE_DET_BUS_LANE,
            UTYPE_DET_SEATBELT
        ]

        if unet.find_unit_r(term_uinfo, utype=UTYPE_RADAR):
            det_types.append(UTYPE_DET_SPOT_SPEED)

        return det_types

    @classmethod
    def get_raterm_pid(cls) -> str | None:
        root_uinfo = rootinfo.decode_info().root
        radar_uinfo = unet.find_unit_r(root_uinfo, utype=UTYPE_RADAR)
        if radar_uinfo:
            return Umod.create(radar_uinfo['ref']).pid

    @classmethod
    def get_uimods(cls, root: unet.UnitInfo, urefs: Iterable[str]) -> set[IUiConfProvider]:
        uimods = set()
        for uref in urefs:
            uinfo = unet.find_unit_r(root, uref=uref)
            if uinfo is None:
                eventlog.log_w(f'fail: find_unit_r (uiref:{uref})')
                continue

            utype = uinfo['type']
            if utype == UTYPE_WEBSERVER:
                uimod = Umod2(uinfo['ref'])

            elif utype == UTYPE_EVENTSTORE:
                uimod = EsUimod.create(uinfo['ref'])

            elif utype in UTYPES_UIMOD2:
                uimod = Umod2(uinfo['ref'])

            else:
                uimod = Umod.create(uinfo['ref'])

            uimods.add(uimod)

        return uimods

    @classmethod
    def add_term(cls, ctrl_uref: str,
                 factory_uinfo: unet.UnitInfo,
                 factory_links: list[unet.UnetLink]) -> tuple[unet.UnitInfo, unet.UnitInfo]:
        """
        decode_term_factory(ctrl_uinfo, factory_name) -> tuple[term_uinfo, term_links, term_conf]
            - ctrl_uinfo is const
            - term_uinfo contains UnitInfos with remapped urefs
            - term_links contains UnitLinks with remmaped urefs
            - term_conf contains remmaped uref_to_defconf_map

        :returns: tuple[root_uinfo, term_uinfo]
            [0] contains the root_uinfo with term_uinfo added (with remapped urefs)
            [1] contains the term_uinfo (with remapped urefs)
        """

        unet_info = rootinfo.decode_info()

        # create a uref map
        uref_gen = rootinfo.UrefGenerator(unet_info.root)
        term_uref = uref_gen.create_term_uref().ref

        uref_gen = rootinfo.UrefGenerator(unet_info.root)
        old_to_new_factory_urefs: dict[str, str] = uref_gen.create_uref_map(factory_uinfo, term_uref)

        # remap urefs in the links
        for factory_link in factory_links:
            factory_link[0] = old_to_new_factory_urefs.get(factory_link[0], factory_link[0])
            factory_link[1] = old_to_new_factory_urefs.get(factory_link[1], factory_link[1])

        unet_info.links.extend(factory_links)

        # remap urefs in the infos
        def replace_uref_fn(uinfo: unet.UnitInfo) -> None:
            uinfo['ref'] = old_to_new_factory_urefs[uinfo['ref']]

        unet.map_units_r(factory_uinfo, replace_uref_fn)

        # update the root-info
        ctrl_uinfo = unet.find_unit_r(unet_info.root, uref=ctrl_uref)
        ctrl_uinfo['units'].append(factory_uinfo)

        def find_link(links_list: list[unet.UnetLink], ref1: str, ref2: str) -> bool:
            for link in links_list:
                if ref1 in link and ref2 in link:
                    return True

            return False

        tr_urefs = unet.find_urefs_r(unet_info.root, utypes=[UTYPE_TRAFFIC_RULES])
        iit_ca_urefs = unet.find_urefs_r(unet_info.root, utypes=[UTYPE_IIT_CA])
        iit_ca_eu_urefs = unet.find_urefs_r(unet_info.root, utypes=[UTYPE_IIT_CA_END_USER])

        for tr_uref in tr_urefs:
            for iit_ca_uref in iit_ca_urefs:
                if not find_link(unet_info.links, tr_uref, iit_ca_uref):
                    unet_info.links.append((tr_uref, iit_ca_uref))

            for iit_ca_eu_uref in iit_ca_eu_urefs:
                if not find_link(unet_info.links, tr_uref, iit_ca_eu_uref):
                    unet_info.links.append((tr_uref, iit_ca_eu_uref))

        rootinfo.encode_info(unet_info)

        return unet_info.root, factory_uinfo

    @classmethod
    def add_term_detector(cls, term_uref: str, det_utype: str, det_name: str) -> tuple[Uref, unet.UnitInfo]:
        root_unet = rootinfo.decode_info()
        root_info, root_links = root_unet.root, root_unet.links

        term_uinfo = unet.find_unit_r(root_info, uref=term_uref)
        if term_uinfo is None:
            eventlog.log_d(f'fail: NotFoundError ({UTYPE_TERMINAL})')
            raise NotFoundError

        rules_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_TRAFFIC_RULES)
        if rules_uinfo is None:
            eventlog.log_d(f'fail: NotFoundError ({UTYPE_TRAFFIC_RULES})')
            raise NotFoundError

        uref_gen = rootinfo.UrefGenerator(rules_uinfo)
        det_uref: Uref = uref_gen.create_unit_uref(term_uref, det_utype)

        rules_uinfo['units'].append({
            'ref': det_uref.ref,
            'type': det_uref.type,
            'name': det_name,
            'role': UROLE_DETECTOR,
            'internal': False
        })

        if det_utype == UTYPE_DET_SPOT_SPEED:
            hstreaming_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_HTTP_STREAMING)
            if hstreaming_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_HTTP_STREAMING})')
            else:
                root_links.append((det_uref.ref, hstreaming_uinfo['ref']))

            rstreaming_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_RTSP_STREAMING)
            if rstreaming_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_RTSP_STREAMING})')
            else:
                root_links.append((det_uref.ref, rstreaming_uinfo['ref']))

        elif det_utype in [UTYPE_DET_REDLIGHT, UTYPE_DET_RAILROAD]:
            hstreaming_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_HTTP_STREAMING)
            if hstreaming_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_HTTP_STREAMING})')
            else:
                root_links.append((det_uref.ref, hstreaming_uinfo['ref']))

            rstreaming_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_RTSP_STREAMING)
            if rstreaming_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_RTSP_STREAMING})')
            else:
                root_links.append((det_uref.ref, rstreaming_uinfo['ref']))

            traffic_light_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_TRAFFIC_LIGHT)
            if traffic_light_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_TRAFFIC_LIGHT})')
            else:
                root_links.append((det_uref.ref, traffic_light_uinfo['ref']))

            rtsp_cam_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_RTSP_CAMERA)
            if rtsp_cam_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_RTSP_CAMERA})')
            else:
                root_links.append((det_uref.ref, rtsp_cam_uinfo['ref']))

            mv_cam_uinfo = unet.find_unit_r(term_uinfo, utype=UTYPE_MV_CAMERA)
            if mv_cam_uinfo is None:
                eventlog.log_d(f'fail: NotFoundError ({UTYPE_MV_CAMERA})')
            else:
                root_links.append((det_uref.ref, mv_cam_uinfo['ref']))

        root_links.append((det_uref.ref, rules_uinfo['ref']))
        rootinfo.encode_info(root_unet)
        return det_uref, root_info

    @classmethod
    def add_term_service(cls,
                         term_uref: str,
                         service_utype: str,
                         service_name: str,
                         utypes_to_link: list[str]) -> tuple[Uref, unet.UnitInfo]:

        unet_info = rootinfo.decode_info()
        term_info = unet.find_unit_r(unet_info.root, uref=term_uref)
        if term_info is None:
            eventlog.log_d(f'fail: NotFoundError ({UTYPE_TERMINAL})')
            raise NotFoundError

        uref_gen = rootinfo.UrefGenerator(term_info)
        service_uref: Uref = uref_gen.create_unit_uref(term_uref, service_utype)

        if term_info.get('units') is None:
            term_info['units'] = []

        term_info['units'].append({
            'ref': service_uref.ref,
            'type': service_uref.type,
            'name': service_name,
            'role': UROLE_SERVICE,
            'internal': False
        })

        for type_to_link in utypes_to_link:
            unit_to_link_info = unet.find_unit_r(term_info, utype=type_to_link)
            unet_info.links.append((service_uref.ref, unit_to_link_info['ref']))

        rootinfo.encode_info(unet_info)
        return service_uref, unet_info.root


def get_detector_limits_stats(
        uref: str,
        vehicle: bool = True,
        scene: bool = True,
        lanes: bool = False,
        classes: bool = False) -> dict:

    root_uinfo = rootinfo.decode_info().root

    term_uinfo = unet.find_parent_r(root_uinfo, unit_uref=uref, parent_utype=UTYPE_TERMINAL)
    term_uref = term_uinfo['ref']

    vs0_stats = get_terminal_vs0_stats(term_uref)
    max_image_w = vs0_stats.get('resolution', [65535, 65535])[0]
    det_ulimits = DetUlimits(max_image_w)

    limits_jnode = {
        'stores': EsUlimits(root_uinfo).store_infos
    }

    limits_jnode.update(det_ulimits.get_image_profiles(vehicle, scene))

    if lanes:
        limits_jnode['lanes'] = RoadwayUlimits(root_uinfo, term_uref).lane_names

    if classes:
        limits_jnode['classes'] = MMR_CLASSES

    return {
        'limits': limits_jnode,
        'stats': {
            'media': [vs0_stats]
        }
    }
