from typing import Self

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import NotFoundError, InternalError
from edge.webserver.sysconf.types.conf import IUiConfProvider
from edge.webserver.sysconf.types.defs import CTRL_BACKUP_DIR, CTRL_UICONF_DIR, SEMVER_PATH
from edge.webserver.sysconf.uiconf.uiconf2 import UiConf2
from edge.webserver.sysconf.uimod.uref import Uref2
from edge.webserver.sysconf.uimod.utype import UTYPES_FILELESS


class Umod2(Uref2, IUiConfProvider):
    def __init__(self, ref: str) -> None:
        super().__init__(ref)
        self._conf_jnode: dict | None = None
        self._container_path: str | None = None

    def __eq__(self, other: Self) -> bool:
        return isinstance(other, self.__class__) and getattr(other, 'ref', None) == self.ref

    def __hash__(self) -> int:
        return hash(self.ref)

    def __str__(self) -> str:
        return self.ref

    @property
    def pid(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def id(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def lock_path(self) -> str:
        if self.type in UTYPES_FILELESS:
            # TODO: migrate to /run/lock/edge/
            return fs.join_path('/var/run/edge', f'{self.ref}.lock')

        return self.uiconf.lock_path

    def create_backup_dirpath(self) -> str:
        import time
        backup_action = '{:d}-{}'.format(int(round(time.time() * 1e6)), self.ref)
        backup_dir = fs.join_path(CTRL_BACKUP_DIR, backup_action)
        return backup_dir

    def assert_ref(self) -> None:
        """
        :raises NotFoundError:
        """
        self.load_conf()

    @property
    def parent(self) -> Self:
        """
        :raises NotFoundError:
        """
        return Umod2(self.parent_ref)

    def child(self, child_type: str, child_chan: str | int) -> Self:
        return Umod2(self.child_ref(child_type, child_chan))

    def load_conf(self) -> dict:
        """
        :raise NotFoundError:
        """
        if None in [self._container_path, self._conf_jnode]:
            container_names_desc = []
            segments = self.ref.split('~')
            for index in range(0, len(segments)):
                container_name = '~'.join(segments[0:index + 1]) + '.json'
                container_names_desc.append(container_name)

            for container_name in reversed(container_names_desc):
                container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
                container_jnode = fs.load_json3(container_path, default=None) or dict()
                conf_jnode = container_jnode.get(self.ref)
                if conf_jnode:
                    self._container_path = container_path
                    self._conf_jnode = conf_jnode
                    return conf_jnode

            raise NotFoundError

        return self._conf_jnode

    @property
    def schema_path(self) -> str:
        eventlog.log_e('fail: unsupported')
        raise NotImplementedError()

    @property
    def conf_path(self) -> str:
        # import warnings
        # warnings.warn('message', PendingDeprecationWarning, stacklevel=2)

        self.assert_ref()
        return self._container_path

    def save_conf(self, conf_jnode: dict) -> None:
        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        container_jnode[self.ref] = conf_jnode
        fs.write_json_file(container_jnode, self._container_path, pretty=True)

        self._conf_jnode = conf_jnode

    @property
    def uiconf(self) -> UiConf2:
        """
        :raise NotFoundError:
        """
        return UiConf2(self.load_conf())

    def save_uiconf(self, conf: UiConf2) -> None:
        """
        :rtype: None
        """
        with open(SEMVER_PATH, 'r') as f:
            ver = f.readlines()[0].strip()

        conf.set_ver(ver)
        self.save_conf(conf.raw)

    def install(self, conf: UiConf2, container_ref: str | None = None) -> None:
        if self.type in UTYPES_FILELESS:
            return

        try:
            self.assert_ref()
            eventlog.log_e('fail: InternalError (already installed)')
            raise InternalError

        except NotFoundError:
            pass

        if not self.ref.startswith(container_ref):
            raise NotImplementedError

        container_name = f'{container_ref or self.ref}.json'
        container_path = fs.join_path(CTRL_UICONF_DIR, container_name)
        container_jnode = fs.load_json3(container_path, default=None) or dict()
        container_jnode[self.ref] = conf.raw
        fs.write_json_file(container_jnode, container_path, pretty=True)

        self._container_path = container_path
        self._conf_jnode = conf.raw

    def uninstall(self) -> None:
        if self.type in UTYPES_FILELESS:
            return

        self.assert_ref()
        container_jnode = fs.load_json3(self._container_path, default=None) or dict()
        del container_jnode[self.ref]
        if len(container_jnode) > 0:
            fs.write_json_file(container_jnode, self._container_path, pretty=True)
        else:
            fs.silent_rm_r(self._container_path)
