from typing import Any, Annotated, Callable, Final, NotRequired, Self, TypedDict

from edge import eventlog
from edge.support import fs
from edge.types.exceptions import InternalError

UnetLink = Annotated[list[str], 2]


UnitInfo = TypedDict('UnitInfo', {
    'ref': str,
    'type': str,
    'role': str,
    'name': Annotated[str, NotRequired],
    'units': Annotated[list[Self], NotRequired],
    'internal': Annotated[bool, NotRequired]
})


class UnetInfo:
    def __init__(self, root: UnitInfo, links: list[UnetLink], ver: str | None = None) -> None:
        self.root: UnitInfo = root
        self.links: list[UnetLink] = links
        self.ver: str | None = ver


class UnetFactoryInfo(UnetInfo):
    def __init__(self,
                 root: UnitInfo,
                 links: list[UnetLink],
                 defconf: dict[str, str | dict],
                 ver: str | None = None) -> None:

        super().__init__(root, links, ver)
        self.defconf: Final[dict[str, str | dict]] = defconf


def decode_factory(file_path: str) -> UnetFactoryInfo:
    uiconf = fs.load_json3(file_path, default=None)
    if uiconf is None:
        eventlog.log_e(f'fail: filepath:<{file_path}>')
        raise InternalError

    root: UnitInfo = uiconf['root']
    links: list[UnetLink] = uiconf['links']
    defconf: dict[str, str | dict] = uiconf['defconf']
    return UnetFactoryInfo(root, links, defconf)


def map_units_r(root: UnitInfo,
                map_fn: Callable[[UnitInfo], Any] | None = None,
                utypes: list[str] | None = None) -> list[Any]:
    """
    Finds and applies a function to all units which satisfy the constraints.

    The root unit is included in the result if it satisfies the constraints.

    :param root:
        The root unit for the operation.
    :param map_fn:
        The function to be applied to each matched unit.
        The default function returns the unit info.
    :param utypes:
        Unit type constraint. All units are matched if not specified.
    """

    def find_units_r(parent: UnitInfo, result: list[UnitInfo]) -> list[UnitInfo]:
        children = parent.get('units', [])
        for child in children:
            find_units_r(child, result)

        if utypes is None:
            result.append(parent)

        elif parent['type'] in utypes:
            result.append(parent)

        return result

    subunits_info = find_units_r(root, [])
    return list(map(map_fn, subunits_info)) if map_fn else subunits_info


def find_urefs_r(root: UnitInfo, utypes: list[str] | None = None) -> list[str]:
    """
    Finds all unit refs which satisfy the constraints.

    The root unit is included in the result if it satisfies the constraints.

    :param root:
        The root unit for the operation.
    :param utypes:
        Unit type constraint. All units are matched if not specified.
    """
    return map_units_r(root, lambda x: x['ref'], utypes)


def find_unit_r(root: UnitInfo, uref: str | None = None, utype: str | None = None) -> UnitInfo | None:
    assert any([uref, utype])

    if root['ref'] == uref or root['type'] == utype:
        return root

    children = root.get('units', [])
    for child in children:
        target_unit: UnitInfo | None = find_unit_r(child, uref=uref, utype=utype)
        if target_unit:
            return target_unit


def find_unit_parent_r(
        root: UnitInfo,
        unit_uref: str | None = None,
        unit_utype: str | None = None,
        parent_utype: str | None = None
        ) -> tuple[UnitInfo, UnitInfo] | tuple[UnitInfo, UnitInfo | None] | tuple[None, None]:
    """
    Finds a unit and one of its parents in the root_info tree.

    At least one of the arguments must be specified: unit-type, unit-ref

    :param root:
        The root unit for the operation.
    :param unit_uref:
        Unit ref constraint.
    :param unit_utype:
        Unit type constraint.
    :param parent_utype:
        Parent unit type constraint.
        If not provided, the most nested parent will be returned.
    """
    assert any([unit_uref, unit_utype])

    def find_unit_parents_r(unit: UnitInfo, parents: list[UnitInfo]) -> tuple[UnitInfo | None, list[UnitInfo]]:
        """
        Finds a unit and all its parents.
        :param unit:
            Current unit info
        :param parents:
            List of the parent units.
            The order of parents is from the most nested to the least nested.
        """
        if unit['ref'] == unit_uref or unit['type'] == unit_utype:
            return unit, parents

        subunits = unit.get('units', [])
        for subunit in subunits:
            found_unit, _ = find_unit_parents_r(subunit, parents)
            if found_unit:
                parents.append(unit)
                return found_unit, parents

        return None, parents

    unit_info, parents_info = find_unit_parents_r(root, [])
    for parent in parents_info:
        if parent_utype is None:
            return unit_info, parent

        elif parent_utype == parent['type']:
            return unit_info, parent

    return unit_info, None


def find_parent_r(
        root: UnitInfo,
        unit_uref: str | None = None,
        unit_utype: str | None = None,
        parent_utype: str | None = None
        ) -> UnitInfo | None:
    _, parent = find_unit_parent_r(root, unit_uref, unit_utype, parent_utype)
    return parent
