from typing import Final, Self

from edge import eventlog
from edge.types.exceptions import NotFoundError
from edge.webserver.sysconf.uimod.utype import *


class Uref:
    def __init__(self, pid: str | int, utype: str, uid: str | int) -> None:
        """
        (pid:1, utype:'mv-camera', uid:'14') => 'mv-camera.1_14'

        The uid part defines the --id value of the edge module.
            'mv-camera.1_14' => /etc/edge/mediagrabber.14/
        """
        self._ref: Final[str] = f'{utype}.{pid}_{uid}'

        self._pid: Final[str] = str(pid)
        self._utype: Final[str] = utype
        self._uid: Final[str] = str(uid)

    def __eq__(self, other: Self) -> bool:
        return (isinstance(other, self.__class__) and
                getattr(other, 'pid', None) == self.pid and
                getattr(other, 'type', None) == self.type and
                getattr(other, 'id', None) == self.id)

    def __hash__(self) -> int:
        return hash(self.type + self.id)

    def __str__(self) -> str:
        return self.ref

    @classmethod
    def create_uref(cls, uref: str) -> Self:
        """
        :param uref:
            'mv-camera.0_4'
            'mv-camera.1_14'
            'mv-camera.2_24'
        """
        try:
            utype, pid_uid = uref.split('.', 1)
            pid, uid = pid_uid.split('_', 1)
        except Exception:
            eventlog.log_d(f'fail: NotFoundError ({uref})')
            raise NotFoundError

        return cls(pid, utype, uid)

    @classmethod
    def create_term_uref(cls, pid: str | int) -> Self:
        """
        terminal.0_0
        terminal.1_10
        terminal.2_20
        """
        return cls(pid, UTYPE_TERMINAL, 10 * int(pid))

    @classmethod
    def create_unit_uref(cls, pid: str | int, unit_utype: str, unit_uid: int) -> Self:
        """
        mv-camera.0_4
        mv-camera.1_14
        mv-camera.2_24
        """
        return cls(pid, unit_utype, 10 * int(pid) + unit_uid)

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def type(self) -> str:
        return self._utype

    @property
    def pid(self) -> str:
        return self._pid

    @property
    def id(self) -> str:
        return self._uid

    def assert_type(self, utype: str) -> None:
        if self.type != utype:
            eventlog.log_d(f'fail: NotFoundError ({self.ref})')
            raise NotFoundError


class Uref2:
    def __init__(self, ref: str) -> None:
        assert ref

        self._ref: Final[str] = ref

    def __eq__(self, other: Self) -> bool:
        return isinstance(other, self.__class__) and getattr(other, 'ref', None) == self.ref

    def __hash__(self) -> int:
        return hash(self.ref)

    def __str__(self) -> str:
        return self.ref

    @classmethod
    def create_uref(cls, root_type: str, root_chan: str | int) -> Self:
        return cls(f'{root_type}.{root_chan}')

    @property
    def ref(self) -> str:
        return self._ref

    @property
    def type(self) -> str:
        segments = self.ref.split('~')
        subtypes = [x.split('.')[0] for x in segments]
        return '~'.join(subtypes)

    @property
    def chan(self) -> str:
        segments = self.ref.split('~')
        return segments[-1].split('.')[1]

    def assert_type(self, other_type: str) -> None:
        """
        :raises NotFoundError:
        """
        if self.type != other_type:
            eventlog.log_d(f'fail: NotFoundError ({self.ref})')
            raise NotFoundError

    @property
    def root_ref(self) -> str:
        """
        :raises NotFoundError:
        """
        return self.ref.split('~', 1)[0]

    @property
    def parent_ref(self) -> str:
        """
        :raises NotFoundError:
        """
        parent_ref = '~'.join(self.ref.split('~')[:-1])
        if not parent_ref:
            raise NotFoundError

        return parent_ref

    def child_ref(self, child_type: str, child_chan: str | int) -> str:
        if not child_type.startswith(self.type):
            raise NotFoundError

        child_type_segments = child_type.split('~')
        if not len(self.type.split('~')) == len(child_type_segments) - 1:
            raise NotFoundError

        return f'{self.ref}~{child_type_segments[-1]}.{child_chan}'
