from edge import eventlog
from edge.services import db_wanted
from edge.support import db as edge_db
from edge.support.isotime import datetime_to_isostr, isostr_to_datetime_utc


def _get_limits() -> dict:
    return {
        'profiles': {
            'max': 2,
        },
        'lists': {
            'max': 2,
            'min-owner': 1,
            'max-owner': 128,
            'min-name': 1,
            'max-name': 128,
            'max-log': 4
        }
    }


class WantedError(Exception):
    pass


class WatchList:
    def __init__(self, services, ref: tuple[str, str]):
        """
        :type services: Services
        :type ref: (str, str)
        """
        self._services = services
        self.ref = ref


class WantedProfile:
    def __init__(self, services, ref: str):
        """
        :type services: Services
        :type ref: str
        """
        self._services = services
        self.ref = ref

    def get(self):
        self._services.db.destroy()


class WantedPack:
    def __init__(self, db_service: db_wanted.DbWantedService):
        self._db = db_service

    def destroy(self) -> None:
        self._db.destroy()

    def get(self) -> dict | None:
        with self._db as api:
            profile_infos: list[db_wanted.ProfileInfo] = api.load_profile_infos()
            list_infos: list[db_wanted.ListInfo] = api.load_list_infos()

            def pack_list(list_info: db_wanted.ListInfo) -> dict:
                list_versions = api.load_list_versions(list_info.owner, list_info.ident)

                list_jnode = {
                    'owner': list_info.owner,
                    'name': list_info.ident,
                    'log': [{
                        'time': datetime_to_isostr(x.utime, True),
                        'comment': x.comment
                    } for x in list_versions]
                }

                return list_jnode

            jnode = {
                'limits': _get_limits(),
                'profiles': [{'name': x.name} for x in profile_infos],
                'lists': list(map(pack_list, list_infos))
            }

            return jnode

    def delete_watch_list(self, owner: str, ident: str) -> dict:
        with self._db as db_service:
            rows_deleted = db_service.drop_list(owner, ident)

        if rows_deleted == 0:
            raise WantedError

        return self.get()

    def load_profile(self, name: str) -> dict:
        with self._db as api:
            profile: db_wanted.Profile = api.load_profile(name)

        if profile is None:
            raise WantedError

        jnode = {
            'args': {
                'name': profile.name,
                'template': profile.tmpl
            }
        }

        if len(profile.desc):
            jnode['args']['comment'] = profile.desc

        return jnode

    def save_profile(self, name: str, tmpl: dict) -> dict:
        with self._db as api:
            api.save_profile(name, '', tmpl)

        return {
            'args': {
                'name': name,
                'template': tmpl
            }
        }

    def load_watch_list(
            self,
            list_owner: str,
            list_name: str,
            list_path: str,
            prof_tmpl: str,
            prof_name: str | None,
            comment: str | None,
            prof_same: str,
            prof_new: str,
            prof_old: str):
        with self._db as api:
            try:
                api.save_list(
                    list_owner=list_owner,
                    list_name=list_name,
                    list_file=list_path,
                    list_sha1=None,

                    prof_tmpl=prof_tmpl,
                    prof_name=prof_name,
                    list_desc=comment,
                    prof_old=prof_old,
                    prof_new=prof_new,
                    prof_same=prof_same)
            except edge_db.DbNotFoundError as x:
                eventlog.log_i(f'fail: {x.hint}')
                raise WantedError

    def drop_profile(self, name: str) -> dict:
        with self._db as db_service:
            try:
                db_service.drop_profile(name)
            except edge_db.DbNotFoundError as x:
                eventlog.log_i(f'fail: {x.hint}')
                raise WantedError

            except edge_db.DbQueryError as x:
                eventlog.log_i(f'fail: {x.hint}')
                raise NotImplementedError

        return self.get()

    def cleanup_list(self, owner: str, name: str, max_time: str):
        max_time_dt = isostr_to_datetime_utc(max_time)
        with self._db as api:
            list_versions = api.load_list_versions(owner, name)
            revision = None
            for index in range(len(list_versions)-1, -1, -1):
                version = list_versions[index]
                if version.utime <= max_time_dt:
                    revision = version.revision
                else:
                    break

            if revision:
                api.cleanup_list(owner, name, ver_max_rev=revision)
            else:
                raise WantedError

    def revert_list(self, owner: str, ident: str, min_time: str):
        min_time_dt = isostr_to_datetime_utc(min_time)
        with self._db as api:
            timeline = api.load_list_versions(owner, ident)
            revision = None
            for index in range(0, len(timeline)):
                item = timeline[index]
                if item.utime <= min_time_dt:
                    revision = item.revision
                    break

            if revision:
                api.revert_list(owner, ident, ver_min_rev=revision)
            else:
                raise WantedError


def init(db_args: edge_db.DbArgs | None = None) -> WantedPack:
    if db_args is None:
        db_args = edge_db.DbArgs.deserialize2()

    db_service = db_wanted.init_service(db_args)

    return WantedPack(db_service)


# noinspection PyUnusedLocal
def fini(watchlist: WantedPack):
    watchlist.destroy()
