from gpsdclient import GPSDClient

from edge import eventlog
from edge.support import cgi
from edge.webserver.systest.misc import Services
from edge.webserver.systest.test import Test, TestStep, Summary, TestArgsError


def _get_geodata() -> tuple[str, str] | None:
    try_count = 0

    with GPSDClient(host='localhost', port='2947') as client:
        for result in client.dict_stream(convert_datetime=False, filter=['DEVICES', 'TPV']):
            msg_class = result.get('class')
            if msg_class == "DEVICES":
                if len(result.get('devices', [])) == 0:
                    eventlog.log_i("fail: gps absent")
                    return None
            elif msg_class == "TPV":
                if result.get('mode') in [2, 3]:
                    latitude = result.get('lat')
                    longitude = result.get('lon')

                    eventlog.log_i(f"{latitude} {longitude}")
                    return latitude, longitude

                try_count += 1
                if try_count > 10:
                    eventlog.log_i('fail: no fix')
                    return None


class GeolocationTestStep(TestStep):
    def __init__(self, services: Services, session_ref: str, test_ref: tuple[str, str], step_ref: str):
        super().__init__(services, session_ref, test_ref, step_ref)

    def _edit_log_item_jnode(self, log_item_jnode: dict, body_jnode: dict) -> dict:
        # unpack #/data/report
        report = body_jnode.get('report', False)
        if type(report) not in (bool,):
            eventlog.log_i('fail: http400. param #/report has invalid value')
            raise cgi.CgiError400

        try:
            data_jnode = body_jnode['data']
            lat_jnode = data_jnode['latitude']
            long_jnode = data_jnode['longitude']

            # unpack #/data/latitude/actual
            if lat_jnode.get('actual'):
                eventlog.log_i('fail: http400. param #/data/latitude/actual is provided')
                raise cgi.CgiError400

            # unpack #/data/latitude/expected
            expected_lat = lat_jnode.get('expected')
            if expected_lat is not None:
                if type(expected_lat) not in [float, int]:
                    eventlog.log_i('fail: http400. param #/data/latitude/expected has invalid value')
                    raise cgi.CgiError400

            # unpack #/data/longitude/actual
            if long_jnode.get('actual'):
                eventlog.log_i('fail: http400. param #/data/longitude/actual is provided')
                raise cgi.CgiError400

            # unpack #/data/longitude/expected
            expected_long = long_jnode.get('expected')
            if expected_long is not None:
                if type(expected_long) not in [float, int]:
                    eventlog.log_i('fail: http400. param #/data/longitude/expected has invalid value')
                    raise cgi.CgiError400

            # unpack #/summary
            summary_jnode = Summary.repack(body_jnode.get('summary'))

            # apply changes
            log_item_jnode['report'] = report
            log_item_jnode['data']['latitude']['expected'] = expected_lat
            log_item_jnode['data']['longitude']['expected'] = expected_long
            log_item_jnode['summary'] = summary_jnode
            return log_item_jnode

        except (KeyError, ValueError):
            eventlog.log_i(f'fail: http400 <x>')
            raise cgi.CgiError400


class GeolocationTest(Test):
    def __init__(self, services: Services, session_ref: str, test_ref: tuple[str, str]):
        super().__init__(services, session_ref, test_ref)

    def _mk_step_jnode(self, test_jnode: dict, step_jnode: dict, args_jnode: dict) -> dict | None:
        if args_jnode:
            eventlog.log_i('fail: http400')
            raise TestArgsError

        result = _get_geodata()
        if result is None:
            return None

        latitude, longitude = result
        step_jnode['data'] = {
            'latitude': {
                'actual': latitude
            },
            'longitude': {
                'actual': longitude
            }
        }

        return step_jnode
