__all__ = [
    'TestStep', 'Test',
    'Summary', 'StatSummary',
    'TestError', 'TestArgsError', 'TestLimitsError'
]

import os

from typing import Self

from edge import eventlog
from edge.webserver.sysconf import ph
from edge.webserver.sysconf.api_terminal import TerminalApi
from edge.webserver.sysconf.uimod.umod import Umod
from edge.webserver.systest.exception import *
from edge.webserver.systest.misc import *

_TEST_TYPE_GEOLOCATION = 'geolocation'
_TEST_TYPE_DATE_TIME = 'datetime'
_TEST_TYPE_SPOT_SPEED = 'spot-speed'
_TEST_TYPE_POSITION = 'position'


class Summary:
    def __init__(self, comment: str = ''):
        self.comment = comment

    @classmethod
    def repack(cls, jnode: dict) -> dict:
        """
        :raise ValueError:
        """
        summary = cls.unpack(jnode)
        return summary.pack()

    @classmethod
    def unpack(cls, jnode: dict | None) -> Self:
        """
        :raise ValueError:
        """
        if not jnode:
            return Summary()

        comment = jnode.get('comment', '')
        if type(comment) not in [str]:
            eventlog.log_i('fail: param #/comment has invalid type')
            raise ValueError

        elif not 0 <= len(comment) <= 512:
            eventlog.log_i('fail: param #/comment has invalid len')
            raise ValueError

        return cls(comment)

    def pack(self) -> dict:
        answer = dict()
        if len(self.comment):
            answer = {
                'comment': self.comment
            }

        return answer


class StatSummary(Summary):
    def __init__(self, status: str, comment: str = ''):
        super().__init__(comment)
        self.status = status

    @classmethod
    def unpack(cls, jnode: dict) -> Self:
        """
        :raise ValueError:
        """
        summary = Summary.unpack(jnode)

        # unpack #/status
        status = jnode.get('status')
        if status not in ['passed', 'failed']:
            eventlog.log_i('fail: param #/status has invalid value')
            raise ValueError

        return cls(status, summary.comment)

    def pack(self) -> dict:
        answer = Summary.pack(self)
        answer['status'] = self.status
        return answer


class TestStep:
    def __init__(self, services: Services, session_ref: str, test_ref: tuple[str, str], step_ref: str):
        self._services = services

        self._session_ref = session_ref
        self._test_ref = test_ref
        self._step_ref = step_ref

    def _edit_log_item_jnode(self, log_item_jnode: dict, body_jnode: dict) -> dict:
        log_item_jnode['data'] = body_jnode['data']
        log_item_jnode['summary'] = body_jnode.get('summary')
        return log_item_jnode

    def edit(self, jnode: dict) -> dict:
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                eventlog.log_e('fail: invalid test-ref')
                raise NotImplementedError

            for log_item_jnode in test_jnode['rw']['log']:
                if log_item_jnode['item'] == self._step_ref:
                    self._edit_log_item_jnode(log_item_jnode, jnode)
                    break

            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return test_jnode['rw']


def _mk_step(services, session_ref: str, test_ref: tuple[str, str], step_ref: str) -> TestStep:
    test_type = test_ref[0]
    if test_type == _TEST_TYPE_GEOLOCATION:
        from .geolocation_test import GeolocationTestStep
        return GeolocationTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_DATE_TIME:
        from .datetime_test import DatetimeTestStep
        return DatetimeTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_SPOT_SPEED:
        from .spot_speed_test import SpotSpeedTestStep
        return SpotSpeedTestStep(services, session_ref, test_ref, step_ref)

    elif test_type == _TEST_TYPE_POSITION:
        from .position_test import PositionTestStep
        return PositionTestStep(services, session_ref, test_ref, step_ref)

    else:
        return _mk_step(services, session_ref, test_ref, step_ref)


class Test:
    _MAX_LOG_ITEMS = 10

    def __init__(self, services: Services, session_ref: str, test_ref: tuple[str, str]):
        self._services = services

        self._session_ref = session_ref
        self._test_ref = test_ref

    @property
    def ref(self) -> tuple[str, str]:
        """
        :returns:
            test-type, test-run
        """
        return self._test_ref

    @staticmethod
    def _get_live_image(ui_pid: str, bs_chan: str = '0') -> tuple[int, str]:
        # get last overlay imagepath
        or_id = str(int(ui_pid) * 10)
        or_store = f"/var/spool/edge/pipeline-{ui_pid}/or{or_id}-or{or_id}-store/"
        file_names = os.listdir(or_store)

        def unpack_ts(file_name: str) -> int:
            return int(file_name.split('-')[-1])

        blob_name = max(file_names, key=unpack_ts)
        blob_path = os.path.join(or_store, blob_name)
        blob_ts = unpack_ts(blob_name)

        # save image to bs.65535
        from edge.support import bs
        bs_session = bs.Session.open('systest')
        try:
            blob_id = bs_session.save_blob2(bs_chan, blob_ts // 1000, blob_path)
            blob_ref = f'{bs_chan}_{blob_id}.jpeg'
            return blob_ts, blob_ref

        finally:
            bs_session.close()

    def _mk_step_jnode(self, test_jnode: dict, step_jnode: dict, args_jnode: dict | None) -> dict:
        return step_jnode

    def set_steps_report(self, report: bool) -> bool:
        with self._services.db as db_service:
            test_jnode = db_service.get_value(DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            for step_jnode in test_jnode['rw']['log']:
                step_jnode['report'] = report

            db_service.set_value(DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref), test_jnode)

        return True

    def add_step(self, jnode: dict) -> None:
        """
        :raise TestError
        """
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                raise TestError

            if len(test_jnode['rw']['log']) >= self._MAX_LOG_ITEMS:
                eventlog.log_i('fail: http409. item limit exceeded')
                raise TestLimitsError

            data_jnode = jnode.get('data')
            args_jnode = jnode.get('args')
            step_ref = mk_ref([x['item'] for x in test_jnode['rw']['log']])
            if data_jnode:
                blob_ts = data_jnode.get('ts')
                if blob_ts is not None:
                    time_isostr = self._services.tz.map_ts_usec_to_isostr(blob_ts)
                else:
                    time_isostr = self._services.tz.now_isostr()

                step_jnode = {
                    'report': True,
                    'item': step_ref,
                    'time': time_isostr,
                    'data': data_jnode
                }

            else:
                step_jnode = self._mk_step_jnode(test_jnode, {
                    'report': True,
                    'item': step_ref,
                    'time': self._services.tz.now_isostr()
                }, args_jnode)

            if step_jnode:
                test_jnode['rw']['log'].append(step_jnode)
                db_service.set_value(
                    DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                    test_jnode)

    def get_step(self, ref: str) -> TestStep | None:
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return None

            log_items = [log_item_jnode for log_item_jnode in test_jnode['rw']['log'] if log_item_jnode['item'] == ref]

            if len(log_items) != 1:
                return None

            return _mk_step(self._services, self._session_ref, self.ref, ref)

    def del_step(self, step_ref: str) -> bool:
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            last_items = test_jnode['rw']['log']
            next_items = [log_item_jnode for log_item_jnode in last_items if log_item_jnode['item'] != step_ref]

            if len(last_items) == len(next_items):
                return False

            test_jnode['rw']['log'] = next_items
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return True

    def del_steps(self, step_refs: list[str]) -> bool:
        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))
            if not test_jnode:
                return False

            if step_refs:
                last_items = test_jnode['rw']['log']
                next_items = [x for x in last_items if x['item'] not in step_refs]

            else:
                next_items = []

            test_jnode['rw']['log'] = next_items
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return True

    def set_summary(self, jnode: dict | None) -> None:
        summary_jnode = None
        if jnode:
            summary_jnode = StatSummary.repack(jnode['summary'])

        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))

            test_jnode['rw']['summary'] = summary_jnode
            db_service.set_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref),
                test_jnode)

            return test_jnode['rw']

    def pack(self) -> dict:
        from edge.webserver.sysconf.dgstats.camera import CamDgStats

        with self._services.db as db_service:
            test_jnode = db_service.get_value(
                DB_SECTION, mk_test_db_key(self._session_ref, self._test_ref))

            test_jnode['rw']['limits'] = {
                'log': {
                    'max-items': self._MAX_LOG_ITEMS
                }
            }
            test_ro_jnode = test_jnode.get('ro') or dict()
            term_uuid = test_ro_jnode.get('origin')
            if term_uuid:
                test_jnode['rw']['limits']['origin'] = {
                    'uuid': term_uuid
                }

                term_uref = ph.map_term_uuid_to_uref(term_uuid)
                if term_uref:
                    term_uimod = Umod.create(term_uref)
                    video_wh = CamDgStats.create(term_uimod.pid).stream_wh
                    if video_wh:
                        video_w, video_h = video_wh
                        test_jnode['rw']['limits']['origin']['live'] = {
                            'wh': [video_w, video_h]
                        }

                    if self._test_ref[0] in [_TEST_TYPE_SPOT_SPEED, _TEST_TYPE_POSITION]:
                        ra_type = TerminalApi.get_ra_type(term_uimod.pid)
                        if ra_type:
                            test_jnode['rw']['limits']['simulator'] = [
                                SIMULATOR_TYPE_EKTSDG,
                                SIMULATOR_TYPE_IS24D,
                                SIMULATOR_TYPE_SAPSAN3_1,
                                SIMULATOR_TYPE_SAPSAN3_2
                            ]
                        else:
                            test_jnode['rw']['limits']['simulator'] = []

            return test_jnode['rw']
