import os
import time

from edge import eventlog
from edge.support import fs
from eventstore.services.types import StatsArgs


class _CountersError(Exception):
    pass


class StoreCounters:
    def __init__(self):
        self.link_present = int()
        self.link_absent = int()
        self.time_now = 0.0

    def inc_link_present(self) -> None:
        self.link_present += 1

    def inc_link_absent(self) -> None:
        self.link_absent += 1

    def last_event_time(self) -> None:
        self.time_now = time.time()


class ReactCounters:
    def __init__(self):
        self.react_count = int()

    def inc_fact_count(self) -> None:
        self.react_count += 1


class EventCounters:
    def __init__(self):
        self.event_count = int()

    def inc_event_count(self) -> None:
        self.event_count += 1


class StatsService:
    def __init__(self, twisted_reactor, args: StatsArgs, dump_event: bool):
        self.store = StoreCounters()
        self.react = ReactCounters()
        self.event = EventCounters() if dump_event else None

        if args.stat_tout_sec:
            twisted_reactor.callLater(
                args.stat_tout_sec, self._log_stat_counters,
                twisted_reactor, args.stat_filepath, args.stat_tout_sec)

    def _log_stat_counters(self, twisted_reactor, stat_filepath: str, stat_tout_sec: int) -> None:
        # noinspection PyBroadException
        try:
            self._dump(stat_filepath, stat_tout_sec)
        except _CountersError:
            pass
        except Exception:
            eventlog.log_x()
        finally:
            twisted_reactor.callLater(
                stat_tout_sec, self._log_stat_counters,
                twisted_reactor, stat_filepath, stat_tout_sec)

    def _dump(self, stat_filepath: str, stat_tout_sec: int) -> None:
        """
        :raises: CountersError
        """
        sys_utc_time = time.time()

        stat = {
            'stat': {
                'mtime': [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                'tout-sec': stat_tout_sec
            },
            'db': {
                'link-present': self.store.link_present,
                'link-absent': self.store.link_absent,
                'last-event': [
                    time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(self.store.time_now)),
                    int(round(self.store.time_now))
                ]
            },
            'react': {
                'count': self.react.react_count
            }
        }

        if self.event is not None:
            stat['event'] = {
                'count': self.event.event_count
            }

        try:
            tmp_path = stat_filepath + '.tmp'
            fs.write_json_file(stat, tmp_path)
            os.rename(tmp_path, stat_filepath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{stat_filepath}> ({x})')
            raise _CountersError

    def destroy(self) -> None:
        pass


def init_service(twisted_reactor, service_args: StatsArgs, dump_event: bool) -> StatsService:
    """
    :param twisted_reactor:
        twisted.internet.interfaces.IReactorTime
    :param service_args:
    :param dump_event:

    :rtype: StatsService
    """
    return StatsService(twisted_reactor, service_args, dump_event)


def fini_service(service) -> None:
    service.destroy()
