from typing import Callable, Dict

from edge import eventlog
from edge.services.react import ReactSource
from edge.support.imloader import ImrefLoader
from edge.types.image import ImRgba
from renderer._args import ImLoaderArgs
from renderer.services.types.react import IReactService, ReactServiceConf
from renderer.types.jnode.tr import decode__tr_scene
from renderer.types.tr import TrScene


class _ReactService(IReactService):
    def __init__(self, twisted_reactor, conf: ReactServiceConf, im_loader_args: ImLoaderArgs) -> None:
        self._source = ReactSource(twisted_reactor)
        self._source.set_handler(self._handle_react)
        self._source.set_endpoints(conf.endpoints)

        self._react_ts: int = 0
        self._handler: Callable[[TrScene, ImRgba], None] | None = None

        self._footer_height = im_loader_args.dh

        self._im_loader = ImrefLoader(im_loader_args.image_wh[0], im_loader_args.image_wh[1])
        twisted_reactor.addSystemEventTrigger('before', 'shutdown', self._im_loader.destroy)
        self._rgba_image = ImRgba()

    def destroy(self) -> None:
        self._source.destroy()

    def set_handler(self, handler: Callable[[TrScene, ImRgba], None]) -> None:
        self._handler = handler

    def _handle_react(self, react_name: str, react_jnode: Dict) -> None:
        if self._handler is None:
            return

        # noinspection PyBroadException
        try:
            tr_scene = decode__tr_scene(react_jnode)

        except (KeyError, ValueError) as x:
            eventlog.log_e(f'fail: decode {react_name} <{x.message}>')
            return

        except Exception:
            eventlog.log_x()
            return

        react_ts = tr_scene.im_ref.ts
        if react_ts <= self._react_ts:
            eventlog.log_w(f'fail: curr-ts: {self._react_ts}, prev-ts:{react_ts}')
            return

        try:
            self._im_loader.load_rgba(self._rgba_image, tr_scene.im_ref, self._footer_height)
        except ValueError as x:
            eventlog.log_e(f'fail: load image {tr_scene.im_ref.url} <{str(x)}> ')
            return

        # noinspection PyBroadException
        try:
            self._handler(tr_scene, self._rgba_image)
            self._react_ts = react_ts
        except Exception:
            eventlog.log_x()
            return


def init_react_service(twisted_reactor, conf: ReactServiceConf, im_loader_args: ImLoaderArgs) -> IReactService:
    return _ReactService(twisted_reactor, conf, im_loader_args)
