from typing import Tuple, Final, Self, Optional, Set

from PIL import ImageDraw, Image, ImageFont

from edge.image.builder import LiveImBuilder
from edge.image.helpers import LineDrawer
from edge.image.types.footer import ImFooterVars
from edge.services.types.about import IAboutService
from edge.services.types.cam import ICamService
from edge.services.types.lane import ILaneService
from edge.services.types.tl import TlState, ITlService
from edge.services.types.tz import ITzService
from edge.support.service import ServicePack
from edge.types.exceptions import ParamsError
from edge.types.image import ImRgba
from edge.types.shape.point import PointI

from renderer._color_store import LrColorToutStore
from renderer.services.types.renderer import IRenderer, RendererConf, RaAnnoConf, LrAnnoConf
from renderer.services.types.tl import TlConf
from renderer.types.face import FaceSceneItem
from renderer.types.lr import LrSceneItem
from renderer.types.ra import RaSceneItem
from renderer.types.tr import TrScene, TrSceneItem

_FONT_PT_SIZE = 20
_RA_BOX_DIR = 1
_RA_BOX_SPEED = 2
_RA_BOX_LENGTH = 3
_RA_BOX_LANE = 4
_LR_BOX_DIR = 5
_LR_BOX_PLATE_TEXT = 6
_LR_BOX_LANE = 7
_LR_BOX_COORD_BBOX = 8
_RA_BOX_SPEED_WITHOUT_LENGTH = 8
_RA_BOX_DIR_WITHOUT_LENGTH = 9
_SYM_W = 12
_SYM_H = 17
_SYM_MARGIN_W = 6
_SYM_MARGIN_H = 3
_MODE_ALWAYS = 'always'
_MODE_NEVER = 'never'
_MODE_IF_PLATE = 'if-plate'
_LANE_UNDEFINED = '0'

COLOR_LIGHT_BLUE = (0, 191, 255)
COLOR_PINK = (255, 192, 203)

COLOR_YELLOW = (255, 255, 0)
COLOR_GREEN = (0, 255, 0)


def _get_ra_board_box_ltrb(ref_point_xy: Tuple[int, int], box_id: int) -> Tuple[int, int, int, int]:
    if box_id == _RA_BOX_DIR:
        ir = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        il = ir - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W

    elif box_id == _RA_BOX_SPEED:
        ir = ref_point_xy[0] - (_SYM_MARGIN_W // 2)
        il = ir - 4 * _SYM_W - _SYM_MARGIN_W // 2

    elif box_id == _RA_BOX_LENGTH:
        il = ref_point_xy[0] + (_SYM_MARGIN_W // 2)
        ir = il + 4 * _SYM_W + _SYM_MARGIN_W

    elif box_id == _RA_BOX_LANE:
        il = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        ir = il + 2 * _SYM_MARGIN_W + _SYM_W

    elif box_id == _RA_BOX_SPEED_WITHOUT_LENGTH:
        ir = ref_point_xy[0] + 2 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W / 4

    elif box_id == _RA_BOX_DIR_WITHOUT_LENGTH:
        ir = ref_point_xy[0] - 2 * _SYM_W - _SYM_MARGIN_W
        il = (ref_point_xy[0] + (_SYM_MARGIN_W // 2)) - 4 * _SYM_W - _SYM_MARGIN_W

    else:
        raise NotImplementedError

    it = ref_point_xy[1] + _SYM_MARGIN_H
    ib = it + _SYM_H + _SYM_MARGIN_H
    return il, it, ir, ib


def _get_lr_board_box_ltrb(ref_point_xy: Tuple[int, int], box_id: int) -> Tuple[int, int, int, int]:
    if box_id == _LR_BOX_DIR:
        ir = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W - _SYM_MARGIN_W
        il = ir - _SYM_MARGIN_W - 2 * _SYM_MARGIN_W

    elif box_id == _LR_BOX_PLATE_TEXT:
        ir = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W

    elif box_id == _LR_BOX_LANE:
        il = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W + _SYM_MARGIN_W
        ir = il + _SYM_MARGIN_W + _SYM_W

    elif box_id == _LR_BOX_COORD_BBOX:
        ir = ref_point_xy[0] + 4 * _SYM_W + _SYM_MARGIN_W
        il = ref_point_xy[0] - 4 * _SYM_W - _SYM_MARGIN_W

    else:
        raise NotImplementedError

    if box_id == _LR_BOX_COORD_BBOX:
        it = ref_point_xy[1] - 2 * _SYM_MARGIN_H - _SYM_H
        ib = it + _SYM_H + _SYM_MARGIN_H
    else:
        it = ref_point_xy[1] + _SYM_MARGIN_H
        ib = it + _SYM_H + _SYM_MARGIN_H
    return il, it, ir, ib


def _render_text(draw: ImageDraw,
                 bg_ltrb: Tuple[int, int, int, int],
                 fg_text: str,
                 bg_color: Tuple[int, int, int] = (255, 255, 255),
                 fg_color: Tuple[int, int, int] = (0, 0, 0)):

    # render background
    draw.rectangle(bg_ltrb, fill=bg_color)

    # render foreground
    text_w = draw.getfont().getlength(fg_text)
    text_il = (bg_ltrb[0] + bg_ltrb[2]) // 2 - text_w // 2
    text_it = (bg_ltrb[1] + bg_ltrb[3]) // 2 - _FONT_PT_SIZE // 2

    draw.text((text_il, text_it), fg_text, fill=fg_color)


def _render_speed(draw: ImageDraw,
                  bg_ltrb: Tuple[int, int, int, int],
                  speed: int,
                  yellow_speed: int,
                  red_speed: int):

    if red_speed and abs(speed) > red_speed:
        bg_color = (208, 100, 89)  # red

    elif yellow_speed and abs(speed) > yellow_speed:
        bg_color = (255, 214, 0)  # amber

    else:
        bg_color = (145, 186, 108)  # green

    _render_text(draw, bg_ltrb, fg_text=str(abs(speed)), bg_color=bg_color)


def _render_dir(draw: Image, bg_ltrb: Tuple[int, int, int, int], fg_dir: int) -> None:
    _DIR_APPROACHING = 11
    _DIR_RECEDING = 12

    if fg_dir == _DIR_APPROACHING:
        fg_text = u'\u2193'
        bg_color = (140, 200, 255)

    elif fg_dir == _DIR_RECEDING:
        fg_text = u'\u2191'
        bg_color = (215, 165, 255)

    else:
        fg_text = '?'
        bg_color = (255, 180, 90)

    _render_text(draw, bg_ltrb, fg_text, bg_color)


def _render_tl(draw: ImageDraw, im_draw_w: int, im_draw_h: int, tl_state: TlState, tl_args: TlConf) -> None:
    if tl_state.r_active and tl_args.red is not None:
        r_ltrb = tl_args.red.get_rect(im_draw_w, im_draw_h)
        draw.rectangle(r_ltrb, fill='red')

    if tl_state.a_active and tl_args.amber is not None:
        a_ltrb = tl_args.amber.get_rect(im_draw_w, im_draw_h)
        draw.rectangle(a_ltrb, fill='yellow')

    if tl_state.g_active and tl_args.green is not None:
        g_ltrb = tl_args.green.get_rect(im_draw_w, im_draw_h)
        draw.rectangle(g_ltrb, fill='green')


class _ServiceDeps:
    def __init__(self,
                 tl_service: ITlService,
                 lane_service: ILaneService,
                 about_service: IAboutService,
                 tz_service: ITzService,
                 cam_service: ICamService | None) -> None:

        self.tl: Final[ITlService] = tl_service
        self.lane: Final[ILaneService] = lane_service
        self.about: Final[IAboutService] = about_service
        self.tz: Final[ITzService] = tz_service
        self.cam: Final[ICamService | None] = cam_service

    @classmethod
    def create(cls, services: ServicePack) -> Self:
        return cls(
            services.get(ITlService),
            services.get(ILaneService),
            services.get(IAboutService),
            services.get(ITzService),
            services.try_get(ICamService))


class Renderer(IRenderer):
    _FONT = ImageFont.truetype('/usr/share/fonts/truetype/croscore/Cousine-Bold.ttf', _FONT_PT_SIZE, encoding='utf-8')

    _PLATE_MARGIN_X = int(2 * _FONT_PT_SIZE / 10.0)
    _PLATE_MARGIN_Y = int(1 * _FONT_PT_SIZE / 10.0)
    _TEXT_BLACK_COLOR = (0, 0, 0)

    _RA_ITEM_RADIUS = 4

    COLOR_BLACK = (0, 0, 0)
    COLOR_WHITE = (255, 255, 255)
    COLOR_BLUE = (0, 255, 255)
    COLOR_RED = (255, 0, 0)

    # http://www.stm.dp.ua/web-design/color-html.php
    _COLORS = [
        (255, 248, 220),  # Cornsilk
        (255, 228, 181),  # Moccasin
        (255, 245, 238),  # Seashell
        (240, 255, 240),  # Honeydew
        (230, 230, 250),  # lavender
        (240, 255, 255),  # Azure
        (255, 228, 225),  # MistyRose
        (152, 251, 152),  # PaleGreen
        (255, 215,   0),  # Gold
        (188, 143, 143),  # RosyBrown
        (245, 245, 220),  # Beige
        (245, 222, 179),  # Wheat
        (193, 255, 193),  # DarkSeaGreen1
        (84,  255, 159)   # SeaGreen1
    ]

    def __init__(self, twisted_reactor, deps: _ServiceDeps, args: RendererConf, im_builder: LiveImBuilder) -> None:
        self._deps: Final[_ServiceDeps] = deps
        self._color_store = LrColorToutStore(twisted_reactor, args.lr_store_tout)

        self._lr_anno_args = args.lr_anno
        self._ra_anno_args = args.ra_anno
        self._tl_anno_args: Optional[TlConf] = args.tl_args[0] if len(args.tl_args) > 0 else None
        self._im_builder: Final[LiveImBuilder] = im_builder

    def render_tr_scene(self, tr_scene: TrScene, im_rgba: ImRgba) -> Image:
        # eventlog.log_i("gr {}, r {}, a {}".format(green, red, amber ))

        w, h = im_rgba.wh
        with Image.frombuffer(
                'RGBA', (w, h + im_rgba.dh), im_rgba.buffer,
                'raw', 'RGBA', im_rgba.stride, 1) as pil_img:

            pil_img.readonly = 0

            im_draw_w, im_draw_h = im_rgba.wh

            im_draw = ImageDraw.Draw(pil_img)

            # TODO: render lamps, add lamps position in conf ?
            if self._tl_anno_args is not None:
                _render_tl(im_draw, im_draw_w, im_draw_h, self._deps.tl.get_state(tr_scene.im_ref.ts),
                           self._tl_anno_args)

            im_draw.font = Renderer._FONT

            list(map(lambda ra_item: self._render_ra_item(
                im_draw, im_draw_w, im_draw_h, self._ra_anno_args, ra_item,
                tr_scene.im_ref.width, tr_scene.im_ref.height),
                     tr_scene.ra_items))

            list(map(lambda lr_item: self._render_lr_item(
                im_draw, im_draw_w, im_draw_h, self._lr_anno_args, lr_item,
                tr_scene.im_ref.width, tr_scene.im_ref.height),
                tr_scene.lr_items))

            list(map(lambda face_item: self._render_face_item(
                im_draw, im_draw_w, im_draw_h, tr_scene.im_ref.width, tr_scene.im_ref.height, face_item),
                tr_scene.face_items))

            list(map(lambda tr_item: self._render_tr_item(
                im_draw, im_draw_w, im_draw_h, tr_item,
                tr_scene.im_ref.width,
                tr_scene.im_ref.height
            ), tr_scene.tr_items))

            origin = self._deps.about.get_origin(im_rgba.ts)
            im_vars = ImFooterVars(
                self._deps.tz, self._deps.about, im_rgba.ts, defvars={},
                geolocation=origin.geolocation, location=origin.location)

            self._im_builder.build_scene(im_rgba, im_vars)
            return pil_img

    def _render_face_item(self, draw: ImageDraw, draw_w: int, draw_h: int,
                          face_scene_w: int, face_scene_h: int, face_item: FaceSceneItem) -> None:

        bbox = face_item.face_bbox

        ra_ltrb = (
            int(float(bbox.x0) / face_scene_w * draw_w),
            int(float(bbox.y0) / face_scene_h * draw_h),
            int(float(bbox.x1) / face_scene_w * draw_w),
            int(float(bbox.y1) / face_scene_h * draw_h)
            )

        draw.rectangle(ra_ltrb, outline=Renderer.COLOR_RED)

        # ltrb = _get_ra_board_box_ltrb((x, y), _RA_BOX_LANE)
        # rgb_image_draw.rectangle(ltrb, outline=Renderer._COLOR_RED)

    @staticmethod
    def _render_ra_box(draw: ImageDraw, draw_w: int, draw_h: int,
                       ra_item: RaSceneItem,
                       lr_scene_w: int,
                       lr_scene_h: int) -> None:

        (x1_1, y1_1, x1_2, y1_2, x1_3, y1_3, x1_4, y1_4,
         x2_1, y2_1, x2_2, y2_2, x2_3, y2_3, x2_4, y2_4) = ra_item.ra_cuboid.coords

        drawer = LineDrawer(draw, draw_w, draw_h, lr_scene_w, lr_scene_h)

        # near plane
        drawer.draw_line((x1_1, y1_1, x1_2, y1_2), Renderer.COLOR_RED)
        drawer.draw_line((x1_2, y1_2, x1_3, y1_3), Renderer.COLOR_RED)
        drawer.draw_line((x1_3, y1_3, x1_4, y1_4), Renderer.COLOR_RED)
        drawer.draw_line((x1_4, y1_4, x1_1, y1_1), Renderer.COLOR_RED)

        # far plane
        drawer.draw_line((x2_1, y2_1, x2_2, y2_2), COLOR_PINK)
        drawer.draw_line((x2_2, y2_2, x2_3, y2_3), COLOR_PINK)
        drawer.draw_line((x2_3, y2_3, x2_4, y2_4), COLOR_PINK)
        drawer.draw_line((x2_4, y2_4, x2_1, y2_1), COLOR_PINK)

        # near to far edges
        drawer.draw_line((x1_1, y1_1, x2_1, y2_1), COLOR_PINK)
        drawer.draw_line((x1_2, y1_2, x2_2, y2_2), COLOR_PINK)
        drawer.draw_line((x1_3, y1_3, x2_3, y2_3), COLOR_PINK)
        drawer.draw_line((x1_4, y1_4, x2_4, y2_4), COLOR_PINK)

    def _render_ra_item(self, draw: Image, draw_w: int, draw_h: int,
                        args: RaAnnoConf,
                        ra_item: RaSceneItem,
                        lr_scene_w: int,
                        lr_scene_h: int,
                        ref_xy: Optional[Tuple[int, int]] = None) -> None:

        if ref_xy:
            lr_ra_match = True
        else:
            ref_xy = ra_item.map_bbox_ref(draw_w, draw_h)
            lr_ra_match = False

        if args.show_target(lr_ra_match):
            abs_x = int(ra_item.ra_rel_im_x * draw_w)
            abs_y = int(ra_item.ra_rel_im_y * draw_h)

            ra_ltrb = (
                abs_x - self._RA_ITEM_RADIUS,
                abs_y - self._RA_ITEM_RADIUS,
                abs_x + self._RA_ITEM_RADIUS,
                abs_y + self._RA_ITEM_RADIUS)

            draw.rectangle(ra_ltrb, fill=Renderer.COLOR_RED)
            #

        if args.show_vehicle(lr_ra_match):
            self._render_ra_box(draw, draw_w,  draw_h, ra_item, lr_scene_w, lr_scene_h)

        if args.show_len(lr_ra_match):
            bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_LENGTH)
            _render_text(draw, bg_ltrb, fg_text=str(ra_item.ra_len))

            if args.show_speed(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_SPEED)
                _render_speed(draw, bg_ltrb, ra_item.ra_speed,
                              args.yellow_speed, args.red_speed)

            if args.show_dir(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_DIR)
                _render_dir(draw, bg_ltrb, ra_item.ra_dir)

        else:
            if args.show_speed(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_SPEED_WITHOUT_LENGTH)
                _render_speed(draw, bg_ltrb, ra_item.ra_speed,
                              args.yellow_speed, args.red_speed)

            if args.show_dir(lr_ra_match):
                bg_ltrb = _get_ra_board_box_ltrb(ref_xy, _RA_BOX_DIR_WITHOUT_LENGTH)
                _render_dir(draw, bg_ltrb, ra_item.ra_dir)

    def _render_lr_item(self, draw: ImageDraw, draw_w: int, draw_h: int,
                        args: LrAnnoConf,
                        lr_item: LrSceneItem,
                        lr_scene_w: int,
                        lr_scene_h: int,
                        ref_point_xy: Optional[Tuple[int, int]] = None) -> None:

        if args.show_trace:
            for x_index in range(0, len(lr_item.lr_trace) - 2, 2):
                # noinspection PyBroadException
                try:
                    image_x = lr_item.lr_trace[x_index] * draw_w / lr_scene_w
                    image_y = lr_item.lr_trace[x_index + 1] * draw_h / lr_scene_h
                    lr_rect_history_ltrb = (image_x - 1, image_y - 1, image_x + 1, image_y + 1)

                    color = self._color_store.get_lr_oid_color(lr_item.lr_oid)
                    draw.rectangle(lr_rect_history_ltrb, fill=color)
                except Exception:
                    pass

        if not ref_point_xy:
            ref_point_xy = lr_item.map_bbox_ref(lr_scene_w, lr_scene_h, draw_w, draw_h)

        if args.show_dir:
            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_DIR)
            _render_dir(draw, bg_ltrb, lr_item.lr_dir)

        if args.show_lane:
            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_LANE)
            lr_center = PointI(*lr_item.lr_quad.center)
            lane_set: Set[str] = self._deps.lane.map_point_abs(lr_center)
            lane: str = lane_set.pop() if len(lane_set) == 1 else _LANE_UNDEFINED
            if len(lane) > 1:
                lane = lane[0]
            _render_text(draw, bg_ltrb, fg_text=lane, bg_color=(206, 180, 141))

        if args.show_text:
            poly = lr_item.lr_quad.map(lr_scene_w, lr_scene_h, draw_w, draw_h)
            draw.polygon(poly, outline=COLOR_YELLOW)

            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy, _LR_BOX_PLATE_TEXT)
            _render_text(draw, bg_ltrb, lr_item.lr_text)

        if args.show_vehicle_bbox and lr_item.vehicle_rect:
            vehicle_poly = lr_item.vehicle_rect.map(lr_scene_w, lr_scene_h, draw_w, draw_h)
            draw.rectangle(vehicle_poly, outline=Renderer.COLOR_BLUE)

        if args.show_vehicle_cuboid and lr_item.vehicle_rect:
            try:
                lr_f, lr_l = self._deps.cam.mk_world_point(lr_item.lr_quad.center_l, lr_item.vehicle_rect.ib, 0)

                x1_1, y1_1 = self._deps.cam.mk_image_point(lr_l - 1, 0, lr_f)
                x1_2, y1_2 = self._deps.cam.mk_image_point(lr_l + 1, 0, lr_f)
                x1_3, y1_3 = self._deps.cam.mk_image_point(lr_l + 1, 2, lr_f)
                x1_4, y1_4 = self._deps.cam.mk_image_point(lr_l - 1, 2, lr_f)

                x2_1, y2_1 = self._deps.cam.mk_image_point(lr_l - 1, 0, lr_f + 4.5)
                x2_2, y2_2 = self._deps.cam.mk_image_point(lr_l + 1, 0, lr_f + 4.5)
                x2_3, y2_3 = self._deps.cam.mk_image_point(lr_l + 1, 2, lr_f + 4.5)
                x2_4, y2_4 = self._deps.cam.mk_image_point(lr_l - 1, 2, lr_f + 4.5)

                # near edge
                drawer = LineDrawer(draw, draw_w, draw_h, lr_scene_w, lr_scene_h)

                drawer.draw_line((x1_1, y1_1, x1_2, y1_2), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_2, y1_2, x1_3, y1_3), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_3, y1_3, x1_4, y1_4), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_4, y1_4, x1_1, y1_1), COLOR_LIGHT_BLUE)

                # far edge
                drawer.draw_line((x2_1, y2_1, x2_2, y2_2), COLOR_LIGHT_BLUE)
                drawer.draw_line((x2_2, y2_2, x2_3, y2_3), COLOR_LIGHT_BLUE)
                drawer.draw_line((x2_3, y2_3, x2_4, y2_4), COLOR_LIGHT_BLUE)
                drawer.draw_line((x2_4, y2_4, x2_1, y2_1), COLOR_LIGHT_BLUE)

                # lines
                drawer.draw_line((x1_1, y1_1, x2_1, y2_1), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_2, y1_2, x2_2, y2_2), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_3, y1_3, x2_3, y2_3), COLOR_LIGHT_BLUE)
                drawer.draw_line((x1_4, y1_4, x2_4, y2_4), COLOR_LIGHT_BLUE)
            except ParamsError:
                pass

        if args.show_center:
            ref_point_xy2 = lr_item.map_bbox_ref2(lr_scene_w, lr_scene_h, draw_w, draw_h)

            x = ref_point_xy2[0]
            y = (ref_point_xy[1] + ref_point_xy2[1]) / 2

            bg_ltrb = _get_lr_board_box_ltrb(ref_point_xy2, _LR_BOX_COORD_BBOX)
            _render_text(draw, bg_ltrb, f'{x} {y}')

    def _render_tr_item(self, draw: Image, draw_w: int, draw_h: int,
                        tr_item: TrSceneItem, lr_scene_w: int, lr_scene_h: int) -> None:

        lr_item = tr_item.lr_item
        ref_point_x, ref_point_y = lr_item.map_bbox_ref(lr_scene_w, lr_scene_h, draw_w, draw_h)
        self._render_lr_item(draw, draw_w, draw_h, self._lr_anno_args,
                             lr_item, lr_scene_w, lr_scene_h,
                             (ref_point_x, ref_point_y + 2 * _SYM_MARGIN_H))

        ra_item = tr_item.ra_item
        ra_ref_xy = ref_point_x, ref_point_y + _SYM_H + 5 * _SYM_MARGIN_H
        self._render_ra_item(draw, draw_w, draw_h, self._ra_anno_args, ra_item, lr_scene_w, lr_scene_h, ra_ref_xy)


def init_renderer(twisted_reactor, services: ServicePack, conf: RendererConf, im_builder: LiveImBuilder) -> IRenderer:
    return Renderer(twisted_reactor, _ServiceDeps.create(services), conf, im_builder)
