import time
import os
from abc import ABCMeta, abstractmethod

from edge.services.types.service import IService
from edge.support import fs
from edge import eventlog
from renderer.services.types.stats import StatsServiceConf


class WhCounters:
    def __init__(self):
        self.w = 0
        self.h = 0
        self.dh = 0

    def im_size(self, w: int, h: int, dh: int) -> None:
        self.w = w
        self.h = h
        self.dh = dh


class IStatService(IService):
    @property
    @abstractmethod
    def wh(self) -> WhCounters:
        pass


class _StubStatService(IStatService):
    def __init__(self) -> None:
        self._wh = WhCounters()

    @property
    def wh(self) -> WhCounters:
        return self._wh


class _StatService(IStatService):
    def __init__(self, twisted_reactor, conf: StatsServiceConf) -> None:
        self._wh = WhCounters()

        self._stat_abspath = conf.stat_filepath
        self._stat_tmp_abspath = conf.stat_filepath + '.tmp'
        self._stat_tout_sec = conf.stat_tout_sec

        if conf.stat_tout_sec > 0:
            twisted_reactor.callLater(conf.stat_tout_sec, self._log_counters, twisted_reactor)

    @property
    def wh(self) -> WhCounters:
        return self._wh

    def _log_counters(self, twisted_reactor) -> None:
        self._report_counters()
        twisted_reactor.callLater(self._stat_tout_sec, self._log_counters, twisted_reactor)

    def _report_counters(self) -> None:
        sys_utc_time = time.time()

        stat = {
            "stat": {
                "mtime": [time.strftime('%d.%m.%Y %H:%M:%S', time.gmtime(sys_utc_time)), int(round(sys_utc_time))],
                "tout-sec": self._stat_tout_sec
            },
            "stream": {
                "wh": [self.wh.w, self.wh.h + self.wh.dh],
                "dh": self.wh.dh
            }
        }

        try:
            fs.write_json_file(stat, self._stat_tmp_abspath, True)
            os.rename(self._stat_tmp_abspath, self._stat_abspath)
        except IOError as x:
            eventlog.log_w(f'fail: write <{self._stat_abspath}> ({x})')


def init_stats_service(twisted_reactor, conf: StatsServiceConf) -> IStatService:
    return _StatService(twisted_reactor, conf) if conf.stat_tout_sec else _StubStatService()
