from typing import Callable, Dict, Optional, Self

from edge import eventlog
from edge.services import gnss, tz, about, db_wanted, lane, cam, tl, plateinfo, systest
from edge.services.db_wanted import DbWantedService
from edge.services.types.about import IAboutService
from edge.services.types.cam import ICamService
from edge.services.types.gnss import IGnssService
from edge.services.types.lane import ILaneService
from edge.services.types.plateinfo import IPlateInfoService
from edge.services.types.systest import ISystestService
from edge.services.types.tl import ITlService
from edge.services.types.tz import ITzService
from edge.services.types.vs import IVsService
from edge.support.service import ServicePack
from edge.types.exceptions import GenericError
from edge.types.trackset import Trackset

from speedcontrol._kernel_args import KernelConf
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.services.event import ev_service
from speedcontrol.services.types.bs import IBsService, BsSenderToChanMap
from speedcontrol.services.types.clock import IClockService
from speedcontrol.services.types.db import IDbService
from speedcontrol.services.types.event import IEventService
from speedcontrol.services.types.leddar import ILeddarService
from speedcontrol.services.types.raas import IRaasService
from speedcontrol.services.types.react_le import ILeReactService
from speedcontrol.services.types.react_tr import ITrReactService
from speedcontrol.services.types.sign import ISignService
from speedcontrol.services.types.stats import IScStatsService
from speedcontrol.services.types.trainer import ITrainerService

from speedcontrol.services import (stats, trainer, raas, bs, vs, leddar, db, clock, sign, react_le, react_tr)
from speedcontrol.detectors import base
from speedcontrol.detectors import buslane
from speedcontrol.detectors import parking
from speedcontrol.detectors import redlight
from speedcontrol.detectors import redlight_leddar
from speedcontrol.detectors import sectionspeed
from speedcontrol.detectors import spotspeed
from speedcontrol.detectors import spotspeed_waga
from speedcontrol.detectors import transit
from speedcontrol.detectors import wanted
from speedcontrol.detectors import wrongway
from speedcontrol.detectors import seatbelt
from speedcontrol.detectors.types.detector import IBaseDetector, DetectorGroups


class KernelDeps:
    def __init__(self,
                 stats_service: stats.IScStatsService,
                 leddar_service: leddar.ILeddarService,
                 tr_react_service: Optional[ITrReactService],
                 le_react_service: Optional[ILeReactService],
                 event_service: ev_service.IEventService):

        self.stats = stats_service
        self.leddar = leddar_service
        self.tr_react = tr_react_service
        self.le_react = le_react_service
        self.event = event_service

    @classmethod
    def create(cls, services: ServicePack) -> Self:
        return cls(
            services.get(IScStatsService),
            services.get(ILeddarService),
            services.try_get(react_tr.ITrReactService),
            services.try_get(react_le.ILeReactService),
            services.get(IEventService))


class BaseKernel(stats.IStatsProvider):
    def __init__(self,
                 services: ServicePack,
                 base_detector: IBaseDetector,
                 detectors: DetectorGroups) -> None:

        self._deps = KernelDeps.create(services)
        self._services = services
        self._detectors = detectors

        list(map(lambda x: x.set_handler(self._handle_event), detectors.plate_absent))
        list(map(lambda x: x.set_handler(self._handle_event), detectors.plate_detected))
        list(map(lambda x: x.set_handler(self._handle_event), detectors.plate_lost))

        self._base_detector = base_detector
        self._base_detector.set_handler(self._handle_base_event)

        if self._deps.le_react:
            self._deps.le_react.set_handler(self._handle_le_react)

        if self._deps.tr_react:
            self._deps.tr_react.set_handler(self._handle_tr_react)

    def destroy(self):
        list(map(lambda x: x.destroy(), self._detectors.plate_lost))
        list(map(lambda x: x.destroy(), self._detectors.plate_absent))
        list(map(lambda x: x.destroy(), self._detectors.plate_detected))

        self._base_detector.destroy()
        self._services.destroy()

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        self._services.collect_stats(json_handler)
        self._base_detector.collect_stats(json_handler)

        list(map(lambda x: x.collect_stats(json_handler), self._detectors.plate_absent))
        list(map(lambda x: x.collect_stats(json_handler), self._detectors.plate_detected))
        list(map(lambda x: x.collect_stats(json_handler), self._detectors.plate_lost))

    def _handle_tr_react(self, trackset: Trackset) -> None:
        lr_track = trackset.lr_track
        if lr_track:
            if lr_track.stat_detected:
                if len(self._detectors.plate_detected) == 0:
                    return

        elif len(self._detectors.plate_absent) == 0:
            return

        # noinspection PyBroadException
        try:
            self._base_detector.analyze(trackset)
        except Exception:
            eventlog.log_x()

    def _handle_le_react(self, trackset: Trackset) -> None:
        if self._detectors.plate_absent:
            # noinspection PyBroadException
            try:
                self._base_detector.analyze(trackset)
            except Exception:
                eventlog.log_x()

    # noinspection PyBroadException
    def _handle_base_event(self, trackset: Trackset, base_event: BaseEvent) -> None:
        if trackset.lr_track is None:
            try:
                list(map(lambda x: x.analyze(trackset, base_event), self._detectors.plate_absent))
            except GenericError:
                eventlog.log_w('fail')
            except Exception:
                eventlog.log_x()

        elif trackset.lr_track.stat_detected:
            try:
                list(map(lambda x: x.analyze(trackset, base_event), self._detectors.plate_detected))
            except GenericError:
                eventlog.log_w('fail')
            except Exception:
                eventlog.log_x()

        elif trackset.lr_track.stat_lost:
            try:
                list(map(lambda x: x.analyze(trackset, base_event), self._detectors.plate_lost))
            except GenericError:
                eventlog.log_w('fail')
            except Exception:
                eventlog.log_x()

    def _handle_event(self, e: Event) -> bool:
        return self._deps.event.emit(e)


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, conf: KernelConf):
        services = ServicePack()

        services.add(IScStatsService, stats.init_service(twisted_reactor, conf.counters))
        stat_service = services.get(IScStatsService)
        stat_service.add_providers([self])

        services.add(DbWantedService, db_wanted.init_service(conf.db.db_args))
        services.add(IBsService, bs.init_service(twisted_reactor, conf.bs, stat_service.bs))
        services.add(IClockService, clock.init_service(twisted_reactor))
        services.add(IDbService, db.init_service(conf.db))
        services.add(ILaneService, lane.init_service(conf.road.lane_service))
        services.add(IPlateInfoService, plateinfo.init_service(conf.plateinfo))
        services.add(IRaasService, raas.init_service(twisted_reactor, conf.raas))
        services.add(ISignService, sign.init_service(twisted_reactor, conf.sign))
        services.add(ISystestService, systest.init_service(conf.systest))
        services.add(ITlService, tl.init_service(twisted_reactor, conf.tl))
        services.add(ITrReactService, react_tr.init_service(twisted_reactor, conf.tr_react, services.get(ILaneService)))
        services.add(ITzService, tz.init_service(conf.tz))
        services.add(IVsService, vs.init_service(twisted_reactor, conf.vs))

        services.add(IGnssService, gnss.init_service(twisted_reactor, conf.gnss, services.get(ITzService)))
        services.add(ITrainerService, trainer.init_service(conf.trainer, services.get(ITzService)))
        services.add(IAboutService, about.init_service(conf.about, services.get(IGnssService)))

        services.add(ILeddarService, leddar.init_service(conf.leddar))
        if conf.leddar:
            services.add(ILeReactService, react_le.init_service(
                twisted_reactor,
                {conf.leddar.endpoint},
                services.get(ILeddarService)))

        services.add(IEventService, ev_service.init_service(
            twisted_reactor,
            stat_service.event,
            conf.event,
            services.get(ITzService),
            services.get(IBsService),
            services.get(ISystestService),
            services.get(IAboutService),
            services.get(ISignService)))

        if conf.cam is not None:
            services.add(ICamService, cam.init_service(conf.cam))

        base_detector = base.create(services, conf.detectors.base[0])

        detectors = DetectorGroups()
        transit.create(detectors, services, conf.detectors.transit)
        if not services.get(ISystestService).session_active:
            buslane.create(detectors, services, conf.detectors.bus_lane)
            parking.create(detectors, services, conf.detectors.parking)
            redlight.create(detectors, services, conf.detectors.red_light)
            redlight_leddar.create(detectors, services, conf.detectors.railroad_crossing)
            sectionspeed.create(detectors, services, conf.detectors.section_speed)
            spotspeed.create(detectors, services, conf.detectors.spot_speed)
            spotspeed_waga.create(detectors, services, conf.detectors.spot_speed_waga)
            wanted.create(detectors, services, conf.detectors.wanted)
            wrongway.create(detectors, services, conf.detectors.wrong_way)
            seatbelt.create(detectors, services, conf.detectors.seatbelt)

        sender_to_chan_map: BsSenderToChanMap = dict()
        for detector in detectors.plate_absent:
            sender_to_chan_map.update({detector.identity: detector.bsconf})

        for detector in detectors.plate_detected:
            sender_to_chan_map.update({detector.identity: detector.bsconf})

        for detector in detectors.plate_lost:
            sender_to_chan_map.update({detector.identity: detector.bsconf})

        services.get(IBsService).setup(sender_to_chan_map)

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', lambda: self.destroy())

        super().__init__(services, base_detector, detectors)
        self._twisted_reactor = twisted_reactor

    def run(self):
        self._twisted_reactor.run()
