from typing import Dict, Optional

from edge import eventlog
from edge.image.types.jnode.builder import decode__im_builder_conf
from edge.types.dir import Dir
from edge.types.exceptions import JnodeError
from edge.types.identity import Identity

from speedcontrol.conf.jnode.mask import decode__mask_conf
from speedcontrol.detectors.types.jnode.conf import decode__base_rule_conf, decode__video_conf
from speedcontrol.detectors.redlight.types.conf import DetectorConf, RuleConf


def decode__rule_conf(rule_jnode: Dict, footer_defvars: Dict) -> Optional[RuleConf]:
    """
    :raise JnodeError:
    """
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/name
    # deserialize #/enabled
    # deserialize #/ticket
    rule = decode__base_rule_conf(rule_jnode)
    if rule.enabled:
        return RuleConf(rule.name, rule.ticket, footer_defvars)


def decode__dir_limit_text(value_text: str) -> Dir:
    text_to_code_map: Dict[str, Dir] = {
        'approaching': Dir.CODE_A,
        'receding': Dir.CODE_R,
    }

    value_code = text_to_code_map.get(value_text)
    if value_code is None:
        eventlog.log_e(f'fail: unsupported (value: {value_text})')
        raise JnodeError

    return Dir(value_code)


def decode__detector_conf(identity: Identity, jnode: Dict, footer_defvars: Dict) -> DetectorConf:
    """
    :raise JnodeError:
    """
    # deserialize #/name
    name = jnode.get('name')
    if not type(name) in [str] or len(name) > 32:
        eventlog.log_e_param('#/name')
        raise JnodeError

    # deserialize #/direction
    direction = jnode.get('direction', 'receding')
    if not type(direction) in [str] and direction not in ['receding', 'approaching']:
        eventlog.log_e_param('#/direction')
        raise JnodeError

    det_dir = decode__dir_limit_text(direction)

    # deserialize #/on-stopbar
    on_stopbar = jnode.get('on-stopbar', False)
    if not type(on_stopbar) in [str] and on_stopbar not in ['required', 'optional']:
        eventlog.log_e_param('#/on-stopbar')
        raise JnodeError

    on_stopbar_required = on_stopbar == 'required'

    # deserialize #/store
    store = jnode.get('store')
    if not (type(store) in [str] and 1 <= len(store) <= 2):
        eventlog.log_e_param('#/store')
        raise JnodeError

    # deserialize #/scene
    scene_jnode = jnode.get('scene')
    if type(scene_jnode) != dict:
        eventlog.log_e_param('#/scene')
        raise JnodeError
    scene = decode__mask_conf(scene_jnode)

    # deserialize #/rules
    rules_jnode = jnode.get('rules')
    if type(rules_jnode) != list:
        eventlog.log_e_param('#/rules')
        raise JnodeError

    rules = []
    for rule_jnode in rules_jnode:
        rule = decode__rule_conf(rule_jnode, footer_defvars)
        if rule is not None:
            rules.append(rule)

    # deserialize #/images
    images_jnode = jnode.get('images')
    if type(images_jnode) != dict:
        eventlog.log_e_param('#/images')
        raise JnodeError

    im_engine = decode__im_builder_conf(images_jnode['violation'])

    # deserialize #/videos/measuments
    video_jnode = jnode.get('videos', {}).get('violation')
    if video_jnode is not None:
        if video_jnode.get('enabled', False):
            video_conf = decode__video_conf(video_jnode)
        else:
            video_conf = None
    else:
        video_conf = None

    return DetectorConf(identity, name, store, scene, rules, im_engine, video_conf, det_dir, on_stopbar_required)
