from typing import Callable, Dict, Final, Optional, List, Tuple, Set

from edge import eventlog
from edge.image.builder import ImBuilderEnv
from edge.services.types.vs import VsImageInfo
from edge.support.schedule import Schedule
from edge.support.speed import map_speed_uint
from edge.types.exceptions import InternalError, GenericError
from edge.types.jnode.dir import encode_dir_text
from edge.types.timespan import TimeSpan
from edge.types.trackset.tr import TrTrackItem, TrTrack
from edge.types.trackset import Trackset
from edge.webserver.sysconf.uiconf.controller import CtrlUiconf

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders, get_tr_item_dist
from speedcontrol.detectors.spotspeed.types.conf import RuleConf
from speedcontrol.detectors.spotspeed.types.deps import RuleDeps
from speedcontrol.detectors.spotspeed.types.footer import EvidenceImFooterVars
from speedcontrol.detectors.spotspeed.types.jnode.stats import encode__rule_stats
from speedcontrol.detectors.spotspeed.types.stats import RuleStats
from speedcontrol.events.detector import DetectorInfo, RuleInfo
from speedcontrol.events.event import BaseEvent
from speedcontrol.events.evidence.spotspeed import Evidence
from speedcontrol.events.position import Position
from speedcontrol.events.schedule import ScheduleInfo
from speedcontrol.events.ticket import TicketInfo
from speedcontrol.events.vehicle import Vehicle, ViolationVehicle
from speedcontrol.events.waypoint import Waypoint
from speedcontrol.events.waypoint.violation import ViolationWaypoint


class _TrItemImageInfo(VsImageInfo):
    def __init__(self, tr_item: TrTrackItem, weight: int = 0):
        super().__init__(tr_item.ts)
        self.tr_item = tr_item
        self.weight: Final[weight] = weight


def _create_detector_info(conf: RuleConf) -> DetectorInfo:
    schedule_info = ScheduleInfo(
        conf.schedule.min_time,
        conf.schedule.max_time,
        conf.schedule.days_to_bool)

    ticket_info = TicketInfo(conf.ticket.name)

    rule_info = RuleInfo(conf.rule_name, schedule_info, ticket_info, {
        'speed-limit': conf.limits.speed_limit,
        'max-overspeed': conf.limits.max_overspeed,
        'types': list(conf.limits.vehicle_types)
    })

    return DetectorInfo(conf.detector_name, rule_info)


def _get_ra_speed_max_error_ua(lane_ra_speed: int) -> int:
    return 2 if lane_ra_speed < 100 else (lane_ra_speed * 0.02)


# noinspection PyUnusedLocal
def _get_ra_speed_max_error_uz(lane_ra_speed: int) -> int:
    return 0


# noinspection PyUnusedLocal
def _get_ra_speed_max_error_kz(lane_ra_speed: int) -> int:
    return 2


LaneToTrTrackItemList = List[Tuple[str, TrTrackItem]]


class Rule:
    def __init__(self,
                 deps: RuleDeps,
                 conf: RuleConf,
                 ime_init: ImBuilderEnv,
                 ime_fini: ImBuilderEnv,
                 ime_best: ImBuilderEnv):

        self._deps: Final[RuleDeps] = deps
        self._conf: Final[RuleConf] = conf
        self._info: Final[DetectorInfo] = _create_detector_info(conf)
        self._stats = RuleStats()

        self._schedule: Final[Schedule] = Schedule(conf.schedule, deps.tz)

        self._ime_init = ime_init
        self._ime_fini = ime_fini
        self._ime_best = ime_best

        bylaws = CtrlUiconf.get_bylaws()
        if bylaws in ['ua', 'default']:
            self._get_max_error_fn = _get_ra_speed_max_error_ua
        elif bylaws == 'kz':
            self._get_max_error_fn = _get_ra_speed_max_error_kz
        elif bylaws == 'uz':
            self._get_max_error_fn = _get_ra_speed_max_error_uz
        else:
            eventlog.log_e(f'fail: unsupported (bylaws:{bylaws})')
            raise NotImplementedError()

    def collect_stats(self, json_handler: Callable[[Dict], None]) -> None:
        json_handler(encode__rule_stats(self._conf.identity.ref, self._stats))

    def _test_suitable_class(self, plate_utf8: bytes, vehicle_class: str) -> bool:
        if vehicle_class not in self._conf.limits.vehicle_types:
            self._stats.class_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_utf8}, reason:class-unmatch, class:{vehicle_class})')
            return False

        return True

    def _test_suitable_schedule(self, plate_utf8: bytes, track: Trackset) -> bool:
        suitable = self._schedule.check_ts_usec(track.lr_track.min_ts) \
                   and self._schedule.check_ts_usec(track.lr_track.max_ts)

        if not suitable:
            self._stats.schedule_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_utf8}, reason:schedule-unmatch)')
            return False

        return True

    def _filter_tr_items_by_lane(self, tr_track: TrTrack, lanes_limit: Set[str]) -> List[TrTrackItem]:
        lane_items: List[TrTrackItem] = []
        for tr_item in tr_track.values():
            if tr_item.lr_item is None or tr_item.ra_item is None:
                continue

            if len(lanes_limit) == 0:
                lane_items.append(tr_item)
                continue

            lane_ref: Optional[str] = tr_item.lane_item.lane if tr_item.lane_item is not None else None

            if lane_ref in lanes_limit:
                lane_items.append(tr_item)

        return lane_items

    def check(self,
              trackset: Trackset,
              base_event: BaseEvent) -> Optional[Tuple[Evidence, TimeSpan]]:

        self._stats.total += 1
        plate_text_utf8: bytes = base_event.plate_utf8

        if not self._test_suitable_class(plate_text_utf8, base_event.best_wp.vehicle.mmr_class):
            return

        elif not self._test_suitable_schedule(plate_text_utf8, trackset):
            return

        spotspeeds = base_event.best_wp.vehicle.spotspeeds

        main_speed: int = spotspeeds.main.measured
        max_speed = self._conf.limits.speed_limit + self._conf.limits.max_overspeed
        max_error = self._get_max_error_fn(main_speed)

        if not spotspeeds.assert_overspeed(main_speed, max_speed + max_error):
            self._stats.raspeed_valid += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:speed-allowed, '
                           f'main-speed:{main_speed}, assert-speed:{max_speed + max_error})')
            return

        lane_items = self._filter_tr_items_by_lane(trackset.tr_track, self._conf.limits.lanes)
        if len(lane_items) < 3:
            self._stats.lane_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:lane-unmatch, '
                           f'source-items:{len(trackset.tr_track)}, target-items:{len(lane_items)})')
            return

        #
        # Create an evidence
        #

        self._stats.evidence_exec += 1

        lr_imref = trackset.lr_track.imref
        image_infos: List[_TrItemImageInfo] = []
        for tr_item in lane_items:
            if check_quad_dist_from_img_borders(tr_item.lr_item.plate_quad, lr_imref.width, lr_imref.height):
                image_infos.append(_TrItemImageInfo(tr_item))

        # TODO: how many images are required?
        if len(image_infos) < 2:
            self._stats.lane_unmatch += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:lane-unmatch, '
                           f'source-items:{len(trackset.tr_track)}, target-items:{len(image_infos)})')
            return

        image_infos.sort(key=lambda ref: ref.ts)

        init_image_infos: List[_TrItemImageInfo] = image_infos[0:5]
        # init_scene_ms_refs.sort(key=lambda ref: ref.ts)
        if not init_image_infos:
            self._stats.evidence_failed_load_images += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:load-images)')
            return

        fini_image_infos: List[_TrItemImageInfo] = image_infos[:-6:-1]
        # fini_scene_ms_refs.sort(key=lambda ref: ref.ts, reverse=True)
        if not fini_image_infos:
            self._stats.evidence_failed_load_images += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:load-images)')
            return

        if trackset.lr_track:
            lr_scene_w = trackset.lr_track.imref.width
            lr_scene_h = trackset.lr_track.imref.height
        else:
            lr_scene_w = None
            lr_scene_h = None

        best_image_infos: List[_TrItemImageInfo] = [
            _TrItemImageInfo(x.tr_item, get_tr_item_dist(x.tr_item, lr_scene_w, lr_scene_h))
            for x in image_infos
        ]

        best_image_infos.sort(key=lambda ref: ref.weight, reverse=True)

        if not best_image_infos:
            self._stats.evidence_failed_load_images += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:load-images)')
            return

        init_image_info = self._deps.vs.export_image_rgba(
            self._ime_init.scene_rgba, init_image_infos, self._ime_init.scene_dh)

        best_image_info = self._deps.vs.export_image_rgba(
            self._ime_best.scene_rgba, best_image_infos, self._ime_best.scene_dh)

        fini_image_info = self._deps.vs.export_image_rgba(
            self._ime_fini.scene_rgba, fini_image_infos, self._ime_fini.scene_dh)

        if None in [init_image_info, fini_image_info, best_image_info]:
            self._stats.evidence_failed_load_images += 1
            eventlog.log_i(f'fail: reject track (plate:{plate_text_utf8}, reason:load-images)')
            return

        return self._mk_evidence(
            base_event, trackset,
            self._conf.limits.speed_limit,
            init_image_info.tr_item,
            fini_image_info.tr_item,
            best_image_info.tr_item)

    def _mk_evidence(
            self,
            base_event: BaseEvent,
            trackset: Trackset,
            speed_limit: int,
            init_tr_item: TrTrackItem,
            fini_tr_item: TrTrackItem,
            best_tr_item: TrTrackItem) -> Optional[Tuple[Evidence, TimeSpan]]:

        # get best vehicle
        best_vehicle: Vehicle = base_event.best_wp.vehicle
        spotspeeds = best_vehicle.spotspeeds

        def mk_waypoint(services: RuleDeps,
                        im_env: ImBuilderEnv,
                        tr_item: TrTrackItem,
                        w_type: str = Waypoint.NAME_VIOLATION) -> ViolationWaypoint:

            waypoint_ts = im_env.scene_rgba.ts
            waypoint_pos = Position(tr_item.ra_item.scene_f_cm, tr_item.ra_item.scene_l_cm)
            wp_dir = trackset.tr_track.get_item_dir(tr_item.ts)

            wp_dir_alias = self._conf.footer_defvars.get(encode_dir_text(wp_dir))

            waypoint_origin = services.about.get_origin(waypoint_ts)
            if tr_item.lane_item is not None:
                waypoint_origin.location.set_lane(tr_item.lane_item.lane)

            footer_vars = EvidenceImFooterVars(
                services.tz, self._deps.about, waypoint_ts, self._conf.footer_defvars,
                origin=waypoint_origin,
                spotspeeds=spotspeeds,
                plate_text=base_event.plate_ucode,
                direction=wp_dir,
                speed_limit=speed_limit,
                ra_overspeed=map_speed_uint(spotspeeds.ra_n) - speed_limit,
                main_overspeed=map_speed_uint(spotspeeds.main.measured) - speed_limit,
                mmr_class=best_vehicle.mmr_class)

            if w_type == Waypoint.NAME_VIOLATION:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True)

                if len(im_blobs) != 3:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise InternalError

                waypoint_v = ViolationVehicle(
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    ViolationWaypoint.NAME_VIOLATION,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            elif w_type == Waypoint.NAME_VIOLATION_BEST:
                scene_dt_utc = self._deps.tz.map_ts_usec_to_datetime_utc(waypoint_ts)
                im_blobs = im_env.builder.build(
                    im_env.scene_rgba, scene_dt_utc, waypoint_origin.geolocation,
                    footer_vars, trackset.lr_track, tr_item.lr_item.plate_quad,
                    vehicle=True, scene=True, plate=True, thumb=True)

                if len(im_blobs) != 4:
                    eventlog.log_w('fail: get all blobs for evidence')
                    raise InternalError

                waypoint_v = ViolationVehicle(
                    direction=wp_dir,
                    direction_alias=wp_dir_alias,
                    position=waypoint_pos,
                    plate_quads=[tr_item.lr_item.plate_quad])

                waypoint = ViolationWaypoint(
                    ViolationWaypoint.NAME_VIOLATION_BEST,
                    waypoint_ts, None, waypoint_origin.location, im_blobs,
                    waypoint_v)

            else:
                eventlog.log_e('fail: unsupported')
                raise InternalError

            return waypoint

        try:
            init_wp = mk_waypoint(self._deps, self._ime_init, init_tr_item)
            fini_wp = mk_waypoint(self._deps, self._ime_fini, fini_tr_item)
            best_wp = mk_waypoint(self._deps, self._ime_best, best_tr_item, Waypoint.NAME_VIOLATION_BEST)
        except GenericError:
            self._stats.evidence_failed_make_images += 1
            eventlog.log_i(f'fail: reject track (plate:{base_event.plate_utf8}, reason:make-images)')
            return None

        event_ts = self._ime_fini.scene_rgba.ts

        evidence = Evidence(
            self._info, self._deps.about.get_origin(event_ts), event_ts,
            spotspeeds.main.measured, spotspeeds.control.measured,
            trackset.get_direction(), self._conf.limits.speed_limit)

        evidence.add_waypoints([init_wp, best_wp, fini_wp])
        return evidence, TimeSpan(init_wp.main_ts, fini_wp.main_ts)
