from typing import Dict, List, Optional

from edge import eventlog
from edge.image.types.jnode.builder import decode__im_builder_conf
from edge.types.exceptions import JnodeError
from edge.types.identity import Identity

from speedcontrol.detectors.spotspeed.types.conf import DetectorConf, RuleLimits, RuleConf
from speedcontrol.detectors.types.detector import IDetector
from speedcontrol.detectors.types.jnode.conf import decode__base_rule_conf, decode__video_conf


def decode__limits_conf(limits_jnode: Dict) -> RuleLimits:
    """
    :raise JnodeError:
    """
    # deserialize #/classes
    types_jnode = limits_jnode.get('classes', [])

    if type(types_jnode) != list:
        eventlog.log_e_param('#/classes')
        raise JnodeError

    elif len(types_jnode) == 0:
        eventlog.log_e_param('#/classes')
        raise JnodeError

    for v_type in types_jnode:
        if not type(v_type) in [str]:
            eventlog.log_e_param('#/classes/%s' % v_type)
            raise JnodeError

    # deserialize #/speed-limit
    speed_limit = limits_jnode.get('speed-limit')
    if not (int == type(speed_limit) and 0 <= speed_limit <= 250):
        eventlog.log_e_param('#/speed-limit')
        raise JnodeError

    # deserialize #/max-overspeed
    max_overspeed = limits_jnode.get('max-overspeed')
    if not (int == type(speed_limit) and 0 <= max_overspeed <= 250):
        eventlog.log_e_param('#/max-overspeed')
        raise JnodeError

    # deserialize #/lanes
    lanes_jnode = limits_jnode.get('lanes', [])

    if type(lanes_jnode) != list:
        eventlog.log_e_param('#/lanes')
        raise JnodeError

    for lane in lanes_jnode:
        if not type(lane) in [str]:
            eventlog.log_e_param('#/lanes/%s' % lane)
            raise JnodeError

    return RuleLimits(set(lanes_jnode), speed_limit, max_overspeed, set(types_jnode))


def decode__rule_conf(identity: Identity,
                      rule_jnode: Dict,
                      detector_name: str,
                      footer_defvars: Dict) -> Optional[RuleConf]:
    """
    :raise JnodeError:
    """
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/')
        raise JnodeError

    # deserialize #/name
    # deserialize #/enabled
    # deserialize #/schedule
    # deserialize #/ticket
    base_rule = decode__base_rule_conf(rule_jnode)
    if not base_rule.enabled:
        return

    # deserialize #/limits
    limits_jnode = rule_jnode.get('limits')
    if type(rule_jnode) != dict:
        eventlog.log_e_param('#/limits')
        raise JnodeError

    limits = decode__limits_conf(limits_jnode)

    return RuleConf(identity, detector_name, base_rule.name, base_rule.enabled, base_rule.schedule, base_rule.ticket,
                    limits, footer_defvars)


def decode__detector_conf(identity: Identity, jnode: Dict, footer_defvars: Dict) -> DetectorConf:
    """
    :raise JnodeError:
    """
    # deserialize #/name
    name = jnode.get('name')
    if not type(name) in [str] or len(name) > 32:
        eventlog.log_e_param('#/name')
        raise JnodeError

    # deserialize #/store
    store = jnode.get('store')
    if not (type(store) in [str] and 1 <= len(store) <= 2):
        eventlog.log_e_param('#/store')
        raise JnodeError

    # deserialize #/videos/violation
    video_jnode = jnode.get('videos', {}).get('violation')
    if video_jnode is not None:
        if video_jnode.get('enabled', False):
            video_conf = decode__video_conf(video_jnode)
        else:
            video_conf = None
    else:
        video_conf = None

    # deserialize #/rules
    rules_jnode = jnode.get('rules')
    if type(rules_jnode) != list:
        eventlog.log_e_param('#/rules')
        raise JnodeError

    rules: List[RuleConf] = []
    rule_index = 0
    for rule_jnode in rules_jnode:
        rule_identity = identity.child(IDetector.TYPE_SPOTSPEED_RULE, rule_index)
        rule_index += 1

        rule_conf = decode__rule_conf(rule_identity, rule_jnode, name, footer_defvars)
        if rule_conf is not None:
            rules.append(rule_conf)

    # deserialize #/images
    images_jnode = jnode.get('images')
    if type(images_jnode) != dict:
        eventlog.log_e_param('#/images')
        raise JnodeError

    im_engine = decode__im_builder_conf(images_jnode['violation'])

    return DetectorConf(identity, name, store, rules, im_engine, video_conf)
