import functools
from typing import Callable, Final, List, Self, Optional

from edge import eventlog
from edge.services.types.about import IAboutService
from edge.services.types.tz import ITzService
from edge.services.types.vs import IVsService, VsSessionAbsentError, VsError
from edge.support.schedule import Schedule
from edge.support.service import ServicePack
from edge.types.blob import BsChanRef, BlobProfileToBsChanMap, FileBlob
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.trackset import Trackset

from speedcontrol.detectors.transit.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.detectors.types.detector import IDetector, DetectorUnitInfo, DetectorGroups
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.transit import TransitEvidence
from speedcontrol.events.waypoint.measurement import BestWaypoint
from speedcontrol.services.types.clock import IClockService


class DetectorDeps:
    def __init__(self,
                 about_service: IAboutService,
                 clock_service: IClockService,
                 tz_service: ITzService,
                 vs_service: IVsService):

        self.about = about_service
        self.clock = clock_service
        self.tz = tz_service
        self.vs: Final[IVsService] = vs_service

    @classmethod
    def create(cls, services: ServicePack) -> Self:
        return cls(
            services.get(IAboutService),
            services.get(IClockService),
            services.get(ITzService),
            services.get(IVsService))


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf

        self._schedule = Schedule(conf.schedule_args, deps.tz)
        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return {
            'args': self._conf.blob_store_args.vehicle_store,
            'plate': self._conf.blob_store_args.vehicle_store,
            'vehicle': self._conf.blob_store_args.vehicle_store,
            'thumb': self._conf.blob_store_args.vehicle_store,
            'scene': self._conf.blob_store_args.scene_store,
            'overview': self._conf.blob_store_args.scene_store,
            'measurement': self._conf.blob_store_args.scene_store
        }

    def set_handler(self, handler: Callable[[Event], None]) -> None:
        if self._handler and self._handler != handler:
            raise NotImplementedError

        self._handler = handler

    def analyze(self, track: Trackset, base_event: BaseEvent) -> None:
        if not self._conf.enabled:
            return

        waypoints_list = base_event.measurements[0].waypoints
        best_points = [waypoint for waypoint in waypoints_list if waypoint.name == waypoint.NAME_BEST]
        if len(best_points) != 1:
            eventlog.log_w('fail: cant find best waypoint')
            return

        best_waypoint: BestWaypoint = best_points[0]
        if self._schedule.check_ts_usec(best_waypoint.main_ts):
            if self._conf.require_plate and best_waypoint.vehicle.vrp_absent:
                eventlog.log_i('fail: plate absent')
                return

            if track.le_track and len(track.le_track):
                if track.le_track.ref != self._conf.le_loop_ref:
                    return

            origin = self._deps.about.get_origin(best_waypoint.main_ts)
            control_ts = self._deps.clock.get_control_ts(best_waypoint.main_ts)
            evidence = TransitEvidence(origin, best_waypoint.main_ts, control_ts=control_ts)

            handler = functools.partial(self._handle_evidence, base_event, evidence)

            if self._conf.with_video and track.lr_track is not None and track.lr_track.vrp.type.template_ref == 0:
                # noinspection PyBroadException
                try:
                    return self._deps.vs.export_video(
                        track.tr_track.min_ts // 1000 - 1,
                        track.tr_track.max_ts // 1000 + 1,
                        0, 0, 0,
                        handler)

                except VsSessionAbsentError:
                    eventlog.log_w(f'fail: VsSessionAbsentError (plate:{base_event.plate_utf8})')

                except VsError:
                    eventlog.log_w(f'fail: VsError (plate:{base_event.plate_utf8})')

                except GenericError:
                    eventlog.log_w(f'fail: plate:{base_event.plate_utf8}')

                except Exception:
                    eventlog.log_x()

            handler(None)

    def _handle_evidence(self, base_event: BaseEvent, evidence: TransitEvidence, video_blob: Optional[FileBlob]) -> None:
        if video_blob is not None:
            evidence.add_videos([video_blob])
            # eventlog.log_i(f'video_blob_type {type(video_blob)}')

        event = Event(self._conf.identity, base_event, evidence)
        self._handler(event)
        # eventlog.log_i(f'done: plate:{base_event.plate_utf8}, with-video:{video_blob is not None}')


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            detector = _Detector(DetectorDeps.create(services), conf)

            if not conf.require_plate:
                detectors.plate_lost.add(detector)
                detectors.plate_absent.add(detector)

            elif conf.plate_detected_mode:
                detectors.plate_detected.add(detector)

            else:
                detectors.plate_lost.add(detector)

        except GenericError:
            eventlog.log_e(f'fail: {info.identity.ref}')
            raise
