import functools
import math

from typing import Callable, Final, List, Optional, Tuple

from edge import eventlog
from edge.image.builder import ImBuilderEnv, EventMeasCamImBuilder
from edge.image.scaler import WhScaler
from edge.services.types.vs import VsSessionAbsentError, VsError
from edge.support.mask import Mask
from edge.support.service import ServicePack
from edge.types.blob import FileBlob, BsChanRef, BlobProfileToBsChanMap
from edge.types.exceptions import GenericError
from edge.types.identity import Identity
from edge.types.shape.quad import QuadI
from edge.types.timespan import TimeSpan
from edge.types.track import TrTrackItem
from edge.types.trackset import Trackset

from speedcontrol.detectors.base.best import check_quad_dist_from_img_borders
from speedcontrol.detectors.types.detector import IDetector, DetectorUnitInfo, DetectorGroups
from speedcontrol.detectors.wrongway.ww_rule import Rule
from speedcontrol.detectors.wrongway.types.deps import DetectorDeps
from speedcontrol.detectors.wrongway.types.jnode.conf import DetectorConf, decode__detector_conf
from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence.wrongway import Evidence


def _test_points_far_enought(q1: QuadI, q2: QuadI) -> bool:
    # 70% of plate height
    min_dist = q1.wh[1] * 0.70

    r1_cx, r1_cy = q1.center
    r2_cx, r2_cy = q2.center

    # points are in same rect
    if math.fabs(r1_cx - r2_cx) < min_dist:
        if math.fabs(r1_cy - r2_cy) < min_dist:
            return False

    # points are far enought
    return True


def filter_tr_items(tr_items: List[TrTrackItem], lr_scene_w: int, lr_scene_h: int) -> List[TrTrackItem]:
    result = []
    # todo: lr_item can be null ?
    init_point = tr_items[0]  # type: TrTrackItem
    result.append(init_point)

    for index in range(1, len(tr_items), 1):
        fini_point = tr_items[index]  # type: TrTrackItem
        if fini_point.lr_item is None:
            continue

        if not check_quad_dist_from_img_borders(fini_point.lr_item.plate_quad, lr_scene_w, lr_scene_h):
            continue

        if _test_points_far_enought(init_point.lr_item.plate_quad, fini_point.lr_item.plate_quad):
            result.append(fini_point)
            init_point = fini_point

    return result


class _Detector(IDetector):
    def __init__(self, deps: DetectorDeps, conf: DetectorConf):
        super().__init__()

        self._deps: Final[DetectorDeps] = deps
        self._conf: Final[DetectorConf] = conf

        imbuilder = EventMeasCamImBuilder(conf.imbuilder)
        self._ime_init = ImBuilderEnv(imbuilder)
        self._ime_fini = ImBuilderEnv(imbuilder)

        self._im_fitter: Optional[WhScaler] = WhScaler(self._conf.video.max_wh) if conf.video else None

        self._rules: List[Rule] = []
        for rule_conf in conf.rules:
            if rule_conf.enabled:
                rule = Rule(deps, rule_conf, conf.name, self._ime_init, self._ime_fini)
                self._rules.append(rule)

        self._scene_mask: Final[Mask] = Mask(conf.scene_mask)
        self._handler: Callable[[Event], None] | None = None

    @property
    def identity(self) -> Identity:
        return self._conf.identity

    @property
    def bsconf(self) -> BsChanRef | BlobProfileToBsChanMap:
        return self._conf.store

    def set_handler(self, event_handler: Callable[[Event], None]) -> None:
        if self._handler:
            raise NotImplementedError

        self._handler = event_handler

    def _map_items_to_lanes(self, trackset: Trackset) -> List[Tuple[str, TrTrackItem]]:
        lane_to_item_map: List[Tuple[str, TrTrackItem]] = []
        for tr_item in trackset.tr_track.values():
            if tr_item.lr_item:
                lr_quad = tr_item.lr_item.plate_quad
                if not self._scene_mask.check_quad_abs(lr_quad):
                    continue

                lane_ref: Optional[str] = tr_item.lane_item.lane if tr_item.lane_item is not None else None
                if lane_ref is None:
                    lane_to_item_map.append(('no_lane', tr_item))
                    continue

                lane_to_item_map.append((lane_ref, tr_item))

        return lane_to_item_map

    def analyze(self, trackset: Trackset, base_event: BaseEvent) -> None:
        if '*' in base_event.plate_ucode:
            eventlog.log_i(f'done: skip trackset ({base_event.plate_utf8}, reason:asterisk)')
            return

        lane_to_tr_item = self._map_items_to_lanes(trackset)

        for rule in self._rules:
            try:
                reply: Optional[Tuple[Evidence, TimeSpan]] = rule.check(trackset, base_event, lane_to_tr_item)
                if reply is None:
                    continue
            except GenericError:
                eventlog.log_i(f'fail: skip trackset (<{base_event.plate_utf8}>)')
                continue

            evidence, time_span = reply
            handler = functools.partial(self._handle_evidence_with_vs, base_event, evidence)
            if self._conf.video:
                scaled_w, scaled_h = self._im_fitter.scaledown(self._ime_init.scene_rgba.wh)

                # noinspection PyBroadException
                try:
                    self._deps.vs.export_video(
                        time_span.min_ts // 1000 - self._conf.video.pre_record,
                        time_span.max_ts // 1000 + self._conf.video.post_record,
                        scaled_w, scaled_h, self._conf.video.quality,
                        handler)

                    eventlog.log_i(f'done: <{base_event.plate_utf8}>, with-video-session:true')
                    return

                except VsSessionAbsentError:
                    eventlog.log_w(f'fail: VsSessionAbsentError (<{base_event.plate_utf8}>)')

                except VsError:
                    eventlog.log_w(f'fail: VsError (<{base_event.plate_utf8}>)')

                except GenericError:
                    eventlog.log_w(f'fail: GenericError (<{base_event.plate_utf8}>)')
                    eventlog.log_w('fail')

                except Exception:
                    eventlog.log_x()

            eventlog.log_i(f'done: <{base_event.plate_utf8}>, with-video-session:false')
            self._handle_evidence_wout_vs(base_event, evidence)
            return

        eventlog.log_i(f'done: skip trackset ({base_event.plate_utf8}, reason:violation-absent)')

    def _handle_evidence_with_vs(self, base: BaseEvent, evidence: Evidence, video_blob: FileBlob | None) -> None:
        eventlog.log_i(f'exec: {base.plate_utf8}, with-video:{video_blob is not None}')
        self._handle_evidence(base, evidence, video_blob)

    def _handle_evidence_wout_vs(self, base: BaseEvent, evidence: Evidence) -> None:
        eventlog.log_i(f'exec: {base.plate_utf8}')
        self._handle_evidence(base, evidence, None)

    def _handle_evidence(self, base: BaseEvent, evidence: Evidence, video_blob: Optional[FileBlob]) -> None:
        eventlog.log_i(f'exec: {base.plate_utf8}, with-video:{video_blob is not None}')

        if video_blob is not None:
            evidence.add_videos([video_blob])

        self._handler(Event(self._conf.identity, base, evidence))


def create(detectors: DetectorGroups, services: ServicePack, infos: List[DetectorUnitInfo]) -> None:
    for info in infos:
        try:
            conf = decode__detector_conf(info.identity, info.conf, info.footer_defvars)
            if conf.rules:
                detector = _Detector(DetectorDeps.create(services), conf)
                detectors.plate_lost.add(detector)

        except GenericError:
            eventlog.log_e(f'fail: {info.identity.ref}')
            raise
