from typing import Dict, Generic, Final, List, Optional, TypeVar, cast

from uuid import uuid4

from edge.types.blob import Blob, MemBlob
from edge.types.identity import Identity

from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.measurement import Measurement
from speedcontrol.events.waypoint.measurement import BestWaypoint


class BaseEvent:
    def __init__(self, measurements: List[Measurement], alarms_jnode: List[Dict] | None):
        self.measurements: Final[List[Measurement]] = measurements
        self.alarms_jnode: Final[List[Dict] | None] = alarms_jnode

    @property
    def best_wp(self) -> Optional[BestWaypoint]:
        waypoints_list = self.measurements[0].waypoints
        best_points = [waypoint for waypoint in waypoints_list if waypoint.name == waypoint.NAME_BEST]
        if len(best_points) != 1:
            return None

        return cast(BestWaypoint, best_points[0])

    def get_blobs(self) -> List[Blob]:
        blobs: List[MemBlob] = []
        for measurement in self.measurements:
            blobs.extend(measurement.get_blobs())

        return blobs

    @property
    def plate_ucode(self) -> Optional[str]:
        waypoints = self.measurements[0].waypoints
        best_waypoints = [waypoint for waypoint in waypoints if waypoint.name == waypoint.NAME_BEST]
        best_waypoint = cast(BestWaypoint, best_waypoints[0])
        return best_waypoint.vehicle.vrp_text_ucode

    @property
    def plate_utf8(self) -> bytes:
        waypoints = self.measurements[0].waypoints
        best_waypoints = [waypoint for waypoint in waypoints if waypoint.name == waypoint.NAME_BEST]
        best_waypoint = cast(BestWaypoint, best_waypoints[0])
        return best_waypoint.vehicle.vrp_text_utf8


TEvidence = TypeVar('TEvidence', bound=BaseEvidence)


class Event(Generic[TEvidence]):
    TYPE_TRANSIT = BaseEvidence.EVENT_TYPE__TRANSIT
    TYPE_PARKING = BaseEvidence.EVENT_TYPE__PARKING
    TYPE_RED_LIGHT = BaseEvidence.EVENT_TYPE__REDLIGHT
    TYPE_WRONG_WAY = BaseEvidence.EVENT_TYPE__WRONGWAY
    TYPE_BUS_LANE = BaseEvidence.EVENT_TYPE__BUSLANE
    TYPE_SPOT_SPEED = BaseEvidence.EVENT_TYPE__SPOTSPEED
    TYPE_WANTED = BaseEvidence.EVENT_TYPE__WANTED
    TYPE_SECTION_SPEED = BaseEvidence.EVENT_TYPE__SECTIONSPEED
    TYPE_SEATBELT = BaseEvidence.EVENT_TYPE__SEATBELT

    def __init__(self, sender: Identity, base_event: BaseEvent, evidence: TEvidence):
        self.uuid: Final[str] = str(uuid4())
        self.sender: Final[Identity] = sender
        self.base_event: Final[BaseEvent] = base_event
        self.evidence: Final[TEvidence] = evidence

    @property
    def type(self) -> str:
        return self.evidence.type

    @property
    def ts(self) -> int:
        return self.evidence.main_ts

    def get_blobs(self) -> List[Blob]:
        blobs = self.base_event.get_blobs()
        blobs.extend(self.evidence.get_blobs())
        return blobs
