from typing import Dict, Final, List, Optional

from edge.types.blob import FileBlob, Blob
from edge.services.types.tz import ITzService
from edge.types.origin import Origin
from speedcontrol.events.jnode.origin import encode__origin
from speedcontrol.events.waypoint import Waypoint


class BaseEvidence:
    EVENT_TYPE__TRANSIT = 'transit'
    EVENT_TYPE__BUSLANE = 'bus-lane'
    EVENT_TYPE__PARKING = 'parking'
    EVENT_TYPE__REDLIGHT = 'red-light'
    EVENT_TYPE__SEATBELT = 'seatbelt'
    EVENT_TYPE__SECTIONSPEED = 'section-speed'
    EVENT_TYPE__SPOTSPEED = 'spot-speed'
    EVENT_TYPE__WANTED = 'wanted'
    EVENT_TYPE__WRONGWAY = 'wrong-way'

    def __init__(self,
                 event_type: str,
                 origin: Origin,
                 main_ts: int,
                 control_ts: Optional[int] = None):

        self._type = event_type
        self.origin: Final[Origin] = origin

        self.main_ts: Final[int] = main_ts
        self.control_ts: Final[int | None] = control_ts

        self._waypoints: List[Waypoint] = []
        self._videos: List[FileBlob] = []

    @property
    def type(self) -> str:
        return self._type

    def add_waypoints(self, waypoints: List[Waypoint]) -> None:
        self._waypoints.extend(waypoints)

    def add_videos(self, videos: List[FileBlob]) -> None:
        self._videos.extend(videos)

    @property
    def videos(self) -> List[Blob]:
        return self._videos

    def get_blobs(self) -> List[Blob]:
        blobs: List[Blob] = []
        blobs.extend(self._videos)

        for waypoint in self._waypoints:
            blobs.extend(waypoint.images)

        return blobs

    @property
    def waypoints(self) -> List[Waypoint]:
        return self._waypoints

    @property
    def min_waypoint(self) -> Waypoint:
        if not self._waypoints:
            raise NotImplementedError

        min_waypoint = self._waypoints[0]
        for waypoint in self._waypoints:
            if waypoint.main_ts < min_waypoint.main_ts:
                min_waypoint = waypoint

        return min_waypoint

    @property
    def max_waypoint(self) -> Waypoint:
        if not self._waypoints:
            raise NotImplementedError

        max_waypoint = self._waypoints[0]
        for waypoint in self._waypoints:
            if waypoint.main_ts > max_waypoint.main_ts:
                max_waypoint = waypoint

        return max_waypoint

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = {
            'type': self._type,
            'origin': encode__origin(self.origin),
            'time': tz_service.map_ts_usec_to_isostr(self.main_ts)
        }
        if self._videos:
            jnode['evidence'] = {
                'measurements': {
                    'videos': [video.serialize(tz_service) for video in self._videos]
                }
            }

        if self.control_ts is not None:
            jnode['control-time'] = tz_service.map_ts_usec_to_isostr(self.control_ts)

        return jnode
