from typing import Dict, Optional

from edge.types.origin import Origin
from edge.services.types.tz import ITzService

from speedcontrol.events.detector import DetectorInfo
from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.jnode.schedule import encode__schedule_info
from speedcontrol.events.waypoint.violation import ViolationWaypoint


class Evidence(BaseEvidence):
    def __init__(self, event_ts: int, origin: Origin, detector_info: DetectorInfo, duration: int):
        super().__init__(BaseEvidence.EVENT_TYPE__BUSLANE, origin, event_ts)
        self.detector = detector_info
        self.duration = duration

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = BaseEvidence.serialize(self, tz_service)

        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector.name,
                'rule': {
                    'name': self.detector.rule.name,
                    'schedule': encode__schedule_info(self.detector.rule.schedule),
                    'ticket': {
                        'name': self.detector.rule.ticket.name
                    },
                    'limits': self.detector.rule.limits
                }
            },
            'measurements': {
                'duration': self.duration,
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if len(self._videos):
            videos_jnode = [x.serialize(tz_service) for x in self._videos]
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode

    @property
    def best_waypoint(self) -> Optional[ViolationWaypoint]:
        bw = [x for x in self.waypoints if x.name == ViolationWaypoint.NAME_VIOLATION_BEST]
        return bw[0] if len(bw) else None
