from typing import Dict, Final, Optional

from edge.services.types.tz import ITzService
from edge.support.speed import map_speed_uint
from edge.types.dir import Dir
from edge.types.jnode.dir import encode_dir_text
from edge.types.origin import Origin

from speedcontrol.events.detector import DetectorInfo
from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.jnode.schedule import encode__schedule_info
from speedcontrol.events.waypoint.violation import ViolationWaypoint


class Evidence(BaseEvidence):
    def __init__(self,
                 detector: DetectorInfo,
                 origin: Origin,
                 event_ts: int,
                 main_speed: int,
                 control_speed: int,
                 direction: Dir,
                 speed_limit: int):

        self.detector: Final[DetectorInfo] = detector

        super().__init__(BaseEvidence.EVENT_TYPE__SPOTSPEED, origin, event_ts)

        self.main_speed: Final[int] = main_speed
        self.control_speed: Final[int] = control_speed
        self.speed_limit: Final[int] = speed_limit
        self.dir: Final[Dir] = direction

    @property
    def best_waypoint(self) -> Optional[ViolationWaypoint]:
        bw = [x for x in self.waypoints if x.name == ViolationWaypoint.NAME_VIOLATION_BEST]
        return bw[0] if len(bw) else None

    def serialize(self, tz_service: ITzService) -> Dict:
        jnode = BaseEvidence.serialize(self, tz_service)
        violation_waypoints_jnode = list()
        best_waypoints_jnode = None
        for waypoint in self._waypoints:
            if waypoint.name == waypoint.NAME_VIOLATION:
                violation_waypoints_jnode.append(waypoint.serialize(tz_service))
            elif waypoint.name == waypoint.NAME_VIOLATION_BEST:
                best_waypoints_jnode = waypoint.serialize(tz_service)

        jnode['evidence'] = {
            'detector': {
                'name': self.detector.name,
                'rule': {
                    'name': self.detector.rule.name,
                    'schedule': encode__schedule_info(self.detector.rule.schedule),
                    'ticket': {
                        'name': self.detector.rule.ticket.name
                    },
                    'limits': self.detector.rule.limits
                }
            },
            'measurements': {
                'speed': {
                  'main': {
                    'measured': map_speed_uint(self.main_speed),
                    'source': 'radar'
                  },
                  'control': [{
                      'measured': map_speed_uint(self.control_speed),
                      'source': 'camera'
                  }],
                  'radar': map_speed_uint(self.main_speed),
                  'camera': map_speed_uint(self.control_speed)
                },
                'overspeed': {
                    'main': {
                        'measured': map_speed_uint(self.main_speed) - self.speed_limit,
                        'source': 'radar'
                    },
                    'control': [{
                        'measured': map_speed_uint(self.control_speed) - self.speed_limit,
                        'source': 'camera'
                    }],
                    'radar': map_speed_uint(self.main_speed) - self.speed_limit,
                    'camera': map_speed_uint(self.control_speed) - self.speed_limit
                },
                'waypoints': {
                    'violation': violation_waypoints_jnode
                }
            }
        }

        if best_waypoints_jnode is not None:
            jnode['evidence']['measurements']['waypoints']['best'] = best_waypoints_jnode

        if self.dir != Dir.CODE_U:
            jnode['evidence']['measurements']['dir'] = encode_dir_text(self.dir)

        if len(self._videos):
            videos_jnode = [x.serialize(tz_service) for x in self._videos]
            jnode['evidence']['measurements']['videos'] = videos_jnode

        return jnode
