from typing import Dict, cast

from edge.services.types.about import IAboutService
from edge.services.types.tz import ITzService
from edge.support.speed import map_speed_uint
from edge.types.blob import Blob
from edge.types.jnode.dir import encode_dir_text

from speedcontrol.events.event import BaseEvent, Event
from speedcontrol.events.evidence import BaseEvidence
from speedcontrol.events.evidence.buslane import Evidence as BusLaneEvidence
from speedcontrol.events.evidence.parking import Evidence as ParkingEvidence
from speedcontrol.events.evidence.redlight import Evidence as RedLightEvidence
from speedcontrol.events.evidence.spotspeed import Evidence as SpotSpeedEvidence
from speedcontrol.events.jnode.geoposition import encode__geoposition
from speedcontrol.events.jnode.measurement import encode__measurement
from speedcontrol.events.vehicle import Vehicle


def encode__base_event(self: BaseEvent, tz_service: ITzService, with_images: bool = True) -> Dict:
    jnode = {
        'measurements': [
            encode__measurement(x, tz_service, with_images) for x in self.measurements
        ]
    }

    if self.alarms_jnode:
        jnode['alarms'] = self.alarms_jnode

    return jnode


def encode__event(self: Event, tz_service: ITzService) -> Dict:
    jnode = {
        'uuid': self.uuid
    }

    with_images = self.evidence.type != BaseEvidence.EVENT_TYPE__SEATBELT

    jnode.update(encode__base_event(self.base_event, tz_service, with_images))
    jnode.update(self.evidence.serialize(tz_service))

    return jnode


def encode__event_ua(self: Event, tz_service: ITzService, about_service: IAboutService) -> Dict:
    base_url = about_service.base_url
    if base_url:
        base_url = f'{base_url}/{self.uuid}/by-ref/'

    base = self.base_event
    bw = base.measurements[0].best_waypoint
    vehicle: Vehicle = bw.vehicle
    ev = self.evidence

    mmr_to_ua_class_map = {
        'CAR': '1',
        'HT': '21',
        'LT': '2',
        'VAN': '11',
        'MINIVAN': '12',
        'CROSSOVER': '13',
        'OFFROAD': '14',
        'BUS': '3',
        'UNKNOWN': '0'
    }

    def ts_to_datetime(ts: int) -> str:
        return tz_service.map_ts_usec_to_isostr(ts)

    direction = vehicle.dir
    jnode = {
        'uuid': self.uuid,
        'datetime':	ts_to_datetime(ev.main_ts),
        'type': ev.type,
        'origin': {
            'uuid': ev.origin.uuid,
            'name': about_service.hw_model,
            'serial': about_service.hw_serial,
            'address': ev.origin.location.place,
            'type': {
                't_install': 1,  # stationary
                't_measure': [1, 3]  # radar + camera
            }
        },
        'vehicle': {
            'plate': {
                'country':	vehicle.vrp_country_num,
                'text': vehicle.vrp_text_ucode
            },
            'params':
            [
                {
                    'key':	'direction',
                    'value': encode_dir_text(direction)
                },
                {
                    'key':	'color',
                    'value': '000'
                },
                {
                    'key':	'type',
                    'value': mmr_to_ua_class_map.get(vehicle.mmr_class, '0')
                },
                {
                    'key':	'mark',
                    'value': vehicle.mmr_make or 'unknown'
                },
                {
                    'key':	'model',
                    'value': vehicle.mmr_model or 'unknown'
                }
            ]
        },
        'limits': [],
        'measurements':	[],
        'media': []
    }

    def append_measurement(key: str, value: str) -> None:
        jnode['measurements'].append({'key': key, 'value': value})

    def append_limits(key: str, value: str) -> None:
        jnode['limits'].append({'key': key, 'value': value})

    def append_blob(blob: Blob, blob_type: str) -> None:
        jnode['media'].append({
            'type': blob_type,
            'baseurl': base_url,
            'name': blob.name,
            'hash': blob.hash,
            'datetime': ts_to_datetime(blob.min_ts)
        })

    # add geolocation if exist
    if ev.origin.geolocation:
        jnode['origin']['location'] = encode__geoposition(ev.origin.geolocation)

    # add limits
    if ev.type == ev.EVENT_TYPE__SPOTSPEED:
        ss_ev = cast(SpotSpeedEvidence, ev)
        append_limits('speed-limit', str(ss_ev.detector.rule.limits['speed-limit']))
        append_limits('max-overspeed', str(ss_ev.detector.rule.limits['max-overspeed']))

    elif ev.type == ev.EVENT_TYPE__REDLIGHT:
        # rl_ev = ev  # type: RedLightEvidence
        append_limits('min-red-time', '0')

    elif ev.type == ev.EVENT_TYPE__PARKING:
        pk_ev = cast(ParkingEvidence, ev)
        append_limits('max-standing-time', str(pk_ev.detector.rule.limits['max-standing-time']))

    # add measurements
    if ev.type == ev.EVENT_TYPE__SPOTSPEED:
        ss_ev = cast(SpotSpeedEvidence, ev)
        append_measurement('speed-main', str(map_speed_uint(ss_ev.main_speed)))
        append_measurement('speed-control', str(map_speed_uint(ss_ev.control_speed)))

    elif ev.type == ev.EVENT_TYPE__REDLIGHT:
        rl_ev = cast(RedLightEvidence, ev)
        halt_span_info = rl_ev.get_ukr_json_ev()
        append_measurement('halt-time', str(halt_span_info.halt_time or 0))
        append_measurement('green-amber-time', str(halt_span_info.green_amber_time or 0))
        append_measurement('red-amber-time', str(halt_span_info.red_amber_time or 0))
        append_measurement('red-time', str(halt_span_info.red_time or 0))

    elif ev.type == ev.EVENT_TYPE__PARKING:
        pk_ev = cast(ParkingEvidence, ev)
        append_measurement('standing-time', str(pk_ev.standing_time))

    # add media
    if ev.type == ev.EVENT_TYPE__SPOTSPEED:
        ss_ev = cast(SpotSpeedEvidence, ev)
        best_blobs = ss_ev.best_waypoint.images
    elif ev.type == ev.EVENT_TYPE__BUSLANE:
        bl_ev = cast(BusLaneEvidence, ev)
        best_blobs = bl_ev.best_waypoint.images
        inside_blobs = bl_ev.waypoints[1].images
        inside_scene = [blob for blob in inside_blobs if blob.profile == 'scene'][0]
        append_blob(inside_scene, 'inside')

    else:
        best_blobs = bw.images

    pl_blob = [blob for blob in best_blobs if blob.profile == 'plate'][0]
    th_blob = [blob for blob in best_blobs if blob.profile == 'thumb'][0]
    vh_blob = [blob for blob in best_blobs if blob.profile == 'vehicle'][0]

    append_blob(pl_blob, 'plate')
    append_blob(th_blob, 'thumb')
    append_blob(vh_blob, 'body')

    ev_wps = ev.waypoints
    near_wp = None
    far_wp = None
    if len(ev_wps) > 1:
        # TODO: find wp with min\max ts instead using index in case of non sorted lists

        if direction == direction.CODE_R:
            if ev_wps[0].main_ts < ev_wps[-1].main_ts:
                near_wp = ev_wps[0]
                far_wp = ev_wps[-1]
            else:
                near_wp = ev_wps[-1]
                far_wp = ev_wps[0]

        elif direction == direction.CODE_A:
            if ev_wps[0].main_ts < ev_wps[-1].main_ts:
                near_wp = ev_wps[-1]
                far_wp = ev_wps[0]
            else:
                near_wp = ev_wps[0]
                far_wp = ev_wps[-1]

    if None not in [near_wp, far_wp]:
        near_full_blob = [blob for blob in near_wp.images if blob.profile == 'scene'][0]
        far_full_blob = [blob for blob in far_wp.images if blob.profile == 'scene'][0]
        append_blob(near_full_blob, 'close')
        append_blob(far_full_blob, 'far')

    video_blobs = ev.videos
    if len(video_blobs):
        append_blob(video_blobs[0], 'video')

    return jnode
