import ctypes
import io
import os
import socket
import tarfile
import ujson

from typing import Dict, List, Optional

from edge import eventlog
from edge.services.types.about import IAboutService
from edge.services.types.systest import ISystestService
from edge.services.types.tz import ITzService
from edge.support.bs import BsError
from edge.support.fs import silent_mkdir_r
from edge.support.queue import JsonFileQueue
from edge.types.blob import MemBlob, Blob, StructBlobMeta
from edge.types.exceptions import BufferSizeError

from speedcontrol.events.event import Event
from speedcontrol.events.jnode.event import encode__event, encode__event_ua
from speedcontrol.services.event.jnode.fact import encode__fact
from speedcontrol.services.types.bs import IBsService
from speedcontrol.services.types.event import EventServiceConf, IEventService
from speedcontrol.services.types.sign import SignError, ISignService
from speedcontrol.services.types.stats import IEventCounters


class _EventService(IEventService):
    def __init__(self, twisted_reactor,
                 counters: IEventCounters,
                 tz_service: ITzService,
                 bs_service: IBsService,
                 systest_service: ISystestService,
                 about_service: IAboutService,
                 sign_service: ISignService,
                 json_q_list: List[JsonFileQueue],
                 args: EventServiceConf):

        IEventService.__init__(self)

        self._reactor = twisted_reactor
        self._counters = counters

        self._tz_service = tz_service
        self._bs_service = bs_service
        self._systest_service = systest_service
        self._about_service = about_service
        self._sign_service = sign_service
        self._json_q_list = json_q_list
        self._devices_jnode = self._get_devices(about_service)

        self._args_ua_required = args.ua_json_required
        self._local_arch_enabled = args.local_arch_enabled
        self._local_arch_base_dir = args.local_arch_base_dir

        self._hash_buffer_size = 100

    @staticmethod
    def _get_devices(about_service: IAboutService) -> List[Dict]:
        device = {'hostname': socket.gethostname()}
        hw = about_service.hw_info
        if any(hw):
            device['hw'] = hw

        sw = about_service.sw_info
        if sw:
            device['sw'] = sw

        sw_name = about_service.sw_name
        if sw_name:
            device['sw-name'] = sw_name

        return [device]

    def _hash_blob(self, blob: Blob) -> None:
        hash_buffer = ctypes.create_string_buffer(b'', size=self._hash_buffer_size)
        hash_size = 0
        try:
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)
        except BufferSizeError as x:
            self._hash_buffer_size = x.min_size
            hash_buffer = ctypes.create_string_buffer(b'', size=self._hash_buffer_size)
            hash_size = self._sign_service.hash_blob(blob.file_data, hash_buffer)

        except SignError:
            eventlog.log_e(f'fail: hash (profile:{blob.profile})')

        if hash_size > 0:
            blob.hash = self._sign_service.hash_type, str(hash_buffer.raw[:hash_size], 'ascii')

    def emit(self, event: Event) -> bool:
        # noinspection PyBroadException
        try:
            # save media
            media_blobs: List[Blob] = event.get_blobs()

            [self._bs_service.save(event.sender, b) for b in media_blobs]
            [self._hash_blob(b) for b in media_blobs]

            # save args.json
            args_jnode: Dict = encode__event(event, self._tz_service)
            args_jnode['origin']['devices'] = self._devices_jnode
            self._systest_service.serialize(args_jnode)

            args_data: bytes = ujson.dumps(args_jnode).encode('utf-8')
            args_blob = MemBlob('args', event.ts, args_data, file_ext='json')

            self._bs_service.save(event.sender, args_blob)
            self._hash_blob(args_blob)

            blobs_to_test = media_blobs + [args_blob]
            if not self._args_ua_required:
                self._emit3(event, blobs_to_test, args_blob, None, media_blobs, args_jnode, False)
                return True

            # save args-ua.json
            args_ua_jnode: Dict = encode__event_ua(event, self._tz_service, self._about_service)
            args_ua_data: bytes = ujson.dumps(args_ua_jnode).encode('utf-8')
            args_ua_blob = MemBlob('args', event.ts, args_ua_data, file_ext='json')

            if event.type not in [Event.TYPE_BUS_LANE,
                                  Event.TYPE_SPOT_SPEED,
                                  Event.TYPE_SECTION_SPEED,
                                  Event.TYPE_WRONG_WAY,
                                  Event.TYPE_RED_LIGHT,
                                  Event.TYPE_PARKING,
                                  Event.TYPE_WANTED]:

                self._emit3(event, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, False)
                return True

            def sign_blob_handler(blob_data: Optional[bytes], sign_data: Optional[bytes]) -> None:
                self._emit2(event, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode,
                            blob_data, sign_data)

            try:
                self._sign_service.sign_blob(args_ua_blob.file_data, sign_blob_handler)

            except SignError:
                sign_blob_handler(None, None)

            except Exception:
                sign_blob_handler(None, None)
                eventlog.log_x()

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')
            return False

        except Exception:
            eventlog.log_x()
            return False

        return True

    def _emit2(
            self,
            event_args: Event,
            blobs_to_test: List[Blob],
            args_blob: MemBlob,
            args_ua_blob: Optional[MemBlob],
            media_blobs: List[Blob],
            args_jnode: Dict,
            blob_data: Optional[bytes],
            sign_data: Optional[bytes]) -> None:

        eventlog.log_d('exec: blob-size:{}, sign-size:{}'.format(
            '' if blob_data is None else len(blob_data),
            '' if sign_data is None else len(sign_data)))

        args_ua_signed = sign_data is not None

        if args_ua_signed:
            if blob_data != args_ua_blob.file_data:
                eventlog.log_w("fail: blob_data != args_ua_blob.file_data")

            crt_data = self._sign_service.cert
            if crt_data:
                args_ua_blob.meta = StructBlobMeta(sign_data, crt_data)
            else:
                args_ua_blob.meta = StructBlobMeta(sign_data)

        self._emit3(event_args, blobs_to_test, args_blob, args_ua_blob, media_blobs, args_jnode, args_ua_signed)

    def _emit3(self,
               event: Event,
               blobs_to_test: List[Blob],
               args_blob: MemBlob,
               args_ua_blob: Optional[MemBlob],
               media_blobs: List[Blob],
               args_jnode: Dict,
               args_ua_signed: bool) -> None:

        def get_tar_path(base_dir: str, ts: int) -> str:
            d = self._tz_service.map_ts_usec_to_datetime_local(ts)
            path = os.path.join(base_dir,
                                str(d.year), str(d.month).zfill(2), str(d.day).zfill(2),
                                str(d.hour).zfill(2), str(d.minute).zfill(2))
            return path

        def export_event_to_gzip() -> None:
            filedir = get_tar_path(self._local_arch_base_dir, event.ts)
            silent_mkdir_r(filedir)
            args_blob_id = args_blob.name.split('_')[1].split(".")[0]
            tar_name = os.path.join(filedir, f'{args_blob_id}.tar')

            with tarfile.open(name=tar_name, mode='w') as tar:
                args_info = tarfile.TarInfo('args.json')
                args_info.size = len(args_blob.file_data)
                tar.addfile(args_info, io.BytesIO(args_blob.file_data))

                if args_ua_blob is not None:
                    args_info = tarfile.TarInfo('args-ua.json')
                    args_info.size = len(args_ua_blob.file_data)
                    tar.addfile(args_info, io.BytesIO(args_ua_blob.file_data))

                for media_blob in media_blobs:
                    if media_blob.file_path is None:
                        tar_info = tarfile.TarInfo(media_blob.name)
                        tar_info.size = len(media_blob.file_data)
                        tar.addfile(tar_info, io.BytesIO(media_blob.file_data))
                    else:
                        tar.add(media_blob.file_path, media_blob.name)

        try:
            if args_ua_blob:
                self._bs_service.save(event.sender, args_ua_blob)
                self._hash_blob(args_ua_blob)

                blobs_to_test.append(args_ua_blob)
                fact_jnode = encode__fact(args_jnode, args_blob.name, args_ua_blob.name, args_ua_signed)

            else:
                fact_jnode = encode__fact(args_jnode, args_blob.name)

            skip_event = False
            eventlog.log_d(f'exec: validate-blobs <{event.uuid}>')
            for blob in blobs_to_test:
                blob_ref = blob.name.split('_')[1].split('.')[0]
                blob_size_tx = os.stat(blob.file_path).st_size if blob.file_path else len(blob.file_data)
                try:
                    blob_size_rx = self._bs_service.get_size(event.sender, blob_ref, blob.profile)
                    if blob_size_rx != blob_size_tx:
                        skip_event = True
                        eventlog.log_w(
                            'fail: validate-blobs <{}> (blob-ref:{}, profile:{}, tx-blob-size:{}, rx-blob-size:{})'
                            .format(event.uuid, hex(int(blob_ref)), blob.profile, blob_size_tx, blob_size_rx))

                except BsError:
                    skip_event = True
                    eventlog.log_w(
                        'fail: validate-blobs <{}> (blob-ref:{}, profile:{}, tx-blob-size:{})'
                        .format(event.uuid, hex(int(blob_ref)), blob.profile, blob_size_tx))

            if skip_event:
                eventlog.log_w(f'fail: validate-blobs <{event.uuid}>')

            else:
                eventlog.log_d(f'done: validate-blobs <{event.uuid}>')
                if self._local_arch_enabled:
                    export_event_to_gzip()

                list(map(lambda json_q: json_q.add_jnode(fact_jnode, args_blob.name), self._json_q_list))
                self._counters.inc_fact()

        except BsError:
            eventlog.log_w('fail: save-blob (skip event)')

        except Exception:
            eventlog.log_x()


def init_service(
        twisted_reactor,
        counters: IEventCounters,
        args: EventServiceConf,
        tz_service: ITzService,
        bs_service: IBsService,
        systest_service: ISystestService,
        about_service: IAboutService,
        sign_service: ISignService) -> IEventService:

    json_q_list = [JsonFileQueue(json_q_args) for json_q_args in args.json_queue_list]
    return _EventService(twisted_reactor, counters, tz_service, bs_service, systest_service,
                         about_service, sign_service, json_q_list, args)
