from collections.abc import Generator
from typing import Final, Callable

from edge import eventlog
from edge.services import cam, lane, plateinfo, systest
from edge.services.types.cam import ICamService
from edge.services.types.jnode.systest import decode__systest_service_conf
from edge.services.types.lane import ILaneService
from edge.services.types.plateinfo import IPlateInfoService
from edge.services.types.stats import IStatsProvider
from edge.services.types.systest import ISystestService
from edge.support.service import ServicePack
from edge.types.dir import Dir
from edge.types.exceptions import GenericError
from edge.types.shape.cuboid import CuboidI
from edge.types.shape.rect import RectI
from edge.types.trend import LrTrend

from tracker.services import scene_emitter
from tracker.services import scene_provider
from tracker.services import timekeeper
from tracker.services import ra_track_provider
from tracker.services import track_emitter
from tracker.services import track_matcher
from tracker.services import lr_track_provider
from tracker.services.stats import TrStatService
from tracker.services.types.scene_emitter import ISceneEmitter
from tracker.services.types.scene_provider import ISceneReactProvider, ISceneReactHandler, SceneReact
from tracker.services.types.timekeeper import ITimekeeper
from tracker.services.types.ra_track_provider import IRaTrackProvider, IRaTrackProviderHandler
from tracker.services.types.track_emitter import ITrackEmitter
from tracker.services.types.track_matcher import ILrRaMatcher
from tracker.services.types.lr_track_provider import ILrTrackProvider, ILrTrackProviderHandler
from tracker.types.kernel import KernelConf, RaCuboidConf
from tracker.types.lr import LrTrackSet, LrSceneItem, LrTrack
from tracker.types.lrra import LrRaSceneItem, LrRaScene, LrRaTrackset
from tracker.types.ra import RaSceneItem, RaTrack


def _get_ra_track_items(ra_items: list[RaSceneItem], conf: RaCuboidConf, camera: ICamService | None) -> list[LrRaSceneItem]:

    tr_items: list[LrRaSceneItem] = []

    # The camera service is always available in the terminal with the radar.
    for ra_item in ra_items:
        ra_f = ra_item.scene_f_cm / 100.0
        ra_l = ra_item.scene_l_cm / 100.0

        # near plane
        x1_1, y1_1 = camera.mk_image_point(ra_l + conf.left_max, conf.up_max, ra_f + conf.forward_min)
        x1_2, y1_2 = camera.mk_image_point(ra_l + conf.left_min, conf.up_max, ra_f + conf.forward_min)
        x1_3, y1_3 = camera.mk_image_point(ra_l + conf.left_min, conf.up_min, ra_f + conf.forward_min)
        x1_4, y1_4 = camera.mk_image_point(ra_l + conf.left_max, conf.up_min, ra_f + conf.forward_min)

        # far plane
        x2_1, y2_1 = camera.mk_image_point(ra_l + conf.left_max, conf.up_max, ra_f + conf.forward_max)
        x2_2, y2_2 = camera.mk_image_point(ra_l + conf.left_min, conf.up_max, ra_f + conf.forward_max)
        x2_3, y2_3 = camera.mk_image_point(ra_l + conf.left_min, conf.up_min, ra_f + conf.forward_max)
        x2_4, y2_4 = camera.mk_image_point(ra_l + conf.left_max, conf.up_min, ra_f + conf.forward_max)

        ra_cuboid = CuboidI((x1_1, y1_1, x1_2, y1_2, x1_3, y1_3, x1_4, y1_4,
                             x2_1, y2_1, x2_2, y2_2, x2_3, y2_3, x2_4, y2_4))

        tr_items.append(LrRaSceneItem(None, ra_item, ra_cuboid))

    return tr_items


def _match_ra_lr_items(lr_items: list[LrSceneItem], ra_items: list[LrRaSceneItem], session_active: bool) -> LrRaScene:
    unused_lr_items: dict[int, LrSceneItem] = dict([(x.oid, x) for x in lr_items])
    unused_ra_items: dict[int, LrRaSceneItem] = dict([(x.ra_item.oid, x) for x in ra_items])

    lrra_items: list[LrRaSceneItem] = []
    for tr_ra_item in ra_items:
        ra_item = tr_ra_item.ra_item
        for lr_item in lr_items:
            if lr_item.dir == ra_item.speed_dir or lr_item.dir == Dir.CODE_U or session_active:
                if lr_item.vrp.text_quad.center in tr_ra_item.ra_cuboid.near_plane:
                    unused_lr_items.pop(lr_item.oid, None)
                    unused_ra_items.pop(ra_item.oid, None)

                    tr_item = LrRaSceneItem(lr_item, ra_item, tr_ra_item.ra_cuboid)
                    lrra_items.append(tr_item)

    return LrRaScene.create(
        lrra_items,
        list(unused_lr_items.values()),
        list(unused_ra_items.values()))


class BaseKernel(IStatsProvider,
                 ISceneReactHandler,
                 ILrTrackProviderHandler,
                 IRaTrackProviderHandler):

    def __init__(self, services: ServicePack, ra_cuboid_conf: RaCuboidConf | None) -> None:
        self._services: Final[ServicePack] = services

        self._lrra_matcher: Final[ILrRaMatcher] = services.get(ILrRaMatcher)
        self._ra_track_provider: Final[IRaTrackProvider] = services.get(IRaTrackProvider)

        self._scene_emitter: Final[ISceneEmitter] = services.get(ISceneEmitter)
        self._track_emitter: Final[ITrackEmitter] = services.get(ITrackEmitter)

        self._ra_cuboid_conf: Final[RaCuboidConf | None] = ra_cuboid_conf
        self._systest = self._services.get(ISystestService).session_active
        self._cam: Final[ICamService | None] = services.try_get(ICamService)

    def destroy(self):
        self._services.destroy()

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        self._services.collect_stats(json_handler)

    def handle_scene_react(self, react: SceneReact) -> None:
        lr_scene = react.lr_scene
        ra_scene = react.ra_scene

        # Update _ra_track_store
        if ra_scene:
            self._ra_track_provider.add_scene(ra_scene.items)

        # Create the tr_scene from an avaliable lr_scene and ra_scene.
        if lr_scene and ra_scene:
            ra_items: list[LrRaSceneItem] = _get_ra_track_items(ra_scene.items, self._ra_cuboid_conf, self._cam)
            lrra_scene: LrRaScene = _match_ra_lr_items(lr_scene, ra_items, self._systest)

            def yield_matches() -> Generator[tuple[int, int], None, None]:
                for tr_item in lrra_scene.lrra_items:
                    lr_oid = tr_item.lr_item.oid
                    ra_oid = tr_item.ra_item.oid
                    yield lr_oid, ra_oid

            self._lrra_matcher.add_matches(lr_scene.sender, yield_matches())

        elif lr_scene:
            lrra_scene = LrRaScene.create_lr(lr_scene)

        elif ra_scene:
            ra_items = _get_ra_track_items(ra_scene.items, self._ra_cuboid_conf, self._cam)
            lrra_scene = LrRaScene.create_ra(ra_items)

        else:
            lrra_scene = LrRaScene.create_empty()

        if react.im_ref:
            self._scene_emitter.consume_scene(react.im_ref, lrra_scene)

    def handle_detected_lr_track(self, lr_trackset: LrTrackSet) -> None:
        # noinspection PyBroadException
        try:
            tr_track = LrRaTrackset(lr_trackset, None)
            self._track_emitter.consume_track(tr_track)
        except GenericError:
            eventlog.log_w('fail')
        except Exception:
            eventlog.log_x()

    def handle_lost_lr_track(self, lr_trackset: LrTrackSet) -> None:
        lr_track: LrTrack = lr_trackset.lr_track
        ra_oids_desc: list[int] = self._lrra_matcher.get_ra_oids_desc(lr_track.source_id, lr_track.dir, lr_track.oids)

        ra_track: RaTrack | None = None
        if ra_oids_desc:
            # Search for time intervals when VRP is in static position
            lost_tout_usec = int(10e6)   # 10 sec
            dead_tout_usec = int(120e6)  #  2 min
            lr_trend = LrTrend(lost_tout_usec, dead_tout_usec)
            lr_trend.extend(lr_trackset.lr_track.plate_trace)

            ra_track = self._ra_track_provider.pop_track(ra_oids_desc, lr_trackset.mmr_class, lr_trend)

        if ra_track is None:
            tr_track = LrRaTrackset(lr_trackset, None)

        else:
            tr_track = LrRaTrackset(lr_trackset, ra_track)

        # noinspection PyBroadException
        try:
            self._track_emitter.consume_track(tr_track)
        except GenericError:
            eventlog.log_w('fail')
        except Exception:
            eventlog.log_x()

    def handle_lost_ra_track(self, ra_track: RaTrack) -> None:
        tr_track = LrRaTrackset(None, ra_track)

        # noinspection PyBroadException
        try:
            self._track_emitter.consume_track(tr_track)
        except GenericError:
            eventlog.log_w('fail')

        except Exception:
            eventlog.log_x()


class Kernel(BaseKernel):
    def __init__(self, twisted_reactor, conf: KernelConf) -> None:
        services = ServicePack()
        stats = TrStatService(twisted_reactor, conf.stats,
                              scene_consumer_count=len(conf.scene_emitter.targets))

        stats.add_providers([self])

        if conf.cam is not None:
            services.add(ICamService, cam.init_service(conf.cam))

        services.add(ITimekeeper, timekeeper.init_service(twisted_reactor, conf.timekeeper))
        services.add(ISystestService, systest.init_service(decode__systest_service_conf()))
        services.add(ILaneService, lane.init_service(conf.lane))
        services.add(IPlateInfoService, plateinfo.init_service(conf.plateinfo))
        services.add(ISceneReactProvider, scene_provider.init_service(twisted_reactor, self, conf.scene_provider, stats.scene_provider))
        services.add(ILrTrackProvider, lr_track_provider.init_service(twisted_reactor, services, self, conf.lr_track_provider))
        services.add(IRaTrackProvider, ra_track_provider.init_service(twisted_reactor, services, self, conf.ra_track_provider))

        services.add(ISceneEmitter, scene_emitter.init_service(conf.scene_emitter, stats.scene_emitter))
        services.add(ITrackEmitter, track_emitter.init_service(services, conf.track_emitter, stats.track_emitter))

        services.add(ILrRaMatcher, track_matcher.init_service(twisted_reactor, conf.lrra_matcher))

        twisted_reactor.addSystemEventTrigger('before', 'shutdown', lambda: self.destroy())

        super().__init__(services, conf.ra_cuboid)

        self._reactor = twisted_reactor

    def run(self):
        self._reactor.run()
