from typing import Final, Callable

from edge import eventlog
from edge.services.react import ReactSource
from edge.services.types.plateinfo import IPlateInfoService
from edge.support.queue import ImageFileQueue
from edge.support.service import ServicePack
from edge.types.exceptions import InternalError, GenericError
from edge.types.imageref import ImageRef
from edge.types.jnode.imageref import decode__image_ref, encode__image_ref
from tracker.services.types.jnode.lr_track_provider import encode__lr_track_provider_stats
from tracker.services.types.lr_track_provider import (
    LrTrackProviderConf, ILrTrackProvider, ILrTrackProviderHandler, LrTrackAccConf,
    LrTrackProviderStats, LrTrackLostAccStats)
from tracker.types.jnode.lr import decode__lr_trackset, decode__lr_track_stat
from tracker.types.lr import LrTrackStat


class _LrReact:
    def __init__(self, jnode: dict) -> None:
        self.jnode: Final[dict] = jnode
        self.lr_jnode: Final[dict] = jnode['Lr']

    @property
    def vrp_text(self) -> str:
        return self.lr_jnode['Plate']['P']

    def get_vrp_template(self, plateinfo_service: IPlateInfoService) -> str | None:
        return plateinfo_service.get_plate_template(self.lr_jnode['Plate']['T'])

    def clone_imref(self) -> ImageRef:
        return decode__image_ref(self.lr_jnode['Image'])

    def set_imref(self, value: ImageRef) -> None:
        self.lr_jnode['Image'] = encode__image_ref(value)

    @property
    def track_stat(self) -> LrTrackStat:
        return decode__lr_track_stat(self.lr_jnode['Stat'])


class _VrpContext:
    def __init__(self, react: _LrReact, rm_delayed_call) -> None:
        self.react: _LrReact = react
        self.rm_delayed_call = rm_delayed_call


class _LrTrackAcc:
    def __init__(self, twisted_reactor, conf: LrTrackAccConf, stats: LrTrackLostAccStats) -> None:
        self._reactor: Final = twisted_reactor
        self._acc_tout_sec: Final[int] = conf.tout_sec
        self._stats: Final[LrTrackLostAccStats] = stats

        self._imq = ImageFileQueue(
            twisted_reactor, None, conf.imq.endpoint,
            conf.imq.min_file_count,
            conf.imq.max_file_count,
            conf.imq.log_level)

        self._vrp_to_context_map: dict[str, _VrpContext] = dict()
        self._handler: Callable[[_LrReact], None] | None = None

    def set_handler(self, handler: Callable[[_LrReact], None]) -> None:
        self._handler = handler

    def consume_track(self, react: _LrReact) -> None:
        """
        :raises RuntimeError
        """
        vrp_text: str = react.vrp_text
        vrp_context: _VrpContext | None = self._vrp_to_context_map.get(vrp_text, None)

        if vrp_context is None:
            imref = react.clone_imref()
            if imref.url is not None:
                try:
                    imq_imref = self._imq.add_image_ref(imref, imref.codec)  # raises RuntimeError
                    react.set_imref(imq_imref)
                except RuntimeError:
                    self._stats.failure_copy_image += 1
                    eventlog.log_e(f'fail: ImageFileQueue::add_image_ref <{imref.url}>')
                    raise InternalError()

            rm_delayed_call = self._reactor.callLater(self._acc_tout_sec, self._handle_trackset_tout, vrp_text)

            self._vrp_to_context_map[vrp_text] = _VrpContext(react, rm_delayed_call)
            self._stats.success_create_count += 1

        else:
            vrp_context.rm_delayed_call.reset(secondsFromNow=self._acc_tout_sec)

            prev_react: _LrReact = vrp_context.react

            # longer track is better
            if len(react.lr_jnode['PlateTrace']) > len(prev_react.lr_jnode['PlateTrace']):
                best_react: _LrReact = react
            else:
                best_react: _LrReact = vrp_context.react

            best_imref = best_react.clone_imref()
            if best_imref.url is not None:
                try:
                    imq_imref = self._imq.add_image_ref(best_imref, best_imref.codec)  # raises RuntimeError
                    self._imq.del_image(prev_react.clone_imref())
                    best_react.set_imref(imq_imref)
                except RuntimeError:
                    self._stats.failure_copy_image += 1
                    eventlog.log_e(f'fail: ImageFileQueue::add_image_ref <{best_imref.url}>')
                    raise InternalError()

            # vrp_context.lr_trackset always appears earlier than lr_trackset
            best_react.lr_jnode['PlateTrace'] = prev_react.lr_jnode['PlateTrace'] + react.lr_jnode['PlateTrace']
            best_react.lr_jnode['Id'] = list(set(prev_react.lr_jnode['Id'] + react.lr_jnode['Id']))

            vrp_context.react = best_react
            vrp_context.rm_delayed_call.reset(secondsFromNow=self._acc_tout_sec)
            self._stats.success_append_count += 1

    def _handle_trackset_tout(self, vrp_text: str) -> None:
        assert self._handler is not None

        vrp_context: _VrpContext | None = self._vrp_to_context_map.get(vrp_text, None)
        if vrp_context:
            try:
                self._handler(vrp_context.react)
            finally:
                self._vrp_to_context_map.pop(vrp_text, None)

        else:
            eventlog.log_w(f'fail: vrp-context:null, vrp-text:{vrp_text.encode("utf-8")}')


class _LrTrackProvider(ILrTrackProvider):
    def __init__(self, twisted_reactor,
                 handler: ILrTrackProviderHandler,
                 conf: LrTrackProviderConf,
                 plateinfo: IPlateInfoService) -> None:

        self._handler: Final[ILrTrackProviderHandler] = handler
        self._stats: Final[LrTrackProviderStats] = LrTrackProviderStats(conf.mode)

        self._conf_mode: Final[set[LrTrackStat]] = conf.mode
        self._conf_track_filter = conf.track_filter
        self._plateinfo: Final[IPlateInfoService] = plateinfo

        self._source = ReactSource(twisted_reactor)
        self._source.set_handler(self._handle_react)
        self._source.set_endpoints({conf.react_dir})

        self._lost_track_acc: Final[_LrTrackAcc] = _LrTrackAcc(twisted_reactor, conf.lost_track_acc, self._stats.acc_lost)
        self._lost_track_acc.set_handler(self._emit_lost_track)

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__lr_track_provider_stats(self._stats))

    # noinspection PyUnusedLocal
    def _handle_react(self, react_name: str, react_jnode: dict) -> None:
        react = _LrReact(react_jnode)

        vrp_template = react.get_vrp_template(self._plateinfo)
        for filter_rule in self._conf_track_filter.rules:
            if vrp_template in filter_rule.vrp_templates:
                self._stats.consume.filtered_count += 1
                return

        # noinspection PyBroadException
        try:
            track_stat = react.track_stat
            if track_stat == LrTrackStat.LOST:
                self._stats.consume.lost_count += 1
                if track_stat in self._conf_mode:
                    self._lost_track_acc.consume_track(react)

            elif track_stat == LrTrackStat.DETECTED:
                self._stats.consume.detected_count += 1
                if track_stat in self._conf_mode:
                    self._emit_detected_track(react)

        except GenericError:
            pass
        except Exception:
            eventlog.log_x()

    def _emit_detected_track(self, react: _LrReact) -> None:
        # noinspection PyBroadException
        try:
            lr_trackset = decode__lr_trackset(react.jnode, self._plateinfo)
            self._handler.handle_detected_lr_track(lr_trackset)
            self._stats.produce.detected_count += 1
        except Exception:
            eventlog.log_x()

    def _emit_lost_track(self, react: _LrReact) -> None:
        # noinspection PyBroadException
        try:
            lr_trackset = decode__lr_trackset(react.jnode, self._plateinfo)
            self._handler.handle_lost_lr_track(lr_trackset)
            self._stats.produce.lost_count += 1
        except Exception:
            eventlog.log_x()


def init_service(twisted_reactor,
                 services: ServicePack,
                 handler: ILrTrackProviderHandler,
                 conf: LrTrackProviderConf) -> ILrTrackProvider:

    plateinfo = services.get(IPlateInfoService)
    return _LrTrackProvider(twisted_reactor, handler, conf, plateinfo)
