from collections import defaultdict
from typing import Final, Callable

from edge import eventlog
from edge.services.types.systest import ISystestService
from edge.support.service import ServicePack
from edge.types.trend import LrTrend, LrTrendItem
from tracker.services.types.jnode.ra_track_provider import encode__ra_track_provider_stats
from tracker.services.types.ra_track_provider import IRaTrackProvider, RaTrackProviderConf, IRaTrackProviderHandler, \
    RaTrackProviderStats
from tracker.types.ra import RaTrack, RaSceneItem


def _debouce_ra_items_pos(ra_track: RaTrack, lr_trend: LrTrend) -> RaTrack:
    """
    The function does not change the number of items in the ra_track.
    """

    # Grouping of ra-items by time intervals, when VRP is in static position
    spanref_to_ra_items_map: dict[int, list[RaSceneItem]] = defaultdict(list[RaSceneItem])
    ra_items: list[RaSceneItem] = []

    for ra_item in ra_track.trace:
        spanref: int | None = None
        for trend_item in lr_trend.items:  # type: LrTrendItem
            if ra_item.frame_ts in trend_item.range:
                spanref = trend_item.range.min_ts
                spanref_to_ra_items_map[spanref].append(ra_item)
                break

        if spanref is None:
            ra_items.append(ra_item)

    def _get_median_fl(items: list[RaSceneItem]) -> tuple[float, float]:
        items.sort(key=lambda x: x.scene_f_cm)
        median_f = items[len(items) // 2].scene_f_cm

        items.sort(key=lambda x: x.scene_l_cm)
        median_l = items[len(items) // 2].scene_l_cm

        return median_f, median_l

    # Debounce the position of the ra-items in each group.
    # In each group the ra-items must be static, because the VRP is static
    for span_items in spanref_to_ra_items_map.values():
        median_fl: tuple[float, float] = _get_median_fl(span_items)
        for ra_item in span_items:  # type: RaSceneItem
            ra_item.scene_f_cm, ra_item.scene_l_cm = median_fl
            ra_items.append(ra_item)

    ra_items.sort(key=lambda x: x.frame_ts)
    return RaTrack.create(ra_items)


class _RaTrackContext:
    def __init__(self, ra_track: RaTrack, rm_delayed_call) -> None:
        self.ra_track: Final[RaTrack] = ra_track
        self.rm_delayed_call = rm_delayed_call
        self.rm_request = False


class _RaTrackProvider(IRaTrackProvider):
    def __init__(self, twisted_reactor,
                 handler: IRaTrackProviderHandler,
                 conf: RaTrackProviderConf,
                 systest: bool) -> None:

        self._reactor = twisted_reactor
        self._handler: Final[IRaTrackProviderHandler] = handler
        self._conf: Final[RaTrackProviderConf] = conf
        self._systest: Final[bool] = systest
        self._stats = RaTrackProviderStats()

        self._oid_to_context_map: dict[int, _RaTrackContext] = {}

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        json_handler(encode__ra_track_provider_stats(self._stats))

    def add_scene(self, items: list[RaSceneItem]) -> None:
        for item in items:
            context: _RaTrackContext | None = self._oid_to_context_map.get(item.oid, None)
            if context is None:
                # Create new ra_track.
                ra_track = RaTrack(item.oid, item)

                # Schedule to remove ra_track after _tout_sec since the last update.
                rm_delayed_call = self._reactor.callLater(
                    self._conf.tout_sec, self._del_track, item.oid)

                self._oid_to_context_map[item.oid] = _RaTrackContext(ra_track, rm_delayed_call)
                self._stats.add_track.count += 1

            else:
                # Update existing ra_track
                context.rm_delayed_call.reset(self._conf.tout_sec)
                context.ra_track.append(item)

    def _del_track(self, ra_oid) -> None:
        context: _RaTrackContext | None = self._oid_to_context_map.get(ra_oid, None)
        if context and not context.rm_request:
            ra_track = context.ra_track
            if len(ra_track) >= self._conf.min_track_len:
                self._stats.del_track.success_count += 1
                if not self._conf.require_plate:
                    self._handler.handle_lost_ra_track(ra_track)

            else:
                self._stats.del_track.failure_too_short_count += 1

            eventlog.log_i('done: oid:{0}, items:{1}, speed-n:{2}, forward-m:({3}, {4})'.format(
                ra_track.oid, len(ra_track.trace), ra_track.speed_n_kmh,
                ra_track.trace[0].scene_f_cm / 100.0,
                ra_track.trace[-1].scene_f_cm / 100.0))

        self._oid_to_context_map.pop(ra_oid, None)

    def pop_track(self, ra_oids_desc: list[int], mmr_class: str, lr_trend: LrTrend) -> RaTrack | None:
        for ra_oid in ra_oids_desc:
            context: _RaTrackContext | None = self._oid_to_context_map.get(ra_oid, None)
            if context is None or context.rm_request:
                self._stats.pop_track.failure_not_found_count += 1
                continue

            if self._systest:
                context.rm_delayed_call.cancel()
                self._oid_to_context_map.pop(ra_oid, None)

            else:
                # ra_track will be removed automaticaly after _tout_sec timeout since the last update.
                # rm_request flag indicates that it's necessary to hide ra_track presence in the store for its clients.
                context.rm_request = True

            ra_track = context.ra_track

            for ra_filter_rule in self._conf.ratrack_filter.rules:
                if mmr_class in ra_filter_rule.mmr_classes:
                    if ra_track.speed_n_kmh > ra_filter_rule.max_speed:
                        self._stats.pop_track.failure_mmr_speed_limit_count += 1
                        return

            if len(ra_track) >= self._conf.min_track_len:
                self._stats.pop_track.success_count += 1
                return _debouce_ra_items_pos(ra_track, lr_trend)

            else:
                self._stats.pop_track.failure_too_short_count += 1
                return


def init_service(twisted_reactor,
                 services: ServicePack,
                 handler: IRaTrackProviderHandler,
                 conf: RaTrackProviderConf) -> IRaTrackProvider:

    systest = services.get(ISystestService).session_active
    return _RaTrackProvider(twisted_reactor, handler, conf, systest)
