import os
import time
from typing import Final, Callable

from edge import eventlog
from edge.types.timespan import TimeSpan
from tracker.services.types.timekeeper import ITimekeeper, ChangelogConf, ChronycConf, TimekeeperConf

from twisted.internet import protocol, task
from twisted.python.failure import Failure


class _Changelog:
    def __init__(self, conf: ChangelogConf) -> None:
        self._conf: Final[ChangelogConf] = conf
        self._spans_asc = list[TimeSpan]()

    def _cleanup(self, time_usec: int) -> None:
        if len(self._spans_asc) > self._conf.max_depth_spans:
            self._spans_asc = self._spans_asc[-self._conf.max_depth_spans:]

        next_min_ts = time_usec - self._conf.max_depth_usec

        next_spans_asc = list[TimeSpan]()
        for record in self._spans_asc:
            if record.max_ts < next_min_ts:
                pass
            elif next_min_ts in record:
                next_spans_asc.append(TimeSpan(next_min_ts, record.max_ts))
            else:
                next_spans_asc.append(record)

        self._spans_asc = next_spans_asc

    def add_record(self, time_usec: int, time_sync: bool) -> None:
        if time_sync:
            newest_timespan: TimeSpan | None = self._spans_asc[-1] if self._spans_asc else None
            if newest_timespan is None:
                self._spans_asc.append(TimeSpan(time_usec, time_usec))
            elif time_usec > newest_timespan.max_ts + self._conf.span_tout_usec:
                self._spans_asc.append(TimeSpan(time_usec, time_usec))
            else:
                newest_timespan.max_ts = time_usec

        self._cleanup(time_usec)

    def test_timesync(self, min_ts: int, max_ts) -> bool:
        for record in reversed(self._spans_asc):
            if min_ts in record and max_ts in record:
                return True

        return False


class _ChonycProtocol(protocol.ProcessProtocol):
    def __init__(self, changelog: _Changelog) -> None:
        self._changelog: Final[_Changelog] = changelog

    def connectionMade(self) -> None:
        """
        Called when the child process is started.
        """
        pass

    def outReceived(self, data: bytes) -> None:
        """
        Called when some data is received from stdout.
        """
        #
        # The 'correction' indicates the adjustment made to the system clock to bring it into alignment
        # with the reference time source. If this value is very close to zero, it means that the system clock
        # was already very close to the reference time source.
        #
        # The 'skew-ppm' refers to the rate at which the system clock is drifting away
        # from the reference time source over time.
        #   drift-sec-per-day = (skew-ppm / 1000000usec) * (60sec * 60min * 24hours)
        #   drift-sec-per-day = (skew-ppm / 1e6) * 86400
        #

        # eventlog.log_i(f'done: data:<{data.decode()}>')
        pass

    def errReceived(self, data: bytes) -> None:
        """
        Called when some data is received from stderr.
        """
        eventlog.log_i(f'done: data:<{data.decode()}>')

    def processExited(self, reason: Failure) -> None:
        """
        Called when the child process exits.
        """
        eventlog.log_d(f'done: reason:{reason.value.exitCode}')
        sys_utc_usec: int = int(time.time() * 1e6)
        self._changelog.add_record(sys_utc_usec, reason.value.exitCode == 0)

    def processEnded(self, reason: Failure) -> None:
        """
        Called when the child process exits and all file descriptors
        associated with it have been closed.
        """
        pass


def _spawn_chronyc(twisted_reactor, changelog: _Changelog, conf: ChronycConf) -> None:
    process_protocol = _ChonycProtocol(changelog)

    cli_args = [
        '/usr/bin/chronyc', '-h', '127.0.0.1,::1', 'waitsync',
        f'{conf.attempt_max_count}',
        f'{conf.max_correction_sec}',
        f'{conf.max_skew_ppm}',
        f'{conf.attempt_interval_sec}'
    ]

    twisted_reactor.spawnProcess(process_protocol, cli_args[0], cli_args, env=os.environ)


class _ChronyTimekeeper(ITimekeeper):
    def __init__(self, twisted_reactor, conf: TimekeeperConf) -> None:
        self._changelog = _Changelog(conf.changelog)
        self._looping_call = task.LoopingCall(_spawn_chronyc, twisted_reactor, self._changelog, conf.chronyc)
        self._looping_call.start(1, now=False)

    def test_timesync(self, min_ts: int, max_ts) -> bool:
        return self._changelog.test_timesync(min_ts, max_ts)

    def collect_stats(self, json_handler: Callable[[dict], None]) -> None:
        pass

    def destroy(self) -> None:
        self._looping_call.stop()


class _StubTimekeeper(ITimekeeper):
    def test_timesync(self, min_ts: int, max_ts) -> bool:
        return True


def init_service(twisted_reactor, conf: TimekeeperConf | None) -> ITimekeeper:
    return _StubTimekeeper() if conf is None else _ChronyTimekeeper(twisted_reactor, conf)
