import os
from errno import ENOENT
from typing import Final, Self

from edge import eventlog
from edge.services.types.cam import ICamService
from edge.services.types.lane import ILaneService
from edge.services.types.systest import ISystestService
from edge.support import fs
from edge.support.dir import get_dir_track_lr
from edge.support.service import ServicePack
from edge.support.speed import assert_speed
from edge.types.dir import Dir
from edge.types.exceptions import ParamsError, NotFoundError
from edge.types.shape.point import PointI, PointF
from edge.types.track import DirTrackItem, LaneTrack, DirTrack
from edge.types.trackset.spotspeed import SpotSpeeds, WimSpotSpeed, CameraSpotSpeed, RadarSpotSpeed

from tracker.services.stats import TrackEventCounters
from tracker.services.types.timekeeper import ITimekeeper
from tracker.services.types.track_emitter import ITrackEmitter, TrackEmitterConf
from tracker.types.jnode.lrra import encode__trackset
from tracker.types.lrra import LrRaTrackset
from tracker.types.ra import RaTrack
from tracker.types.lr import LrTrack


def create_lane_track_lrra(service: ILaneService, lrra_track: LrRaTrackset) -> LaneTrack | None:

    if lrra_track.lr_trackset is not None:
        point_provider = ((item.ts, PointI(*item.plate_quad.center))
                          for item in lrra_track.lr_trackset.lr_track.plate_trace)
        return LaneTrack(*service.map_points_abs(point_provider))

    if lrra_track.ra_track is not None:
        point_provider = ((item.frame_ts, PointF(item.frame_l_rel, item.frame_t_rel))
                          for item in lrra_track.ra_track.trace)
        return LaneTrack(*service.map_points_rel(point_provider))


def create_dir_track_lrra(ra_track: RaTrack | None, lr_track: LrTrack | None) -> DirTrack | None:
    if ra_track is not None:
        dir_items = [
            DirTrackItem(item.frame_ts, item.speed_dir)
            for item in ra_track.trace if item.speed_dir in (Dir.CODE_A, Dir.CODE_R)
        ]
        if dir_items:
            return DirTrack(dir_items, ra_track.dir_2)

    if lr_track is not None:
        dir_items = get_dir_track_lr(lr_track.plate_trace, lr_track.dir)
        if dir_items:
            return DirTrack(dir_items, lr_track.dir)


class _Deps:
    def __init__(self,
                 timekeeper: ITimekeeper,
                 lane_service: ILaneService,
                 cam_service: ICamService | None,
                 systest_service: ISystestService | None) -> None:

        self.timekeeper: Final[ITimekeeper] = timekeeper
        self.lane: Final[ILaneService] = lane_service
        self.cam: Final[ICamService | None] = cam_service
        self.systest: Final[ISystestService | None] = systest_service

    @classmethod
    def create(cls, services: ServicePack) -> Self:
        return cls(
            services.get(ITimekeeper),
            services.get(ILaneService),
            services.try_get(ICamService),
            services.try_get(ISystestService))


class TrackEmitter(ITrackEmitter):
    def __init__(self, deps: _Deps, conf: TrackEmitterConf, stats: TrackEventCounters):

        self._deps: Final[_Deps] = deps
        self._stats = stats

        self._event_dirs = conf.event_dirs
        for event_dirpath in self._event_dirs:
            fs.silent_mkdir_r(event_dirpath)

        self._event_file_prefix: Final[int] = os.getpid()
        self._event_file_seq: int = 0

    def mk_spotspeeds_wim_camera2(self, trackset: LrRaTrackset, trackset_dir: Dir) -> SpotSpeeds:
        """
        :raises NotFoundError:
            The trackset does not contain the necessary data.
        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        # wim + camera_2
        if trackset.lr_trackset is None:
            raise NotFoundError

        lr_track: LrTrack = trackset.lr_trackset.lr_track
        vrp_text_utf8: bytes = lr_track.plate_text.encode('utf-8')

        if trackset.lr_trackset.wa_jnode is None:
            raise NotFoundError

        wim_speed: int | None = trackset.lr_trackset.wa_jnode.get('speed')
        if wim_speed is None:
            raise NotFoundError

        cam_speed_2 = self._deps.cam.assert_cam_speed_2(
            vrp_text_utf8, wim_speed,
            lr_track.plate_trace,
            trackset_dir,
            main_dir_u=True)

        return SpotSpeeds(
            main_speed=WimSpotSpeed(wim_speed),
            control_speed=CameraSpotSpeed(cam_speed_2),
            cam_speed_2=cam_speed_2,
            wim_speed=wim_speed)

    def mk_spotspeeds_radarn_systest(self, trackset: LrRaTrackset) -> SpotSpeeds:
        """
        :raises NotFoundError:
            The trackset does not contain the necessary data.
        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        # radar_n + systest
        if not self._deps.systest.session_active:
            raise NotFoundError

        if trackset.ra_track is None:
            raise NotFoundError

        ra_speed_n: int = trackset.ra_track.speed_n_kmh
        ra_speed_2: int | None = trackset.ra_track.speed_2_kmh
        return SpotSpeeds(main_speed=RadarSpotSpeed(ra_speed_n), ra_speed_2=ra_speed_2, ra_speed_n=ra_speed_n)

    def mk_spotspeeds_radarn_camera2(self, trackset: LrRaTrackset, trackset_dir: Dir) -> SpotSpeeds:
        """
        :raises NotFoundError:
            The trackset does not contain the necessary data.
        :raises EstimateError:
            Failed to estimate the control speed.
        :raises ValidateError:
            Failed to validate the speed.
        """
        if trackset.lr_trackset is None:
            raise NotFoundError

        if trackset.ra_track is None:
            raise NotFoundError

        lr_track: LrTrack = trackset.lr_trackset.lr_track
        vrp_text_utf8: bytes = lr_track.plate_text.encode('utf-8')

        ra_speed_n: int = trackset.ra_track.speed_n_kmh
        ra_speed_2: int | None = trackset.ra_track.speed_2_kmh

        # radar_n + radar_2
        assert_speed(ra_speed_n, ra_speed_2)

        # radar_n + camera_2
        cam_speed_2: int = self._deps.cam.assert_cam_speed_2(vrp_text_utf8, ra_speed_n, lr_track.plate_trace, trackset_dir)
        cam_speed_n: int | None = trackset.lr_trackset.lr_jnode.get('SpotSpeed') or None
        return SpotSpeeds(
            main_speed=RadarSpotSpeed(ra_speed_n),
            control_speed=CameraSpotSpeed(cam_speed_2),
            ra_speed_2=ra_speed_2,
            ra_speed_n=ra_speed_n,
            cam_speed_2=cam_speed_2,
            cam_speed_n=cam_speed_n)

    def mk_spotspeeds(self, trackset: LrRaTrackset, trackset_dir: Dir) -> SpotSpeeds | None:
        """
        The main speed measurement channel is defined automatically.
        """
        if trackset_dir == Dir.CODE_U:
            eventlog.log_d('fail: trackset-dir:unknown')
            return

        try:
            return self.mk_spotspeeds_wim_camera2(trackset, trackset_dir)
        except NotFoundError:
            pass
        except ParamsError:
            return

        try:
            return self.mk_spotspeeds_radarn_systest(trackset)
        except NotFoundError:
            pass
        except ParamsError:
            return

        try:
            return self.mk_spotspeeds_radarn_camera2(trackset, trackset_dir)
        except NotFoundError:
            pass
        except ParamsError:
            return

    def consume_track(self, lrra_track: LrRaTrackset) -> None:
        self._stats.inc_track_count(lrra_track)

        # noinspection PyBroadException
        try:
            ra_track: RaTrack | None = lrra_track.ra_track
            lr_track: LrTrack | None = lrra_track.lr_trackset and lrra_track.lr_trackset.lr_track
            dir_track: DirTrack | None = create_dir_track_lrra(lrra_track.ra_track, lr_track)

            if lr_track:
                min_ts = lr_track.plate_trace[0].ts
                max_ts = lr_track.plate_trace[-1].ts
                time_sync = self._deps.timekeeper.test_timesync(min_ts, max_ts)
            elif ra_track:
                min_ts = ra_track.trace[0].frame_ts
                max_ts = ra_track.trace[-1].frame_ts
                time_sync = self._deps.timekeeper.test_timesync(min_ts, max_ts)
            else:
                time_sync = False

            lane_track: LaneTrack | None = create_lane_track_lrra(self._deps.lane, lrra_track)
            spotspeeds = self.mk_spotspeeds(lrra_track, dir_track.dir if dir_track else Dir.CODE_U)
            trackset = encode__trackset(lrra_track.lr_trackset, ra_track,
                                        lane_track, dir_track, spotspeeds, time_sync)

            self._emit(trackset)

        except Exception:
            eventlog.log_x()

    def _emit(self, trackset: str) -> None:
        event_filename = f'{self._event_file_prefix}-{self._event_file_seq:d}'
        self._event_file_seq += 1

        for event_dir in self._event_dirs:
            try_count = 2
            track_dir_present = True
            event_abspath = os.path.join(event_dir, event_filename)
            while try_count > 0 and track_dir_present:
                try:
                    with open(event_abspath, 'w') as event_file:
                        event_file.write(trackset)

                    break

                except IOError as e:
                    if e.errno == ENOENT:
                        track_dir_present = False

                if not track_dir_present:
                    fs.silent_mkdir_r(event_dir)
                    track_dir_present = True
                try_count -= 1


def init_service(services: ServicePack, conf: TrackEmitterConf, stats: TrackEventCounters) -> ITrackEmitter:
    return TrackEmitter(_Deps.create(services), conf, stats)
