import os
import ujson

from typing import Final

from edge import eventlog
from edge.services.types.cam import CamServiceConf
from edge.services.types.jnode.cam import decode__cam_service_conf
from edge.services.types.jnode.plate import decode__plate_info_service_conf
from edge.services.types.jnode.stats import decode__stats_service_conf
from edge.services.types.lane import LaneServiceConf
from edge.services.types.plateinfo import PlateInfoServiceConf
from edge.services.types.stats import StatsServiceConf
from edge.types.exceptions import ParamsError

from tracker.services.types.jnode.lane import decode__road_conf_file
from tracker.services.types.jnode.ra_track_provider import decode__ra_track_provider_conf
from tracker.services.types.jnode.scene_emitter import decode__scene_emitter_conf
from tracker.services.types.jnode.timekeeper import decode__timekeeper_conf
from tracker.services.types.jnode.track_emitter import decode__track_emitter_conf
from tracker.services.types.jnode.lr_track_provider import decode__lr_track_provider_conf
from tracker.services.types.scene_emitter import SceneEmitterConf
from tracker.services.types.scene_provider import SceneProviderConf
from tracker.services.types.jnode.scene_provider import decode__scene_provider_conf
from tracker.services.types.timekeeper import TimekeeperConf
from tracker.services.types.ra_track_provider import RaTrackProviderConf
from tracker.services.types.track_emitter import TrackEmitterConf
from tracker.services.types.track_matcher import LrRaMatcherConf
from tracker.services.types.lr_track_provider import LrTrackProviderConf


class RaCuboidConf:
    def __init__(self,
                 forward_min: float,
                 forward_max: float,
                 left_min: float,
                 left_max: float,
                 up_min: float,
                 up_max: float) -> None:

        self.forward_min: Final[float] = forward_min
        self.forward_max: Final[float] = forward_max
        self.left_min: Final[float] = left_min
        self.left_max: Final[float] = left_max
        self.up_min: Final[float] = up_min
        self.up_max: Final[float] = up_max


def decode__ra_cuboid_conf(jnode: dict) -> RaCuboidConf:
    """
    :param jnode:
        main.conf#/ra-vehicle-ltrb
    """
    forward_jnode = jnode.get('forward')
    if forward_jnode is None or len(forward_jnode) != 2:
        eventlog.log_e_param('#/forward')
        raise ParamsError

    left_jnode = jnode.get('left')
    if left_jnode is None or len(left_jnode) != 2:
        eventlog.log_e_param('#/left')
        raise ParamsError

    up_jnode = jnode.get('up')
    if up_jnode is None or len(up_jnode) != 2:
        eventlog.log_e_param('#/up')
        raise ParamsError

    return RaCuboidConf(forward_jnode[0], forward_jnode[1], left_jnode[0], left_jnode[1], up_jnode[0], up_jnode[1])


class KernelConf:
    def __init__(self, stats_service: StatsServiceConf,
                 scene_provider: SceneProviderConf,
                 track_provider: LrTrackProviderConf,
                 ra_track_provider: RaTrackProviderConf,
                 scene_emitter: SceneEmitterConf,
                 track_emitter: TrackEmitterConf,
                 lane_service: LaneServiceConf | None,
                 plateinfo_service: PlateInfoServiceConf,
                 ra_cuboid_conf: RaCuboidConf | None = None,
                 cam_service: CamServiceConf | None = None,
                 timekeeper_conf: TimekeeperConf | None = None) -> None:

        self.stats: Final[StatsServiceConf] = stats_service

        self.scene_provider = scene_provider
        self.ra_track_provider: Final[RaTrackProviderConf] = ra_track_provider
        self.lr_track_provider: Final[LrTrackProviderConf] = track_provider

        self.scene_emitter = scene_emitter
        self.track_emitter = track_emitter

        self.lrra_matcher = LrRaMatcherConf()
        self.ra_cuboid: Final[RaCuboidConf | None] = ra_cuboid_conf
        self.cam: Final[CamServiceConf | None] = cam_service
        self.lane: Final[LaneServiceConf | None] = lane_service
        self.plateinfo: Final[PlateInfoServiceConf] = plateinfo_service
        self.timekeeper = timekeeper_conf


def decode__kernel_conf(main_conf_jnode: dict) -> KernelConf:
    """
    :param main_conf_jnode:
        main.conf#/
    """
    services_jnode = main_conf_jnode['services']

    scene_provider_conf = decode__scene_provider_conf(services_jnode['scene-provider'])
    lr_track_provider_conf = decode__lr_track_provider_conf(services_jnode['lr-track-provider'])
    ra_track_provider_conf = decode__ra_track_provider_conf(services_jnode['ra-track-provider'])

    scene_emitter_conf = decode__scene_emitter_conf(services_jnode['scene-emitter'])
    track_emitter_conf = decode__track_emitter_conf(services_jnode['track-emitter'])

    lane_service_conf = decode__road_conf_file(services_jnode.get('road.conf'))
    plateinfo_service_conf = decode__plate_info_service_conf(services_jnode.get('plate-info'))

    ra_cuboid_jnode = main_conf_jnode.get('ra-cuboid')
    ra_cuboid_conf: RaCuboidConf | None = None
    if ra_cuboid_jnode:
        try:
            ra_cuboid_conf = decode__ra_cuboid_conf(ra_cuboid_jnode)
        except ParamsError:
            eventlog.log_e_param('#/ra-vehicle-ltrb')
            raise ValueError()

    cam_conf: CamServiceConf | None = None
    if ra_cuboid_conf is not None:
        cam_jnode = services_jnode.get('cam')
        if cam_jnode is None:
            eventlog.log_e_param('#/services/cam')
            raise ValueError()

        # radar terminal
        cam_conf = decode__cam_service_conf(cam_jnode)

    timekeeper_conf: TimekeeperConf | None = None
    timekeeper_type = os.environ.get('EDGE_TIMEKEEPER__TYPE')
    if timekeeper_type == 'chrony':
        time_path = services_jnode['time.conf']
        with open(time_path) as jnode:
            timekeeper_jnode = ujson.load(jnode)

        timekeeper_pri_jnode = timekeeper_jnode['ro'].get('timekeeper')
        if timekeeper_pri_jnode is not None:
            timekeeper_conf = decode__timekeeper_conf(timekeeper_pri_jnode)

    # define stat-counters-args
    monitoring_jnode = services_jnode.get('stat')
    if not monitoring_jnode:
        raise ValueError('#/services/stat')

    stats = decode__stats_service_conf('tracker~stats', 'tracker.0~stats.0', monitoring_jnode)

    return KernelConf(
        stats, scene_provider_conf, lr_track_provider_conf, ra_track_provider_conf, scene_emitter_conf, track_emitter_conf,
        lane_service_conf, plateinfo_service_conf, ra_cuboid_conf, cam_conf, timekeeper_conf)
