from enum import Enum
from typing import Final

from edge.types.dir import Dir
from edge.types.imageref import ImageRef
from edge.types.jnode.imageref import decode__image_ref
from edge.types.shape.quad import QuadI
from edge.types.shape.rect import RectI
from edge.types.track import LrTrackItem


class LrVrp:
    """
    Vehicle registration plate
    """
    def __init__(self, text_ucode: str, text_quad: QuadI, vehicle_rect: RectI | None) -> None:
        self.text_ucode: Final[str] = text_ucode
        self.text_quad: Final[QuadI] = text_quad
        self.vehicle_rect: RectI | None = vehicle_rect


class LrSceneItem:
    def __init__(self, item_ts: int, item_oid: int, item_vrp: LrVrp, item_dir: Dir, item_trace: list[int]) -> None:
        """
        :param item_ts:
            The tracking object timestamp.
        :param item_oid:
            The tracking object ID.
        :param item_vrp:
            The vehicle registration plate.
        :param item_dir:
            The direction of the vehicle's movement relative to the camera.
        :param item_trace:
        """
        self.ts: Final[int] = item_ts
        self.oid: Final[int] = item_oid
        self.vrp: Final[LrVrp] = item_vrp
        self.dir: Final[Dir] = item_dir
        self.trace: Final[list] = item_trace


class LrSceneStat(Enum):
    SUCCESS = 1
    TIMEOUT = 2
    FAILURE = 3


class LrScene(list[LrSceneItem]):
    def __init__(self, sender: tuple[int, int], stat: LrSceneStat, imref: ImageRef, items: list[LrSceneItem]) -> None:
        super().__init__(items)

        self.sender: Final[tuple[int, int]] = sender
        self.stat: Final[LrSceneStat] = stat

        self.imref: Final[ImageRef] = imref


class LrTrackStat(Enum):
    UNKNOWN = 0
    DETECTED = 1
    LOST = 2


class LrTrack:
    def __init__(self,
                 source_id: tuple[int, int],
                 vrp_oids: list[int],
                 vrp_text: str,
                 vrp_template: str,
                 vpr_trace: list[LrTrackItem],
                 status: LrTrackStat,
                 vrp_dir: Dir) -> None:

        self.status: Final[LrTrackStat] = status
        self.oids: Final[list[int]] = vrp_oids
        self.dir: Final[Dir] = vrp_dir
        self.source_id: Final[tuple[int, int]] = source_id
        self.plate_text: Final[str] = vrp_text
        self.plate_template: Final[str] = vrp_template
        self.plate_trace: Final[list[LrTrackItem]] = vpr_trace


class LrTrackSet:
    def __init__(self,
                 lr_track: LrTrack,
                 lr_jnode: dict,
                 wa_jnode: dict | None,
                 alarms_jnode: list[dict] | None) -> None:

        self.lr_track: Final[LrTrack] = lr_track
        self.lr_jnode: Final[dict] = lr_jnode
        self.wa_jnode: Final[dict | None] = wa_jnode

        self.alarms_jnode: Final[list[dict] | None] = alarms_jnode

    def clone_imref(self) -> ImageRef:
        return decode__image_ref(self.lr_jnode['Image'])

    @property
    def mmr_class(self) -> str:
        vehicle: dict = self.lr_jnode.get('Vehicle') or dict()
        vehicle_type: dict = vehicle.get('type') or dict()
        return vehicle_type.get('class') or 'UNKNOWN'
